//#define LOG_NDEBUG 0
#define LOG_TAG "CameraHardware3Base"

#include <string.h>
#include <cmath>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <log/log.h>
#ifdef HAL3_HWC_BUFFER
#include <utils/backtrace.h>
#endif
#include <hardware/gralloc.h>
#include <cutils/properties.h>
#include <ui/Fence.h>
#include <ui/Rect.h>
#include <sw_sync.h>
#include <ui/GraphicBufferMapper.h>
#include "misc.h"
#include "MJPGFmt.h"
#include "Debug.h"
#include "ExifHelper.h"
#include "CameraHardware3Base.h"

#define FCC2ASCII(fmt)  ((char *)(&fmt))[0], \
    ((char *)(&fmt))[1], \
    ((char *)(&fmt))[2], \
    ((char *)(&fmt))[3]

#ifndef MIN
#define MIN(a, b)               ((a) < (b) ? (a) : (b))
#endif
#define MIN_FPS_FOR_BATCH_MODE (120)
#define PREVIEW_FPS_FOR_HFR    (30)
#define METADATA_MAP_SIZE(MAP) (sizeof(MAP)/sizeof(MAP[0]))

#define MAX_RAW_STREAMS         1
#define MAX_PROCESSED_STREAMS   2
#define MAX_STALLING_STREAMS    1
#define EMPTY_PIPELINE_DELAY    2
#define FRAME_SKIP_DELAY        0
#define PARTIAL_RESULT_COUNT    1
#define CAM_MAX_SYNC_LATENCY    4

#define DUMP_TID(x)      \
    do { \
        pthread_t ptid = pthread_self(); \
        ALOGD("[%s] TID: %lld [%s]",__FUNCTION__,(uint64_t)ptid,x); \
    } while(0)

#define IS_YUV_BUFFER(x)        (x != ANDROID_SCALER_AVAILABLE_FORMATS_BLOB && \
                                 x != HAL_PIXEL_FORMAT_RGBA_8888)

/**
 * Do not touch these compile options!
 */
#define H264_FIRST              (0)
#define NV21_FIRST              (0)
#define AE_AVAILABLE_FPS_FORMAT (0)
#define REPORT_ALL_SIZE         (0)
#define NEW_HAL_PIXEL_FORMAT    (1)
#define BLOCK_HIGH_RESOLUTION   (1)
#define DEFAULT_MAX_FPS         (1)
#define BLOCK_NG_RESOLUTIONS    (0)

using namespace android_camera_hal3;

namespace android
{

//const unsigned int kResolution[2]  = {160, 120};
const unsigned int kResolution[2]  = {2, 2};

const int64_t USEC      = 1000LL;
const int64_t MSEC      = USEC * 1000LL;
const int64_t SEC       = MSEC * 1000LL;

// fake sensor related control variables
const static int64_t exposure_time_range[2]     = {1000, 300000000};
const static int64_t frame_duration_range[2]    = {33331760, 300000000};
const static int32_t pixel_array_size[2]        = {1, 1};

// fake value
const static int32_t sensitivity_range_min = 100;
const static int32_t sensitivity_range_max = 800;

const static Res blocked_sizes[] = {
    Res(176, 144)
    , Res(432, 240)
    , Res(720, 480)
    , Res(720, 576)
};

const static int32_t available_thumbnail_sizes[] = {
    0,   0,
    160,  90,
    160, 120
};
/**
 * 3A constants
 */
const nsecs_t CameraHardware3Base::kNormalExposureTime       = 10 * MSEC;
const nsecs_t CameraHardware3Base::kFacePriorityExposureTime = 30 * MSEC;
const int     CameraHardware3Base::kNormalSensitivity        = 100;
const int     CameraHardware3Base::kFacePrioritySensitivity  = 400;
const float   CameraHardware3Base::kExposureTrackRate        = 0.1;
const int     CameraHardware3Base::kPrecaptureMinFrames      = 10;
const int     CameraHardware3Base::kStableAeMaxFrames        = 100;
const float   CameraHardware3Base::kExposureWanderMin        = -2;
const float   CameraHardware3Base::kExposureWanderMax        = 1;

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_control_effect_mode_t, cam_effect_mode_type > CameraHardware3Base::EFFECT_MODES_MAP[] = {
    { ANDROID_CONTROL_EFFECT_MODE_OFF,       CAM_EFFECT_MODE_OFF },
    { ANDROID_CONTROL_EFFECT_MODE_MONO,       CAM_EFFECT_MODE_MONO },
    { ANDROID_CONTROL_EFFECT_MODE_NEGATIVE,   CAM_EFFECT_MODE_NEGATIVE },
    { ANDROID_CONTROL_EFFECT_MODE_SOLARIZE,   CAM_EFFECT_MODE_SOLARIZE },
    { ANDROID_CONTROL_EFFECT_MODE_SEPIA,      CAM_EFFECT_MODE_SEPIA },
    { ANDROID_CONTROL_EFFECT_MODE_POSTERIZE,  CAM_EFFECT_MODE_POSTERIZE },
    { ANDROID_CONTROL_EFFECT_MODE_WHITEBOARD, CAM_EFFECT_MODE_WHITEBOARD },
    { ANDROID_CONTROL_EFFECT_MODE_BLACKBOARD, CAM_EFFECT_MODE_BLACKBOARD },
    { ANDROID_CONTROL_EFFECT_MODE_AQUA,       CAM_EFFECT_MODE_AQUA }
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_control_awb_mode_t, cam_wb_mode_type > CameraHardware3Base::WHITE_BALANCE_MODES_MAP[] = {
    { ANDROID_CONTROL_AWB_MODE_OFF,             CAM_WB_MODE_OFF },
    { ANDROID_CONTROL_AWB_MODE_AUTO,            CAM_WB_MODE_AUTO },
    { ANDROID_CONTROL_AWB_MODE_INCANDESCENT,    CAM_WB_MODE_INCANDESCENT },
    { ANDROID_CONTROL_AWB_MODE_FLUORESCENT,     CAM_WB_MODE_FLUORESCENT },
    { ANDROID_CONTROL_AWB_MODE_WARM_FLUORESCENT, CAM_WB_MODE_WARM_FLUORESCENT},
    { ANDROID_CONTROL_AWB_MODE_DAYLIGHT,        CAM_WB_MODE_DAYLIGHT },
    { ANDROID_CONTROL_AWB_MODE_CLOUDY_DAYLIGHT, CAM_WB_MODE_CLOUDY_DAYLIGHT },
    { ANDROID_CONTROL_AWB_MODE_TWILIGHT,        CAM_WB_MODE_TWILIGHT },
    { ANDROID_CONTROL_AWB_MODE_SHADE,           CAM_WB_MODE_SHADE }
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_control_scene_mode_t, cam_scene_mode_type > CameraHardware3Base::SCENE_MODES_MAP[] = {
    { ANDROID_CONTROL_SCENE_MODE_DISABLED,       CAM_SCENE_MODE_OFF },      // added by Bruce
    { ANDROID_CONTROL_SCENE_MODE_FACE_PRIORITY,  CAM_SCENE_MODE_FACE_PRIORITY },
    { ANDROID_CONTROL_SCENE_MODE_ACTION,         CAM_SCENE_MODE_ACTION },
    { ANDROID_CONTROL_SCENE_MODE_PORTRAIT,       CAM_SCENE_MODE_PORTRAIT },
    { ANDROID_CONTROL_SCENE_MODE_LANDSCAPE,      CAM_SCENE_MODE_LANDSCAPE },
    { ANDROID_CONTROL_SCENE_MODE_NIGHT,          CAM_SCENE_MODE_NIGHT },
    { ANDROID_CONTROL_SCENE_MODE_NIGHT_PORTRAIT, CAM_SCENE_MODE_NIGHT_PORTRAIT },
    { ANDROID_CONTROL_SCENE_MODE_THEATRE,        CAM_SCENE_MODE_THEATRE },
    { ANDROID_CONTROL_SCENE_MODE_BEACH,          CAM_SCENE_MODE_BEACH },
    { ANDROID_CONTROL_SCENE_MODE_SNOW,           CAM_SCENE_MODE_SNOW },
    { ANDROID_CONTROL_SCENE_MODE_SUNSET,         CAM_SCENE_MODE_SUNSET },
    { ANDROID_CONTROL_SCENE_MODE_STEADYPHOTO,    CAM_SCENE_MODE_ANTISHAKE },
    { ANDROID_CONTROL_SCENE_MODE_FIREWORKS ,     CAM_SCENE_MODE_FIREWORKS },
    { ANDROID_CONTROL_SCENE_MODE_SPORTS ,        CAM_SCENE_MODE_SPORTS },
    { ANDROID_CONTROL_SCENE_MODE_PARTY,          CAM_SCENE_MODE_PARTY },
    { ANDROID_CONTROL_SCENE_MODE_CANDLELIGHT,    CAM_SCENE_MODE_CANDLELIGHT },
    { ANDROID_CONTROL_SCENE_MODE_BARCODE,        CAM_SCENE_MODE_BARCODE}
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_control_af_mode_t, cam_focus_mode_type > CameraHardware3Base::FOCUS_MODES_MAP[] = {
    { ANDROID_CONTROL_AF_MODE_OFF,                CAM_FOCUS_MODE_OFF },
    { ANDROID_CONTROL_AF_MODE_OFF,                CAM_FOCUS_MODE_FIXED },
    { ANDROID_CONTROL_AF_MODE_AUTO,               CAM_FOCUS_MODE_AUTO },
    { ANDROID_CONTROL_AF_MODE_MACRO,              CAM_FOCUS_MODE_MACRO },
    { ANDROID_CONTROL_AF_MODE_EDOF,               CAM_FOCUS_MODE_EDOF },
    { ANDROID_CONTROL_AF_MODE_CONTINUOUS_PICTURE, CAM_FOCUS_MODE_CONTINOUS_PICTURE },
    { ANDROID_CONTROL_AF_MODE_CONTINUOUS_VIDEO,   CAM_FOCUS_MODE_CONTINOUS_VIDEO }
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_color_correction_aberration_mode_t, cam_aberration_mode_t > CameraHardware3Base::COLOR_ABERRATION_MAP[] = {
    {
        ANDROID_COLOR_CORRECTION_ABERRATION_MODE_OFF,
        CAM_COLOR_CORRECTION_ABERRATION_OFF
    },
    {
        ANDROID_COLOR_CORRECTION_ABERRATION_MODE_FAST,
        CAM_COLOR_CORRECTION_ABERRATION_FAST
    },
    {
        ANDROID_COLOR_CORRECTION_ABERRATION_MODE_HIGH_QUALITY,
        CAM_COLOR_CORRECTION_ABERRATION_HIGH_QUALITY
    },
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_control_ae_antibanding_mode_t, cam_antibanding_mode_type > CameraHardware3Base::ANTIBANDING_MODES_MAP[] = {
    { ANDROID_CONTROL_AE_ANTIBANDING_MODE_OFF,  CAM_ANTIBANDING_MODE_OFF },
    { ANDROID_CONTROL_AE_ANTIBANDING_MODE_50HZ, CAM_ANTIBANDING_MODE_50HZ },
    { ANDROID_CONTROL_AE_ANTIBANDING_MODE_60HZ, CAM_ANTIBANDING_MODE_60HZ },
    { ANDROID_CONTROL_AE_ANTIBANDING_MODE_AUTO, CAM_ANTIBANDING_MODE_AUTO }
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_control_ae_mode_t, cam_flash_mode_t > CameraHardware3Base::AE_FLASH_MODE_MAP[] = {
    { ANDROID_CONTROL_AE_MODE_OFF,                  CAM_FLASH_MODE_OFF },
    { ANDROID_CONTROL_AE_MODE_ON,                   CAM_FLASH_MODE_OFF },
    { ANDROID_CONTROL_AE_MODE_ON_AUTO_FLASH,        CAM_FLASH_MODE_AUTO},
    { ANDROID_CONTROL_AE_MODE_ON_ALWAYS_FLASH,      CAM_FLASH_MODE_ON  },
    { ANDROID_CONTROL_AE_MODE_ON_AUTO_FLASH_REDEYE, CAM_FLASH_MODE_AUTO}
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_flash_mode_t, cam_flash_mode_t > CameraHardware3Base::FLASH_MODES_MAP[] = {
    { ANDROID_FLASH_MODE_OFF,    CAM_FLASH_MODE_OFF  },
    { ANDROID_FLASH_MODE_SINGLE, CAM_FLASH_MODE_SINGLE },
    { ANDROID_FLASH_MODE_TORCH,  CAM_FLASH_MODE_TORCH }
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_statistics_face_detect_mode_t, cam_face_detect_mode_t > CameraHardware3Base::FACEDETECT_MODES_MAP[] = {
    { ANDROID_STATISTICS_FACE_DETECT_MODE_OFF,    CAM_FACE_DETECT_MODE_OFF     },
    { ANDROID_STATISTICS_FACE_DETECT_MODE_SIMPLE, CAM_FACE_DETECT_MODE_SIMPLE  },
    { ANDROID_STATISTICS_FACE_DETECT_MODE_FULL,   CAM_FACE_DETECT_MODE_FULL    }
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_lens_info_focus_distance_calibration_t, cam_focus_calibration_t > CameraHardware3Base::FOCUS_CALIBRATION_MAP[] = {
    {
        ANDROID_LENS_INFO_FOCUS_DISTANCE_CALIBRATION_UNCALIBRATED,
        CAM_FOCUS_UNCALIBRATED
    },
    {
        ANDROID_LENS_INFO_FOCUS_DISTANCE_CALIBRATION_APPROXIMATE,
        CAM_FOCUS_APPROXIMATE
    },
    {
        ANDROID_LENS_INFO_FOCUS_DISTANCE_CALIBRATION_CALIBRATED,
        CAM_FOCUS_CALIBRATED
    }
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_lens_state_t, cam_af_lens_state_t > CameraHardware3Base::LENS_STATE_MAP[] = {
    { ANDROID_LENS_STATE_STATIONARY,    CAM_AF_LENS_STATE_STATIONARY},
    { ANDROID_LENS_STATE_MOVING,        CAM_AF_LENS_STATE_MOVING}
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_sensor_test_pattern_mode_t, cam_test_pattern_mode_t > CameraHardware3Base::TEST_PATTERN_MAP[] = {
    { ANDROID_SENSOR_TEST_PATTERN_MODE_OFF,          CAM_TEST_PATTERN_OFF   },
    { ANDROID_SENSOR_TEST_PATTERN_MODE_SOLID_COLOR,  CAM_TEST_PATTERN_SOLID_COLOR },
    { ANDROID_SENSOR_TEST_PATTERN_MODE_COLOR_BARS,   CAM_TEST_PATTERN_COLOR_BARS },
    { ANDROID_SENSOR_TEST_PATTERN_MODE_COLOR_BARS_FADE_TO_GRAY, CAM_TEST_PATTERN_COLOR_BARS_FADE_TO_GRAY },
    { ANDROID_SENSOR_TEST_PATTERN_MODE_PN9,          CAM_TEST_PATTERN_PN9 },
};

const CameraHardware3Base::QCameraMap <camera_metadata_enum_android_sensor_reference_illuminant1_t, cam_illuminat_t > CameraHardware3Base::REFERENCE_ILLUMINANT_MAP[] = {
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_FLUORESCENT, CAM_AWB_WARM_FLO},
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_DAYLIGHT_FLUORESCENT, CAM_AWB_CUSTOM_DAYLIGHT },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_COOL_WHITE_FLUORESCENT, CAM_AWB_COLD_FLO },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_STANDARD_A, CAM_AWB_A },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_D55, CAM_AWB_NOON },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_D65, CAM_AWB_D65 },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_D75, CAM_AWB_D75 },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_D50, CAM_AWB_D50 },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_ISO_STUDIO_TUNGSTEN, CAM_AWB_CUSTOM_A},
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_DAYLIGHT, CAM_AWB_D50 },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_TUNGSTEN, CAM_AWB_A },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_FINE_WEATHER, CAM_AWB_D50 },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_CLOUDY_WEATHER, CAM_AWB_D65 },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_SHADE, CAM_AWB_D75 },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_DAY_WHITE_FLUORESCENT, CAM_AWB_CUSTOM_DAYLIGHT },
    { ANDROID_SENSOR_REFERENCE_ILLUMINANT1_WHITE_FLUORESCENT, CAM_AWB_COLD_FLO},
};

template <typename halType, class mapType> int lookupFwkName(const mapType* arr, size_t len, halType hal_name)
{

    for (size_t i = 0; i < len; i++) {
        if (arr[i].hal_name == hal_name) {
            return arr[i].fwk_name;
        }
    }

    /* Not able to find matching framework type is not necessarily
     * an error case. This happens when mm-camera supports more attributes
     * than the frameworks do */
    ALOGE("%s[%d]: Cannot find matching framework type", __FUNCTION__, __LINE__);
    return NAME_NOT_FOUND;
}

CameraHardware3Base::CameraHardware3Base(int camId, int v4l2Id, int facing)
    : mFormatList(),
      mSizeList(),
      mFpsList(),
      mStaticMetadata(NULL),
      mCamOpened(false),
      mCamId(camId),
      mV4l2Id(v4l2Id),
      mFacing(facing),
      mFormat(-1),
#if (DEFAULT_MAX_FPS)
      mPerferedFps(-1),
#else
      mPerferedFps(99999),
#endif
      mMaxFps(-1),
      mMinFps(0xffff),  // just give it a big value as default
      mMaxWidth(0),
      mMaxHeight(0),
      mMinWidth(9999),
      mMinHeight(9999),
      mGrallocUsage(0),
      mDumpCount(0),
      mFlushed(false),
      mDoFlush(false),
      mInputStream(NULL),
      mStatus(STATUS_ERROR),
      mRuntimeFps0(0),
      mRuntimeFps1(0)
{
    LOG_FUNCTION_NAME();
    ALOGV("%s[%d] mCamId %d v4l2Id %d", __FUNCTION__, __LINE__, mCamId, mV4l2Id);

    mGrallocUsage = CAMERA_GRALLOC_USAGE;

    memset(mDevStr, 0x0, DEV_STR_LEN);
    snprintf(mDevStr, DEV_STR_LEN, "/dev/video%d", mV4l2Id);

    mOps = NULL;
    mStatus = STATUS_ERROR;
#ifdef HAL3_HWC_BUFFER
    pGrallocWrapper = GrallocWrapper_new();
#endif
    memset(&mCamCapability, 0x0, sizeof(cam_capability_t_rtk));

    for(int i = 0; i < CAMERA3_TEMPLATE_COUNT; i++) {
        mDefaultMetadata[i] = NULL;
    }

    for (int i = 0; i < 10; i++) {
        if (queryV4L2Info(mV4l2Id) >= 0) {
            break;
        }
        usleep(500000);
    }

    initCamCap();
    initStaticMetadata();

    mFormat = choiceBestV4L2Format();
    getPerferedFps();
}

CameraHardware3Base::~CameraHardware3Base()
{
    ALOGD("~CameraHardware3Base()");
#ifdef HAL3_HWC_BUFFER
    GrallocWrapper_delete(pGrallocWrapper);
#endif
    if(mStaticMetadata != NULL) {
        free_camera_metadata(mStaticMetadata);
    }

    for(int i = 0; i < CAMERA3_TEMPLATE_COUNT; i++) {
        if(mDefaultMetadata[i] != NULL) {
            free_camera_metadata(mDefaultMetadata[i]);
        }
    }
}

void CameraHardware3Base::cleanupOnPlugout()
{
    mDoFlush = true;
}

bool CameraHardware3Base::isValidSize(int w, int h)
{
    int length = mSizeList.size();
    for (int i = 0; i < length; i++) {
        Res size = mSizeList.keyAt(i);
        if (w == size.width && h == size.height) {
            ALOGV("[%s] w %d h %d match", __FUNCTION__, w, h);
            return true;
        }
    }
    ALOGW("[%s] w %d h %d not match", __FUNCTION__, w, h);
    return false;
}

camera_metadata_t* CameraHardware3Base::getStaticInfo()
{
    LOG_FUNCTION_NAME();
    return mStaticMetadata;
}

size_t CameraHardware3Base::calcMaxJpegSize()
{
    ALOGD("CameraHardware3Base::calcMaxJpegSize");
    size_t max_jpeg_size = 0;
    size_t temp_width, temp_height;

    cam_capability_t_rtk* gCamCapability = &mCamCapability;

    size_t count = MIN(gCamCapability->picture_sizes_tbl_cnt,
                       MAX_SIZES_CNT);
    for (size_t i = 0; i < count; i++) {
        temp_width = (size_t)gCamCapability->picture_sizes_tbl[i].width;
        temp_height = (size_t)gCamCapability->picture_sizes_tbl[i].height;
        if (temp_width * temp_height > max_jpeg_size ) {
            max_jpeg_size = temp_width * temp_height;
        }
    }
    // enlarge size for exif area
    max_jpeg_size = max_jpeg_size * 1 + sizeof(camera3_jpeg_blob_t);
    //ALOGD("%s size %d",__FUNCTION__,max_jpeg_size);
    return max_jpeg_size;
}

void CameraHardware3Base::makeFPSTable(
    cam_fps_range_t* fpsTable,
    size_t size,
    size_t max_size,
    int32_t* fpsRangesTable)
{
    ALOGD("makeFPSTable size %d max_size %d", size, max_size);
    size_t j = 0;
    if (size > max_size) {
        size = max_size;
    }

    for (size_t i = 0; i < size; i++) {
        fpsRangesTable[j] = (int32_t)fpsTable[i].min_fps;
        fpsRangesTable[j + 1] = (int32_t)fpsTable[i].max_fps;
        j += 2;
    }

    // finally, add (0, max_fps)
    // for android.hardware.camera2.cts.CameraDeviceTest#testCameraDevicePreviewTemplate
    fpsRangesTable[j] = (int32_t) mMinFps;
    fpsRangesTable[j + 1] = (int32_t) mMaxFps;

#if 0
    // dump result
    for(size_t i = 0; i < ((size + 1) * 2); i++) {
        ALOGD("fpsRangesTable [%d] : %d", i, fpsRangesTable[i]);
    }
#endif
}

void CameraHardware3Base::makeTable(
    cam_dimension_t* dimTable,
    size_t size,
    size_t max_size,
    int32_t* sizeTable)
{
    ALOGD("CameraHardware3Base::makeTable");
    size_t j = 0;
    if (size > max_size) {
        size = max_size;
    }
    for (size_t i = 0; i < size; i++) {
        sizeTable[j] = dimTable[i].width;
        sizeTable[j + 1] = dimTable[i].height;
        j += 2;
    }
}

size_t CameraHardware3Base::filterJpegSizes(
    int32_t* jpegSizes,
    int32_t* processedSizes,
    size_t processedSizesCnt,
    size_t maxCount,
    cam_rect_t active_array_size,
    uint8_t downscale_factor)
{
    ALOGD("CameraHardware3Base::filterJpegSizes");
    if (0 == downscale_factor) {
        downscale_factor = 1;
    }

    int32_t min_width = active_array_size.width / downscale_factor;
    int32_t min_height = active_array_size.height / downscale_factor;
    size_t jpegSizesCnt = 0;
    if (processedSizesCnt > maxCount) {
        processedSizesCnt = maxCount;
    }
    for (size_t i = 0; i < processedSizesCnt; i += 2) {
        if (processedSizes[i] >= min_width && processedSizes[i + 1] >= min_height) {
            jpegSizes[jpegSizesCnt] = processedSizes[i];
            jpegSizes[jpegSizesCnt + 1] = processedSizes[i + 1];
            jpegSizesCnt += 2;
        }
    }
    return jpegSizesCnt;
}

void CameraHardware3Base::addStreamConfig(
    Vector<int32_t>& available_stream_configs,
    int32_t scalar_format,
    const cam_dimension_t& dim,
    int32_t config_type)
{
    ALOGV("%s[%d]: dim: [%dx%d]", __FUNCTION__, __LINE__, dim.width, dim.height);
    available_stream_configs.add(scalar_format);
    available_stream_configs.add(dim.width);
    available_stream_configs.add(dim.height);
    available_stream_configs.add(config_type);
}

void CameraHardware3Base::makeOverridesList(
    cam_scene_mode_overrides_t* overridesTable,
    size_t size,
    size_t max_size,
    uint8_t* overridesList,
    uint8_t* supported_indexes)
{
    /*daemon will give a list of overrides for all scene modes.
      However we should send the fwk only the overrides for the scene modes
      supported by the framework*/
    size_t j = 0;
    ALOGD("CameraHardware3Base::makeOverridesList");
    cam_capability_t_rtk* gCamCapability = &mCamCapability;

    if (size > max_size) {
        size = max_size;
    }
    size_t focus_count = CAM_FOCUS_MODE_MAX;
    focus_count = MIN(gCamCapability->supported_focus_modes_cnt,
                      focus_count);

    for (size_t i = 0; i < size; i++) {
        bool supt = false;
        size_t index = supported_indexes[i];
        overridesList[j] = gCamCapability->flash_available ?
                           ANDROID_CONTROL_AE_MODE_ON_AUTO_FLASH : ANDROID_CONTROL_AE_MODE_ON;
        int val = lookupFwkName(WHITE_BALANCE_MODES_MAP,
                                METADATA_MAP_SIZE(WHITE_BALANCE_MODES_MAP),
                                overridesTable[index].awb_mode);
        if (NAME_NOT_FOUND != val) {
            overridesList[j + 1] = (uint8_t)val;
        }
        uint8_t focus_override = overridesTable[index].af_mode;
        for (size_t k = 0; k < focus_count; k++) {
            if (gCamCapability->supported_focus_modes[k] == focus_override) {
                supt = true;
                break;
            }
        }
        if (supt) {
            val = lookupFwkName(FOCUS_MODES_MAP, METADATA_MAP_SIZE(FOCUS_MODES_MAP),
                                focus_override);
            if (NAME_NOT_FOUND != val) {
                overridesList[j + 2] = (uint8_t)val;
            }
        } else {
            overridesList[j + 2] = ANDROID_CONTROL_AF_MODE_OFF;
        }
        j += 3;
    }
}

camera_metadata_t* CameraHardware3Base::translateCapabilityToMetadata(int type)
{
    LOG_FUNCTION_NAME();
    cam_capability_t_rtk* gCamCapability = &mCamCapability;

    if (mDefaultMetadata[type] != NULL) {
        return mDefaultMetadata[type];
    }
    //first time we are handling this request
    //fill up the metadata structure using the wrapper class
    CameraMetadata settings;
    //translate from cam_capability_t to camera_metadata_tag_t
    static const uint8_t requestType = ANDROID_REQUEST_TYPE_CAPTURE;
    settings.update(ANDROID_REQUEST_TYPE, &requestType, 1);
    int32_t defaultRequestID = 0;
    settings.update(ANDROID_REQUEST_ID, &defaultRequestID, 1);

    const bool facingBack = gCamCapability->position == CAM_POSITION_BACK;
    ALOGV("%s[%d]: type:%d", __FUNCTION__, __LINE__, type);
#if 0
    // This is a bit hacky. EIS is enabled only when the above setprop
    // is set to non-zero value and on back camera (for 2015 Nexus).
    // Ideally, we should rely on m_bEisEnable, but we cannot guarantee
    // configureStream is called before this function. In other words,
    // we cannot guarantee the app will call configureStream before
    // calling createDefaultRequest.
    const bool eisEnabled = facingBack && eis_prop_set;
#endif

    const bool eisEnabled = false; //facingBack && eis_prop_set;
    uint8_t controlIntent = 0;
    uint8_t focusMode;
    uint8_t vsMode;
    uint8_t optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_OFF;
    uint8_t cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_OFF;
    uint8_t edge_mode;
    uint8_t noise_red_mode;
    uint8_t tonemap_mode;
    vsMode = ANDROID_CONTROL_VIDEO_STABILIZATION_MODE_OFF;
    optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_OFF;
    switch (type) {
        case CAMERA3_TEMPLATE_PREVIEW:
            controlIntent = ANDROID_CONTROL_CAPTURE_INTENT_PREVIEW;
            focusMode = ANDROID_CONTROL_AF_MODE_AUTO;
            optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_ON;
            cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_OFF; //ANDROID_COLOR_CORRECTION_ABERRATION_MODE_FAST;
            edge_mode = ANDROID_EDGE_MODE_OFF; //ANDROID_EDGE_MODE_FAST;
            noise_red_mode = ANDROID_NOISE_REDUCTION_MODE_OFF; //ANDROID_NOISE_REDUCTION_MODE_FAST;
            tonemap_mode = ANDROID_TONEMAP_MODE_FAST;
            break;
        case CAMERA3_TEMPLATE_STILL_CAPTURE:
            controlIntent = ANDROID_CONTROL_CAPTURE_INTENT_STILL_CAPTURE;
            focusMode = ANDROID_CONTROL_AF_MODE_AUTO;
            optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_ON;
            cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_OFF; //ANDROID_COLOR_CORRECTION_ABERRATION_MODE_HIGH_QUALITY;
            edge_mode = ANDROID_EDGE_MODE_HIGH_QUALITY;
            noise_red_mode = ANDROID_NOISE_REDUCTION_MODE_HIGH_QUALITY;
            tonemap_mode = ANDROID_TONEMAP_MODE_HIGH_QUALITY;
            break;
        case CAMERA3_TEMPLATE_VIDEO_RECORD:
            controlIntent = ANDROID_CONTROL_CAPTURE_INTENT_VIDEO_RECORD;
            focusMode = ANDROID_CONTROL_AF_MODE_AUTO;
            optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_OFF;
            if (eisEnabled) {
                vsMode = ANDROID_CONTROL_VIDEO_STABILIZATION_MODE_ON;
            }
            cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_OFF; //ANDROID_COLOR_CORRECTION_ABERRATION_MODE_FAST;
            edge_mode = ANDROID_EDGE_MODE_OFF; //ANDROID_EDGE_MODE_FAST;
            noise_red_mode = ANDROID_NOISE_REDUCTION_MODE_OFF; // ANDROID_NOISE_REDUCTION_MODE_FAST;
            tonemap_mode = ANDROID_TONEMAP_MODE_FAST;
            //if (forceVideoOis)
            //    optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_ON;
            break;
        case CAMERA3_TEMPLATE_VIDEO_SNAPSHOT:
            controlIntent = ANDROID_CONTROL_CAPTURE_INTENT_VIDEO_SNAPSHOT;
            focusMode = ANDROID_CONTROL_AF_MODE_CONTINUOUS_VIDEO;
            optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_OFF;
            if (eisEnabled) {
                vsMode = ANDROID_CONTROL_VIDEO_STABILIZATION_MODE_ON;
            }
            cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_FAST;
            edge_mode = ANDROID_EDGE_MODE_FAST;
            noise_red_mode = ANDROID_NOISE_REDUCTION_MODE_FAST;
            tonemap_mode = ANDROID_TONEMAP_MODE_FAST;
            //if (forceVideoOis)
            //    optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_ON;
            break;
        case CAMERA3_TEMPLATE_ZERO_SHUTTER_LAG:
            controlIntent = ANDROID_CONTROL_CAPTURE_INTENT_ZERO_SHUTTER_LAG;
            focusMode = ANDROID_CONTROL_AF_MODE_CONTINUOUS_PICTURE;
            optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_ON;
            cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_FAST;
            edge_mode = ANDROID_EDGE_MODE_ZERO_SHUTTER_LAG;
            noise_red_mode = ANDROID_NOISE_REDUCTION_MODE_ZERO_SHUTTER_LAG;
            tonemap_mode = ANDROID_TONEMAP_MODE_FAST;
            break;
        case CAMERA3_TEMPLATE_MANUAL:
            edge_mode = ANDROID_EDGE_MODE_FAST;
            noise_red_mode = ANDROID_NOISE_REDUCTION_MODE_FAST;
            //tonemap_mode = ANDROID_TONEMAP_MODE_CONTRAST_CURVE; //ANDROID_TONEMAP_MODE_FAST;
            tonemap_mode = ANDROID_TONEMAP_MODE_FAST;
            cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_FAST;
            controlIntent = ANDROID_CONTROL_CAPTURE_INTENT_MANUAL;
            focusMode = ANDROID_CONTROL_AF_MODE_OFF;
            optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_OFF;
            break;
        default:
            edge_mode = ANDROID_EDGE_MODE_FAST;
            noise_red_mode = ANDROID_NOISE_REDUCTION_MODE_FAST;
            tonemap_mode = ANDROID_TONEMAP_MODE_CONTRAST_CURVE; //ANDROID_TONEMAP_MODE_FAST;
            cacMode = ANDROID_COLOR_CORRECTION_ABERRATION_MODE_FAST;
            controlIntent = ANDROID_CONTROL_CAPTURE_INTENT_CUSTOM;
            optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_OFF;
            break;
    }
    settings.update(ANDROID_COLOR_CORRECTION_ABERRATION_MODE, &cacMode, 1);
    settings.update(ANDROID_CONTROL_CAPTURE_INTENT, &controlIntent, 1);
    settings.update(ANDROID_CONTROL_VIDEO_STABILIZATION_MODE, &vsMode, 1);

    // reset... bad coding style, but forgot it.
    focusMode = ANDROID_CONTROL_AF_MODE_AUTO;

    settings.update(ANDROID_CONTROL_AF_MODE, &focusMode, 1);

    if (gCamCapability->optical_stab_modes_count == 1 &&
        gCamCapability->optical_stab_modes[0] == CAM_OPT_STAB_ON) {
        optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_ON;
    } else if ((gCamCapability->optical_stab_modes_count == 1 &&
                gCamCapability->optical_stab_modes[0] == CAM_OPT_STAB_OFF)
               /*|| ois_disable*/) {
        optStabMode = ANDROID_LENS_OPTICAL_STABILIZATION_MODE_OFF;
    }

    settings.update(ANDROID_LENS_OPTICAL_STABILIZATION_MODE, &optStabMode, 1);

    settings.update(ANDROID_CONTROL_AE_EXPOSURE_COMPENSATION,
                    &gCamCapability->exposure_compensation_default, 1);

    static const uint8_t aeLock = ANDROID_CONTROL_AE_LOCK_OFF;
    settings.update(ANDROID_CONTROL_AE_LOCK, &aeLock, 1);

    static const uint8_t awbLock = ANDROID_CONTROL_AWB_LOCK_OFF;
    settings.update(ANDROID_CONTROL_AWB_LOCK, &awbLock, 1);

    static const uint8_t awbMode = ANDROID_CONTROL_AWB_MODE_AUTO;
    settings.update(ANDROID_CONTROL_AWB_MODE, &awbMode, 1);

    static const uint8_t controlMode = ANDROID_CONTROL_MODE_AUTO;
    settings.update(ANDROID_CONTROL_MODE, &controlMode, 1);

    static const uint8_t effectMode = ANDROID_CONTROL_EFFECT_MODE_OFF;
    settings.update(ANDROID_CONTROL_EFFECT_MODE, &effectMode, 1);

    static const uint8_t sceneMode = ANDROID_CONTROL_SCENE_MODE_FACE_PRIORITY;
    settings.update(ANDROID_CONTROL_SCENE_MODE, &sceneMode, 1);

    static const uint8_t aeMode = ANDROID_CONTROL_AE_MODE_ON;
    settings.update(ANDROID_CONTROL_AE_MODE, &aeMode, 1);

    /*flash*/
    static const uint8_t flashMode = ANDROID_FLASH_MODE_OFF;
    settings.update(ANDROID_FLASH_MODE, &flashMode, 1);

    static const uint8_t flashFiringLevel = CAM_FLASH_FIRING_LEVEL_4;
    settings.update(ANDROID_FLASH_FIRING_POWER,
                    &flashFiringLevel, 1);

    /* lens */
    float default_aperture = gCamCapability->apertures[0];
    settings.update(ANDROID_LENS_APERTURE, &default_aperture, 1);

    if (gCamCapability->filter_densities_count) {
        float default_filter_density = gCamCapability->filter_densities[0];
        settings.update(ANDROID_LENS_FILTER_DENSITY, &default_filter_density,
                        gCamCapability->filter_densities_count);
    }

    float default_focal_length = gCamCapability->focal_length;
    settings.update(ANDROID_LENS_FOCAL_LENGTH, &default_focal_length, 1);

    float default_focus_distance = 0;
    settings.update(ANDROID_LENS_FOCUS_DISTANCE, &default_focus_distance, 1);

    static const uint8_t demosaicMode = ANDROID_DEMOSAIC_MODE_FAST;
    settings.update(ANDROID_DEMOSAIC_MODE, &demosaicMode, 1);

    static const uint8_t hotpixelMode = ANDROID_HOT_PIXEL_MODE_OFF;
    settings.update(ANDROID_HOT_PIXEL_MODE, &hotpixelMode, 1);

    static const int32_t testpatternMode = ANDROID_SENSOR_TEST_PATTERN_MODE_OFF;
    settings.update(ANDROID_SENSOR_TEST_PATTERN_MODE, &testpatternMode, 1);

    /* face detection (default to OFF) */
    static const uint8_t faceDetectMode = ANDROID_STATISTICS_FACE_DETECT_MODE_OFF;
    settings.update(ANDROID_STATISTICS_FACE_DETECT_MODE, &faceDetectMode, 1);

    static const uint8_t histogramMode = ANDROID_STATISTICS_HISTOGRAM_MODE_OFF;
    settings.update(ANDROID_STATISTICS_HISTOGRAM_MODE, &histogramMode, 1);

    static const uint8_t sharpnessMapMode = ANDROID_STATISTICS_SHARPNESS_MAP_MODE_OFF;
    settings.update(ANDROID_STATISTICS_SHARPNESS_MAP_MODE, &sharpnessMapMode, 1);

    static const uint8_t hotPixelMapMode = ANDROID_STATISTICS_HOT_PIXEL_MAP_MODE_OFF;
    settings.update(ANDROID_STATISTICS_HOT_PIXEL_MAP_MODE, &hotPixelMapMode, 1);

    static const uint8_t lensShadingMode = ANDROID_STATISTICS_LENS_SHADING_MAP_MODE_OFF;
    settings.update(ANDROID_STATISTICS_LENS_SHADING_MAP_MODE, &lensShadingMode, 1);

    static const uint8_t blackLevelLock = ANDROID_BLACK_LEVEL_LOCK_OFF;
    settings.update(ANDROID_BLACK_LEVEL_LOCK, &blackLevelLock, 1);

    /* Exposure time(Update the Min Exposure Time)*/
    int64_t default_exposure_time = gCamCapability->exposure_time_range[0];
    settings.update(ANDROID_SENSOR_EXPOSURE_TIME, &default_exposure_time, 1);

    /* frame duration */
    static const int64_t default_frame_duration = NSEC_PER_33MSEC;
    settings.update(ANDROID_SENSOR_FRAME_DURATION, &default_frame_duration, 1);

    /* sensitivity */
    static const int32_t default_sensitivity = 100;
    settings.update(ANDROID_SENSOR_SENSITIVITY, &default_sensitivity, 1);

    /*edge mode*/
    settings.update(ANDROID_EDGE_MODE, &edge_mode, 1);
    /*noise reduction mode*/
    settings.update(ANDROID_NOISE_REDUCTION_MODE, &noise_red_mode, 1);

    /*color correction mode*/
    static const uint8_t color_correct_mode = ANDROID_COLOR_CORRECTION_MODE_FAST;
    settings.update(ANDROID_COLOR_CORRECTION_MODE, &color_correct_mode, 1);

    /*transform matrix mode*/
    settings.update(ANDROID_TONEMAP_MODE, &tonemap_mode, 1);

    int32_t scaler_crop_region[4];
    scaler_crop_region[0] = gCamCapability->active_array_size.left;
    scaler_crop_region[1] = gCamCapability->active_array_size.top;
    scaler_crop_region[2] = gCamCapability->active_array_size.width + gCamCapability->active_array_size.left;
    scaler_crop_region[3] = gCamCapability->active_array_size.height + gCamCapability->active_array_size.top;
    settings.update(ANDROID_SCALER_CROP_REGION, scaler_crop_region, 4);

    static const uint8_t antibanding_mode = ANDROID_CONTROL_AE_ANTIBANDING_MODE_AUTO;
    settings.update(ANDROID_CONTROL_AE_ANTIBANDING_MODE, &antibanding_mode, 1);

    /*focus distance*/
    float focus_distance = 0.0;
    settings.update(ANDROID_LENS_FOCUS_DISTANCE, &focus_distance, 1);

    /*target fps range: use maximum range for picture, and maximum fixed range for video*/
    float max_range = 0.0;
    float max_fixed_fps = 0.0;
    int32_t fps_range[2] = {0, 0};
#if 0
    for (uint32_t i = 0; i < gCamCapability->fps_ranges_tbl_cnt;
         i++) {
        float range = gCamCapability->fps_ranges_tbl[i].max_fps -
                      gCamCapability->fps_ranges_tbl[i].min_fps;
        if (type == CAMERA3_TEMPLATE_PREVIEW ||
            type == CAMERA3_TEMPLATE_STILL_CAPTURE ||
            type == CAMERA3_TEMPLATE_ZERO_SHUTTER_LAG) {
            if (range > max_range) {
                fps_range[0] =
                    (int32_t)gCamCapability->fps_ranges_tbl[i].min_fps;
                fps_range[1] =
                    (int32_t)gCamCapability->fps_ranges_tbl[i].max_fps;
                max_range = range;
            }
        } else {
            if (range < 0.01 && max_fixed_fps <
                gCamCapability->fps_ranges_tbl[i].max_fps) {
                fps_range[0] =
                    (int32_t)gCamCapability->fps_ranges_tbl[i].min_fps;
                fps_range[1] =
                    (int32_t)gCamCapability->fps_ranges_tbl[i].max_fps;
                max_fixed_fps = gCamCapability->fps_ranges_tbl[i].max_fps;
            }
        }
    }
#endif

#if 0
    if(type == CAMERA3_TEMPLATE_PREVIEW || type == CAMERA3_TEMPLATE_STILL_CAPTURE || type == CAMERA3_TEMPLATE_ZERO_SHUTTER_LAG) {
        fps_range[0] = mMinFps;
        fps_range[1] = mMaxFps;
    } else {
        fps_range[0] = mMaxFps;
        fps_range[1] = mMaxFps;
    }
#endif

    fps_range[0] = mMinFps;
    fps_range[1] = mMaxFps;

    //ALOGD("[%s][%d]fps_range: {%d %d}",__FUNCTION__,__LINE__,fps_range[0],fps_range[1]);

    settings.update(ANDROID_CONTROL_AE_TARGET_FPS_RANGE, fps_range, 2);

    /*precapture trigger*/
    uint8_t precapture_trigger = ANDROID_CONTROL_AE_PRECAPTURE_TRIGGER_IDLE;
    settings.update(ANDROID_CONTROL_AE_PRECAPTURE_TRIGGER, &precapture_trigger, 1);

    /*af trigger*/
    uint8_t af_trigger = ANDROID_CONTROL_AF_TRIGGER_IDLE;
    settings.update(ANDROID_CONTROL_AF_TRIGGER, &af_trigger, 1);

    /* ae & af regions */
    int32_t active_region[] = {
        gCamCapability->active_array_size.left,
        gCamCapability->active_array_size.top,
        gCamCapability->active_array_size.left + gCamCapability->active_array_size.width,
        gCamCapability->active_array_size.top + gCamCapability->active_array_size.height,
        0
    }; // the last key is enabled key

#if 1
    ALOGV("%s[%d]: active_region {%d %d %d %d}", __FUNCTION__, __LINE__,
          active_region[0], active_region[1], active_region[2], active_region[3]);
#endif

    settings.update(ANDROID_CONTROL_AE_REGIONS, active_region,
                    sizeof(active_region) / sizeof(active_region[0]));
    settings.update(ANDROID_CONTROL_AF_REGIONS, active_region,
                    sizeof(active_region) / sizeof(active_region[0]));
    settings.update(ANDROID_CONTROL_AWB_REGIONS, active_region,
                    sizeof(active_region) / sizeof(active_region[0]));

    /* black level lock */
    uint8_t blacklevel_lock = ANDROID_BLACK_LEVEL_LOCK_OFF;
    settings.update(ANDROID_BLACK_LEVEL_LOCK, &blacklevel_lock, 1);

    /* lens shading map mode */
    uint8_t shadingmap_mode = ANDROID_STATISTICS_LENS_SHADING_MAP_MODE_OFF;
    if (CAM_SENSOR_RAW == gCamCapability->sensor_type.sens_type) {
        shadingmap_mode = ANDROID_STATISTICS_LENS_SHADING_MAP_MODE_ON;
    }
    settings.update(ANDROID_STATISTICS_LENS_SHADING_MAP_MODE, &shadingmap_mode, 1);

    //special defaults for manual template
    if (type == CAMERA3_TEMPLATE_MANUAL) {
        static const uint8_t manualControlMode = ANDROID_CONTROL_MODE_OFF;
        settings.update(ANDROID_CONTROL_MODE, &manualControlMode, 1);

        static const uint8_t manualFocusMode = ANDROID_CONTROL_AF_MODE_OFF;
        settings.update(ANDROID_CONTROL_AF_MODE, &manualFocusMode, 1);

        static const uint8_t manualAeMode = ANDROID_CONTROL_AE_MODE_OFF;
        settings.update(ANDROID_CONTROL_AE_MODE, &manualAeMode, 1);

        static const uint8_t manualAwbMode = ANDROID_CONTROL_AWB_MODE_OFF;
        settings.update(ANDROID_CONTROL_AWB_MODE, &manualAwbMode, 1);

        // cts testCameraDeviceManualTemplate
        static const uint8_t manualTonemapMode = ANDROID_TONEMAP_MODE_FAST;
        settings.update(ANDROID_TONEMAP_MODE, &manualTonemapMode, 1);

        static const uint8_t manualColorCorrectMode = ANDROID_COLOR_CORRECTION_MODE_TRANSFORM_MATRIX;
        settings.update(ANDROID_COLOR_CORRECTION_MODE, &manualColorCorrectMode, 1);
    }

    mDefaultMetadata[type] = settings.release();
    return mDefaultMetadata[type];
}

int CameraHardware3Base::initStaticMetadata()
{
    LOG_FUNCTION_NAME();
    int rc = 0;
    CameraMetadata staticInfo;
    size_t count = 0;
    bool limitedDevice = false;
    char prop[PROPERTY_VALUE_MAX];
    // ugly workaround, but don't care
    cam_capability_t_rtk* gCamCapability = &mCamCapability;

    /* If sensor is YUV sensor (no raw support) or if per-frame control is not
     * guaranteed, its advertised as limited device */
    limitedDevice = gCamCapability->no_per_frame_control_support ||
                    (CAM_SENSOR_YUV == gCamCapability->sensor_type.sens_type);

    /*
     * Candidate values:
     * ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL_LIMITED
     * ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL_FULL
     * ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL_LEGACY
     * ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL_3
     */
#if PLATFORM_SDK_VERSION < 28
    uint8_t supportedHwLvl = ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL_LEGACY;
#else
    uint8_t supportedHwLvl = ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL_EXTERNAL;
#endif

    staticInfo.update(ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL,
                      &supportedHwLvl, 1);

    bool facingBack = gCamCapability->position == CAM_POSITION_BACK;
    /*HAL 3 only*/
    staticInfo.update(ANDROID_LENS_INFO_MINIMUM_FOCUS_DISTANCE,
                      &gCamCapability->min_focus_distance, 1);

    staticInfo.update(ANDROID_LENS_INFO_HYPERFOCAL_DISTANCE,
                      &gCamCapability->hyper_focal_distance, 1);

    /*should be using focal lengths but sensor doesn't provide that info now*/
    staticInfo.update(ANDROID_LENS_INFO_AVAILABLE_FOCAL_LENGTHS,
                      &gCamCapability->focal_length,
                      1);

    staticInfo.update(ANDROID_LENS_INFO_AVAILABLE_APERTURES,
                      gCamCapability->apertures,
                      gCamCapability->apertures_count);

    staticInfo.update(ANDROID_LENS_INFO_AVAILABLE_FILTER_DENSITIES,
                      gCamCapability->filter_densities,
                      gCamCapability->filter_densities_count);


    staticInfo.update(ANDROID_LENS_INFO_AVAILABLE_OPTICAL_STABILIZATION,
                      (uint8_t*)gCamCapability->optical_stab_modes,
                      gCamCapability->optical_stab_modes_count);

    int32_t lens_shading_map_size[] = {gCamCapability->lens_shading_map_size.width,
                                       gCamCapability->lens_shading_map_size.height
                                      };
    staticInfo.update(ANDROID_LENS_INFO_SHADING_MAP_SIZE,
                      lens_shading_map_size,
                      sizeof(lens_shading_map_size) / sizeof(int32_t));

    staticInfo.update(ANDROID_SENSOR_INFO_PHYSICAL_SIZE,
                      gCamCapability->sensor_physical_size, SENSOR_PHYSICAL_SIZE_CNT);

    staticInfo.update(ANDROID_SENSOR_INFO_EXPOSURE_TIME_RANGE,
                      gCamCapability->exposure_time_range, EXPOSURE_TIME_RANGE_CNT);

    staticInfo.update(ANDROID_SENSOR_INFO_MAX_FRAME_DURATION,
                      &gCamCapability->max_frame_duration, 1);

    // needed if support RAW
    staticInfo.update(ANDROID_SENSOR_INFO_COLOR_FILTER_ARRANGEMENT,
                      (uint8_t*)&gCamCapability->color_arrangement, 1);

#if 0
    int32_t pixel_array_size[] = {gCamCapability->pixel_array_size.width,
                                  gCamCapability->pixel_array_size.height
                                 };
#else
    int32_t pixel_array_size[] = {mMaxWidth, mMaxHeight};
#endif
    ALOGV("%s[%d]: pixel_array_size:[%dx%d]", __FUNCTION__, __LINE__, mMaxWidth, mMaxHeight);
    staticInfo.update(ANDROID_SENSOR_INFO_PIXEL_ARRAY_SIZE,
                      pixel_array_size, sizeof(pixel_array_size) / sizeof(pixel_array_size[0]));

#if 0
    int32_t active_array_size[] = {gCamCapability->active_array_size.left,
                                   gCamCapability->active_array_size.top,
                                   gCamCapability->active_array_size.width,
                                   gCamCapability->active_array_size.height
                                  };
#else

    int32_t offset = 0;
    int32_t active_array_size[] = {offset, offset, mMaxWidth - (offset * 2), mMaxHeight - (offset * 2)};
    //gCamCapability->active_array_size.top,
    //gCamCapability->active_array_size.width,
    //gCamCapability->active_array_size.height};
#endif
    ALOGV("%s[%d]: active_array_size:(%d, %d), [%dx%d]", __FUNCTION__, __LINE__, offset, offset, mMaxWidth - (offset * 2), mMaxHeight - (offset * 2));
    staticInfo.update(ANDROID_SENSOR_INFO_ACTIVE_ARRAY_SIZE,
                      active_array_size, sizeof(active_array_size) / sizeof(active_array_size[0]));

    uint8_t timestampSource = ANDROID_SENSOR_INFO_TIMESTAMP_SOURCE_REALTIME;
    staticInfo.update(ANDROID_SENSOR_INFO_TIMESTAMP_SOURCE,
                      &timestampSource, 1);


    int32_t scalar_formats[] = {
        //ANDROID_SCALER_AVAILABLE_FORMATS_YV12,
        ANDROID_SCALER_AVAILABLE_FORMATS_BLOB,
        // for hal1 callback stuff... pass CTS purpose
        ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888,
        //HAL_PIXEL_FORMAT_Y16,
        //ANDROID_SCALER_AVAILABLE_FORMATS_YCrCb_420_SP,
        ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED
    }; // StreamConfigurationMap.java : 174
    size_t scalar_formats_count = sizeof(scalar_formats) / sizeof(int32_t);

    // Bruce test CTSTODO -> seems only used in camera hal2
#if 0
    staticInfo.update(ANDROID_SCALER_AVAILABLE_FORMATS,
                      scalar_formats,
                      scalar_formats_count);

#endif

    // reporting a fake data
    int32_t available_processed_sizes[MAX_SIZES_CNT * 2];
    count = MIN(gCamCapability->picture_sizes_tbl_cnt, MAX_SIZES_CNT);
    makeTable(gCamCapability->picture_sizes_tbl,
              count, MAX_SIZES_CNT, available_processed_sizes);
    // COMMENT already been Deprecated
    staticInfo.update(ANDROID_SCALER_AVAILABLE_PROCESSED_SIZES,
                      available_processed_sizes, count * 2);


#if 1

    int32_t available_fps_ranges[MAX_SIZES_CNT * 2];
    count = MIN(gCamCapability->fps_ranges_tbl_cnt, MAX_SIZES_CNT);
    makeFPSTable(gCamCapability->fps_ranges_tbl,
                 count, MAX_SIZES_CNT, available_fps_ranges);
    count = count + 1;

#if 0
    for(int i = 0; i < (count * 2); i++) {
        ALOGD("[%s] available_fps_ranges %d", __FUNCTION__, available_fps_ranges[i]);
    }
#endif

    // For YUV_420_888 burst capture use case,
    // this list will always include (min, max) and (max, max)
    // where min <= 15 and max = the maximum output frame rate of
    // the maximum YUV_420_888 output size

    // for cts android.hardware.camera2.cts.CaptureRequestTest#testEdgeModeControl
    // for cts android.hardware.camera2.cts.CameraDeviceTest#testCameraDevicePreviewTemplate
#if (AE_AVAILABLE_FPS_FORMAT)
    int32_t fps_range_2[2];
    fps_range_2[0] = 0; //mMinFps;
    fps_range_2[1] = mMaxFps;
    staticInfo.update(ANDROID_CONTROL_AE_AVAILABLE_TARGET_FPS_RANGES,
                      fps_range_2, 2);
#else

#if 0
    ALOGD("DBG9 dump fps start ,size %d", count * 2);
    // dump fps range table
    for(int i = 0; i < (count * 2); i++) {
        ALOGD("DBG9 FPS table : %d", available_fps_ranges[i]);
    }
#endif

    // old implementation, refer to available_fps_ranges
    staticInfo.update(ANDROID_CONTROL_AE_AVAILABLE_TARGET_FPS_RANGES,
                      available_fps_ranges, count * 2);
#endif

#else

    // just for testing
    int32_t fps_range[10];
    fps_range[0] = 0;
    fps_range[1] = 0;

    fps_range[2] = 60;
    fps_range[3] = 60;

    fps_range[4] = 0;
    fps_range[5] = 60;
    staticInfo.update(ANDROID_CONTROL_AE_AVAILABLE_TARGET_FPS_RANGES,
                      fps_range, 6);

#endif

    Vector<uint8_t> availableVstabModes;
    availableVstabModes.add(ANDROID_CONTROL_VIDEO_STABILIZATION_MODE_OFF);
    staticInfo.update(ANDROID_CONTROL_AVAILABLE_VIDEO_STABILIZATION_MODES,
                      availableVstabModes.array(), availableVstabModes.size());

    /*HAL 1 and HAL 3 common*/
    // For CTS, max zoom is 1
    float maxZoom = 2;
    staticInfo.update(ANDROID_SCALER_AVAILABLE_MAX_DIGITAL_ZOOM,
                      &maxZoom, 1);

    // crop is center only
    uint8_t croppingType = ANDROID_SCALER_CROPPING_TYPE_CENTER_ONLY;
    staticInfo.update(ANDROID_SCALER_CROPPING_TYPE, &croppingType, 1);

    int32_t ae = 1;
    int32_t af = 1;
    int32_t awb = 1;

    int32_t max3aRegions[3] = {/*AE*/ae,/*AWB*/ awb,/*AF*/ af};
    if (gCamCapability->supported_focus_modes_cnt == 1) {
        //max3aRegions[2] = 0; /* AF not supported */
    }
    staticInfo.update(ANDROID_CONTROL_MAX_REGIONS,
                      max3aRegions, 3);

    /* 0: OFF, 1: OFF+SIMPLE, 2: OFF+FULL, 3: OFF+SIMPLE+FULL */
    memset(prop, 0, sizeof(prop));
    // report no face detection support
    property_get("persist.camera.facedetect", prop, "0");
    uint8_t supportedFaceDetectMode = 0; //(uint8_t)atoi(prop);
    ALOGV("%s: Support face detection mode: %d",
          __FUNCTION__, supportedFaceDetectMode);

    int32_t maxFaces = gCamCapability->max_num_roi;
    Vector<uint8_t> availableFaceDetectModes;
    availableFaceDetectModes.add(ANDROID_STATISTICS_FACE_DETECT_MODE_OFF);

    if (supportedFaceDetectMode == 1) {
        availableFaceDetectModes.add(ANDROID_STATISTICS_FACE_DETECT_MODE_SIMPLE);
    } else if (supportedFaceDetectMode == 2) {
        availableFaceDetectModes.add(ANDROID_STATISTICS_FACE_DETECT_MODE_FULL);
    } else if (supportedFaceDetectMode == 3) {
        availableFaceDetectModes.add(ANDROID_STATISTICS_FACE_DETECT_MODE_SIMPLE);
        availableFaceDetectModes.add(ANDROID_STATISTICS_FACE_DETECT_MODE_FULL);
    } else {
        maxFaces = 0;
    }
    staticInfo.update(ANDROID_STATISTICS_INFO_AVAILABLE_FACE_DETECT_MODES,
                      availableFaceDetectModes.array(),
                      availableFaceDetectModes.size());
    staticInfo.update(ANDROID_STATISTICS_INFO_MAX_FACE_COUNT,
                      (int32_t*)&maxFaces, 1);


    int32_t exposureCompensationRange[] = {gCamCapability->exposure_compensation_min,
                                           gCamCapability->exposure_compensation_max
                                          };
    staticInfo.update(ANDROID_CONTROL_AE_COMPENSATION_RANGE,
                      exposureCompensationRange,
                      sizeof(exposureCompensationRange) / sizeof(int32_t));

    // TODO FACING_ISSUE
    uint8_t lensFacing = (facingBack) ?
                         ANDROID_LENS_FACING_BACK : ANDROID_LENS_FACING_FRONT;
    staticInfo.update(ANDROID_LENS_FACING, &lensFacing, 1);

    staticInfo.update(ANDROID_JPEG_AVAILABLE_THUMBNAIL_SIZES,
                      available_thumbnail_sizes,
                      sizeof(available_thumbnail_sizes) / sizeof(int32_t));

    // (COMMENT) provides stream configurations...
    /*all sizes will be clubbed into this tag*/
    int32_t available_jpeg_sizes[MAX_SIZES_CNT * 2];
    count = MIN(gCamCapability->picture_sizes_tbl_cnt, MAX_SIZES_CNT);
    size_t jpeg_sizes_cnt = filterJpegSizes(available_jpeg_sizes, available_processed_sizes,
                                            count * 2, MAX_SIZES_CNT * 2, gCamCapability->active_array_size,
                                            gCamCapability->max_downscale_factor);

    ALOGV("%s[%d]: jpeg_sizes_cnt %d count %d, scalar_formats_count:%d", __FUNCTION__, __LINE__, jpeg_sizes_cnt, count, scalar_formats_count);

    /*android.scaler.availableStreamConfigurations*/
    size_t max_stream_configs_size = count * scalar_formats_count * 4;
    Vector<int32_t> available_stream_configs;

    /* Add input/output stream configurations for each scalar formats*/
    for (size_t j = 0; j < scalar_formats_count; j++) {
        switch (scalar_formats[j]) {
            case ANDROID_SCALER_AVAILABLE_FORMATS_RAW16:
            case ANDROID_SCALER_AVAILABLE_FORMATS_RAW_OPAQUE:
            case HAL_PIXEL_FORMAT_RAW10:
                for (size_t i = 0; i < gCamCapability->supported_raw_dim_cnt; i++) {
                    addStreamConfig(available_stream_configs, scalar_formats[j],
                                    gCamCapability->raw_dim[i],
                                    ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS_OUTPUT);
                }
                ALOGE("ANDROID_SCALER_AVAILABLE_FORMATS_RAW16 ANDROID_SCALER_AVAILABLE_FORMATS_RAW_OPAQUE -> should not reach here");
                break;
            case HAL_PIXEL_FORMAT_BLOB:
                cam_dimension_t jpeg_size;
                for (size_t i = 0; i < jpeg_sizes_cnt / 2; i++) {
                    jpeg_size.width  = available_jpeg_sizes[i * 2];
                    jpeg_size.height = available_jpeg_sizes[i * 2 + 1];
                    addStreamConfig(available_stream_configs, scalar_formats[j],
                                    jpeg_size,
                                    ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS_OUTPUT);
                }
                //ALOGE("HAL_PIXEL_FORMAT_BLOB -> should not reach here");
                break;
            //case HAL_PIXEL_FORMAT_YCbCr_420_888:
            //case HAL_PIXEL_FORMAT_IMPLEMENTATION_DEFINED:
            case ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED:
            case ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888:
            //case HAL_PIXEL_FORMAT_Y16:
            default:
                cam_dimension_t largest_picture_size;
                memset(&largest_picture_size, 0, sizeof(cam_dimension_t));
                for (size_t i = 0; i < gCamCapability->picture_sizes_tbl_cnt; i++) {

                    ALOGV("%s[%d]: add format 0x%x to OUTPUT_STREAM_CONFIG", __FUNCTION__, __LINE__, scalar_formats[j]);
                    addStreamConfig(available_stream_configs, scalar_formats[j],
                                    gCamCapability->picture_sizes_tbl[i],
                                    ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS_OUTPUT);
                    /* Book keep largest */
                    if (gCamCapability->picture_sizes_tbl[i].width
                        >= largest_picture_size.width &&
                        gCamCapability->picture_sizes_tbl[i].height
                        >= largest_picture_size.height) {
                        largest_picture_size = gCamCapability->picture_sizes_tbl[i];
                    }
                }
                /*For below 2 formats we also support i/p streams for reprocessing advertise those*/
                if (scalar_formats[j] == HAL_PIXEL_FORMAT_IMPLEMENTATION_DEFINED ||
                    scalar_formats[j] == ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED ||
                    scalar_formats[j] == ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888 ||
                    scalar_formats[j] == HAL_PIXEL_FORMAT_YCbCr_420_888) {
                    addStreamConfig(available_stream_configs, scalar_formats[j],
                                    largest_picture_size,
                                    ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS_INPUT);
                }
                break;
        }
    }


    // (COMMENT) provide ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS info
    // used in preview and record size
    // Bruce -> report ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS CTSTODO
    staticInfo.update(ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS,
                      available_stream_configs.array(), available_stream_configs.size());

    // report i/o format map, format
    // input, nums_of_output, output_1, output_2, ...
    int32_t io_format_map[] = {

        ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888, //ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED,
        2,
        ANDROID_SCALER_AVAILABLE_FORMATS_BLOB,
        ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888,

        ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED,
        2,
        ANDROID_SCALER_AVAILABLE_FORMATS_BLOB,
        ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888
    };

    staticInfo.update(ANDROID_SCALER_AVAILABLE_INPUT_OUTPUT_FORMATS_MAP,
                      io_format_map, sizeof(io_format_map) / sizeof(io_format_map[0]));

    // (COMMENT) not support hot pixel
    static const uint8_t hotpixelMode = ANDROID_HOT_PIXEL_MODE_OFF;
    staticInfo.update(ANDROID_HOT_PIXEL_MODE, &hotpixelMode, 1);

    // (COMMENT) statistics hot pixel map : off
    static const uint8_t hotPixelMapMode = ANDROID_STATISTICS_HOT_PIXEL_MAP_MODE_OFF;
    staticInfo.update(ANDROID_STATISTICS_HOT_PIXEL_MAP_MODE, &hotPixelMapMode, 1);

    /* android.scaler.availableMinFrameDurations */
    int64_t available_min_durations[max_stream_configs_size];
    size_t idx = 0;
    for (size_t j = 0; j < scalar_formats_count; j++) {
        switch (scalar_formats[j]) {
#if 0
            case ANDROID_SCALER_AVAILABLE_FORMATS_RAW16:
            case ANDROID_SCALER_AVAILABLE_FORMATS_RAW_OPAQUE:
            case HAL_PIXEL_FORMAT_RAW10:
                for (size_t i = 0; i < gCamCapability->supported_raw_dim_cnt; i++) {
                    available_min_durations[idx] = scalar_formats[j];
                    available_min_durations[idx + 1] =
                        gCamCapability->raw_dim[i].width;
                    available_min_durations[idx + 2] =
                        gCamCapability->raw_dim[i].height;
                    available_min_durations[idx + 3] =
                        gCamCapability->raw_min_duration[i];
                    idx += 4;
                }
                break;
#endif
            default:
                for (size_t i = 0; i < gCamCapability->picture_sizes_tbl_cnt; i++) {
                    available_min_durations[idx] = scalar_formats[j];
                    available_min_durations[idx + 1] =
                        gCamCapability->picture_sizes_tbl[i].width;
                    available_min_durations[idx + 2] =
                        gCamCapability->picture_sizes_tbl[i].height;
                    available_min_durations[idx + 3] =
                        gCamCapability->picture_min_duration[i];
                    idx += 4;
                }
                break;
        }
    }
    staticInfo.update(ANDROID_SCALER_AVAILABLE_MIN_FRAME_DURATIONS,
                      &available_min_durations[0], idx);

    // hfr is not supported
    Vector<int32_t> available_hfr_configs;
    for (size_t i = 0; i < gCamCapability->hfr_tbl_cnt; i++) {
        int32_t fps = 0;
        switch (gCamCapability->hfr_tbl[i].mode) {
            case CAM_HFR_MODE_60FPS:
                fps = 60;
                break;
            case CAM_HFR_MODE_90FPS:
                fps = 90;
                break;
            case CAM_HFR_MODE_120FPS:
                fps = 120;
                break;
            case CAM_HFR_MODE_150FPS:
                fps = 150;
                break;
            case CAM_HFR_MODE_180FPS:
                fps = 180;
                break;
            case CAM_HFR_MODE_210FPS:
                fps = 210;
                break;
            case CAM_HFR_MODE_240FPS:
                fps = 240;
                break;
            case CAM_HFR_MODE_480FPS:
                fps = 480;
                break;
            case CAM_HFR_MODE_OFF:
            case CAM_HFR_MODE_MAX:
            default:
                break;
        }

        // MIN_FPS_FOR_BATCH_MODE : 120

        /* Advertise only MIN_FPS_FOR_BATCH_MODE or above as HIGH_SPEED_CONFIGS */
        if (fps >= MIN_FPS_FOR_BATCH_MODE) {
            /* For each HFR frame rate, need to advertise one variable fps range
             * and one fixed fps range. Eg: for 120 FPS, advertise [30, 120] and
             * [120, 120]. While camcorder preview alone is running [30, 120] is
             * set by the app. When video recording is started, [120, 120] is
             * set. This way sensor configuration does not change when recording
             * is started */

            /* (width, height, fps_min, fps_max, batch_size_max) */
            available_hfr_configs.add(
                gCamCapability->hfr_tbl[i].dim.width);
            available_hfr_configs.add(
                gCamCapability->hfr_tbl[i].dim.height);
            available_hfr_configs.add(PREVIEW_FPS_FOR_HFR);
            available_hfr_configs.add(fps);
            available_hfr_configs.add(fps / PREVIEW_FPS_FOR_HFR);

            /* (width, height, fps_min, fps_max, batch_size_max) */
            available_hfr_configs.add(
                gCamCapability->hfr_tbl[i].dim.width);
            available_hfr_configs.add(
                gCamCapability->hfr_tbl[i].dim.height);
            available_hfr_configs.add(fps);
            available_hfr_configs.add(fps);
            available_hfr_configs.add(fps / PREVIEW_FPS_FOR_HFR);
        }
    }

    // (COMMENT) Advertise HFR capability IS OFF
    //memset(prop, 0, sizeof(prop));
    //property_get("persist.camera.hal3hfr.enable", prop, "1");
    uint8_t hfrEnable = 0; // (uint8_t)atoi(prop);

    if(hfrEnable && available_hfr_configs.array()) {
        staticInfo.update(
            ANDROID_CONTROL_AVAILABLE_HIGH_SPEED_VIDEO_CONFIGURATIONS,
            available_hfr_configs.array(), available_hfr_configs.size());
    }

    int32_t max_jpeg_size = (int32_t)calcMaxJpegSize();
    ALOGV("%s[%d]: max_jpeg_size:%i", __FUNCTION__, __LINE__, max_jpeg_size);
    staticInfo.update(ANDROID_JPEG_MAX_SIZE, &max_jpeg_size, 1);

    // (COMMENT) only support CAM_EFFECT_MODE_OFF
    uint8_t avail_effects[CAM_EFFECT_MODE_MAX];
    size_t size = 0;
    count = CAM_EFFECT_MODE_MAX;
    count = MIN(gCamCapability->supported_effects_cnt, count);
    for (size_t i = 0; i < count; i++) {
        int val = lookupFwkName(EFFECT_MODES_MAP, METADATA_MAP_SIZE(EFFECT_MODES_MAP),
                                gCamCapability->supported_effects[i]);
        if (NAME_NOT_FOUND != val) {
            avail_effects[size] = (uint8_t)val;
            size++;
        }
    }
    staticInfo.update(ANDROID_CONTROL_AVAILABLE_EFFECTS,
                      avail_effects,
                      size);

    // (COMMENT) : CAM_SCENE_MODE_OFF
    uint8_t avail_scene_modes[CAM_SCENE_MODE_MAX];
    uint8_t supported_indexes[CAM_SCENE_MODE_MAX];
    size_t supported_scene_modes_cnt = 0;
    count = CAM_SCENE_MODE_MAX;
    count = MIN(gCamCapability->supported_scene_modes_cnt, count);
    for (size_t i = 0; i < count; i++) {
        //if (gCamCapability->supported_scene_modes[i] != CAM_SCENE_MODE_OFF) {
        int val = lookupFwkName(SCENE_MODES_MAP,
                                METADATA_MAP_SIZE(SCENE_MODES_MAP),
                                gCamCapability->supported_scene_modes[i]);
        if (NAME_NOT_FOUND != val) {
            avail_scene_modes[supported_scene_modes_cnt] = (uint8_t)val;
            supported_indexes[supported_scene_modes_cnt] = (uint8_t)i;
            supported_scene_modes_cnt++;
        }
        //}
    }

    ALOGV("%s[%d]: supported_scene_modes_cnt %d", __FUNCTION__, __LINE__, supported_scene_modes_cnt);

    // so , if gCamCapability set supported_scene_modes only CAM_SCENE_MODE_OFF,
    // supported_scene_modes_cnt would be 0
    staticInfo.update(ANDROID_CONTROL_AVAILABLE_SCENE_MODES,
                      avail_scene_modes,
                      supported_scene_modes_cnt);

    // (TODO) check how overrided list works
    uint8_t scene_mode_overrides[CAM_SCENE_MODE_MAX  * 3];
    makeOverridesList(gCamCapability->scene_mode_overrides,
                      supported_scene_modes_cnt,
                      CAM_SCENE_MODE_MAX,
                      scene_mode_overrides,
                      supported_indexes);

    // Scene mode override list is an unexpected size: 3 (expected 0)
    staticInfo.update(ANDROID_CONTROL_SCENE_MODE_OVERRIDES,
                      scene_mode_overrides, supported_scene_modes_cnt * 3);

    uint8_t available_control_modes[] = {ANDROID_CONTROL_MODE_AUTO,
                                         ANDROID_CONTROL_MODE_USE_SCENE_MODE
                                        };
    //ANDROID_CONTROL_MODE_OFF,
    staticInfo.update(ANDROID_CONTROL_AVAILABLE_MODES,
                      available_control_modes,
                      2);

    // (COMMENT) anti banding set to auto
    uint8_t avail_antibanding_modes[CAM_ANTIBANDING_MODE_MAX];
    size = 0;
    count = CAM_ANTIBANDING_MODE_MAX;
    count = MIN(gCamCapability->supported_antibandings_cnt, count);
    for (size_t i = 0; i < count; i++) {
        int val = lookupFwkName(ANTIBANDING_MODES_MAP, METADATA_MAP_SIZE(ANTIBANDING_MODES_MAP),
                                gCamCapability->supported_antibandings[i]);
        if (NAME_NOT_FOUND != val) {
            avail_antibanding_modes[size] = (uint8_t)val;
            size++;
        }

    }
    staticInfo.update(ANDROID_CONTROL_AE_AVAILABLE_ANTIBANDING_MODES,
                      avail_antibanding_modes,
                      size);

    // (COMMENT) aberration_modes_count = 0
    uint8_t avail_abberation_modes[CAM_COLOR_CORRECTION_ABERRATION_MAX];
    size = 0;
    count = CAM_COLOR_CORRECTION_ABERRATION_MAX;
    count = MIN(gCamCapability->aberration_modes_count, count);
    if (0 == count) {
        avail_abberation_modes[0] =
            ANDROID_COLOR_CORRECTION_ABERRATION_MODE_OFF;
        size++;
    } else {
        for (size_t i = 0; i < count; i++) {
            int val = lookupFwkName(COLOR_ABERRATION_MAP, METADATA_MAP_SIZE(COLOR_ABERRATION_MAP),
                                    gCamCapability->aberration_modes[i]);
            if (NAME_NOT_FOUND != val) {
                avail_abberation_modes[size] = (uint8_t)val;
                size++;
            } else {
                ALOGE("%s: Invalid CAC mode %d", __func__,
                      gCamCapability->aberration_modes[i]);
                break;
            }
        }

    }
    staticInfo.update(ANDROID_COLOR_CORRECTION_AVAILABLE_ABERRATION_MODES,
                      avail_abberation_modes,
                      size);

    // (COMMENT) cam focus mode CAM_FOCUS_MODE_OFF
    uint8_t avail_af_modes[CAM_FOCUS_MODE_MAX];
    size = 0;
    count = CAM_FOCUS_MODE_MAX;
    count = MIN(gCamCapability->supported_focus_modes_cnt, count);
    for (size_t i = 0; i < count; i++) {
        int val = lookupFwkName(FOCUS_MODES_MAP, METADATA_MAP_SIZE(FOCUS_MODES_MAP),
                                gCamCapability->supported_focus_modes[i]);
        if (NAME_NOT_FOUND != val) {
            avail_af_modes[size] = (uint8_t)val;
            size++;
        }
    }
    staticInfo.update(ANDROID_CONTROL_AF_AVAILABLE_MODES,
                      avail_af_modes,
                      size);

    // (COMMENT) [CHECK] awb: auto white balence , report auto
    uint8_t avail_awb_modes[CAM_WB_MODE_MAX];
    size = 0;
    count = CAM_WB_MODE_MAX;
    count = MIN(gCamCapability->supported_white_balances_cnt, count);
    for (size_t i = 0; i < count; i++) {
        int val = lookupFwkName(WHITE_BALANCE_MODES_MAP,
                                METADATA_MAP_SIZE(WHITE_BALANCE_MODES_MAP),
                                gCamCapability->supported_white_balances[i]);
        if (NAME_NOT_FOUND != val) {
            avail_awb_modes[size] = (uint8_t)val;
            size++;
        }
    }
    staticInfo.update(ANDROID_CONTROL_AWB_AVAILABLE_MODES,
                      avail_awb_modes,
                      size);

    // (COMMENT)
    uint8_t available_flash_levels[CAM_FLASH_FIRING_LEVEL_MAX];
    count = CAM_FLASH_FIRING_LEVEL_MAX;
    count = MIN(gCamCapability->supported_flash_firing_level_cnt,
                count);
    for (size_t i = 0; i < count; i++) {
        available_flash_levels[i] =
            gCamCapability->supported_firing_levels[i];
    }
    staticInfo.update(ANDROID_FLASH_FIRING_POWER,
                      available_flash_levels, count);

    uint8_t flashAvailable;
    if (gCamCapability->flash_available) {
        flashAvailable = ANDROID_FLASH_INFO_AVAILABLE_TRUE;
    } else {
        flashAvailable = ANDROID_FLASH_INFO_AVAILABLE_FALSE;
    }
    staticInfo.update(ANDROID_FLASH_INFO_AVAILABLE,
                      &flashAvailable, 1);

    // (COMMENT) ae mode
    Vector<uint8_t> avail_ae_modes;
    count = CAM_AE_MODE_MAX;
    count = MIN(gCamCapability->supported_ae_modes_cnt, count);
    for (size_t i = 0; i < count; i++) {
        avail_ae_modes.add(gCamCapability->supported_ae_modes[i]);
    }
    if (flashAvailable) {
        avail_ae_modes.add(ANDROID_CONTROL_AE_MODE_ON_AUTO_FLASH);
        avail_ae_modes.add(ANDROID_CONTROL_AE_MODE_ON_ALWAYS_FLASH);
    }
    staticInfo.update(ANDROID_CONTROL_AE_AVAILABLE_MODES,
                      avail_ae_modes.array(),
                      avail_ae_modes.size());

    // (COMMENT) sensitivity_range (100,800)
    int32_t sensitivity_range[2];
    sensitivity_range[0] = gCamCapability->sensitivity_range.min_sensitivity;
    sensitivity_range[1] = gCamCapability->sensitivity_range.max_sensitivity;
    staticInfo.update(ANDROID_SENSOR_INFO_SENSITIVITY_RANGE,
                      sensitivity_range,
                      sizeof(sensitivity_range) / sizeof(int32_t));

    staticInfo.update(ANDROID_SENSOR_MAX_ANALOG_SENSITIVITY,
                      &gCamCapability->max_analog_sensitivity,
                      1);

    // (COMMENT) sensor_orientation:
    int32_t sensor_orientation = (int32_t)gCamCapability->sensor_mount_angle;
    staticInfo.update(ANDROID_SENSOR_ORIENTATION,
                      &sensor_orientation,
                      1);

    // (COMMENT) adjust stream id sequence.
    // raw
    // processed stream
    // jpeg stream (stalling)
#if 0
    int32_t max_output_streams[] = {
        MAX_STALLING_STREAMS,
        MAX_PROCESSED_STREAMS,
        MAX_RAW_STREAMS
    };
#else
    int32_t max_output_streams[] = {
        MAX_RAW_STREAMS,
        MAX_PROCESSED_STREAMS,
        MAX_STALLING_STREAMS
    };
#endif
    staticInfo.update(ANDROID_REQUEST_MAX_NUM_OUTPUT_STREAMS,
                      max_output_streams,
                      sizeof(max_output_streams) / sizeof(max_output_streams[0]));

    // (COMMENT) no leds
    uint8_t avail_leds = 0;
    staticInfo.update(ANDROID_LED_AVAILABLE_LEDS,
                      &avail_leds, 0);

    // (COMMENT) dist calibrated
    uint8_t focus_dist_calibrated;
    int val = lookupFwkName(FOCUS_CALIBRATION_MAP, METADATA_MAP_SIZE(FOCUS_CALIBRATION_MAP),
                            gCamCapability->focus_dist_calibrated);
    if (NAME_NOT_FOUND != val) {
        focus_dist_calibrated = (uint8_t)val;
        staticInfo.update(ANDROID_LENS_INFO_FOCUS_DISTANCE_CALIBRATION,
                          &focus_dist_calibrated, 1);
    }

    // (COMMENT)
    int32_t avail_testpattern_modes[MAX_TEST_PATTERN_CNT];
    size = 0;
    count = MIN(gCamCapability->supported_test_pattern_modes_cnt,
                MAX_TEST_PATTERN_CNT);
    for (size_t i = 0; i < count; i++) {
        int testpatternMode = lookupFwkName(TEST_PATTERN_MAP, METADATA_MAP_SIZE(TEST_PATTERN_MAP),
                                            gCamCapability->supported_test_pattern_modes[i]);
        if (NAME_NOT_FOUND != testpatternMode) {
            avail_testpattern_modes[size] = testpatternMode;
            size++;
        }
    }
    staticInfo.update(ANDROID_SENSOR_AVAILABLE_TEST_PATTERN_MODES,
                      avail_testpattern_modes,
                      size);

    // (COMMENT) (TODO) need check later
    // cts : provide a bigger pipeline depth
    uint8_t max_pipeline_depth = 8; //MAX_BUFFERS_DEQUEUE + 2;
    //(uint8_t)(MAX_INFLIGHT_REQUESTS + EMPTY_PIPELINE_DELAY + FRAME_SKIP_DELAY);
    ALOGV("ANDROID_REQUEST_PIPELINE_MAX_DEPTH %d", max_pipeline_depth);
    staticInfo.update(ANDROID_REQUEST_PIPELINE_MAX_DEPTH,
                      &max_pipeline_depth,
                      1);

    // (COMMENT) (TODO) need check later
    int32_t partial_result_count = PARTIAL_RESULT_COUNT;
    staticInfo.update(ANDROID_REQUEST_PARTIAL_RESULT_COUNT,
                      &partial_result_count,
                      1);

    // (COMMENT) (TODO) need check later
    int32_t max_stall_duration = MAX_REPROCESS_STALL;
    staticInfo.update(ANDROID_REPROCESS_MAX_CAPTURE_STALL, &max_stall_duration, 1);

    // (COMMENT) ANDROID_REQUEST_AVAILABLE_CAPABILITIES
    Vector<uint8_t> available_capabilities;

    available_capabilities.add(ANDROID_REQUEST_AVAILABLE_CAPABILITIES_BACKWARD_COMPATIBLE);
    // CTSTODO - remove REPROCESSING for Now

    //available_capabilities.add(ANDROID_REQUEST_AVAILABLE_CAPABILITIES_PRIVATE_REPROCESSING);
    available_capabilities.add(ANDROID_REQUEST_AVAILABLE_CAPABILITIES_YUV_REPROCESSING);
    //available_capabilities.add(ANDROID_REQUEST_AVAILABLE_CAPABILITIES_READ_SENSOR_SETTINGS);
    //available_capabilities.add(ANDROID_REQUEST_AVAILABLE_CAPABILITIES_CONSTRAINED_HIGH_SPEED_VIDEO);
    // (COMMENT) we do not support much capabilities
    staticInfo.update(ANDROID_REQUEST_AVAILABLE_CAPABILITIES,
                      available_capabilities.array(),
                      available_capabilities.size());

    //aeLockAvailable to be set to true if capabilities has MANUAL_SENSOR and/or
    //BURST_CAPTURE.
    uint8_t aeLockAvailable = (gCamCapability->sensor_type.sens_type == CAM_SENSOR_RAW) ?
                              ANDROID_CONTROL_AE_LOCK_AVAILABLE_TRUE : ANDROID_CONTROL_AE_LOCK_AVAILABLE_FALSE;

    staticInfo.update(ANDROID_CONTROL_AE_LOCK_AVAILABLE,
                      &aeLockAvailable, 1);

    //awbLockAvailable to be set to true if capabilities has
    //MANUAL_POST_PROCESSING and/or BURST_CAPTURE.
    uint8_t awbLockAvailable = (gCamCapability->sensor_type.sens_type == CAM_SENSOR_RAW) ?
                               ANDROID_CONTROL_AWB_LOCK_AVAILABLE_TRUE : ANDROID_CONTROL_AWB_LOCK_AVAILABLE_FALSE;

    // (COMMENT) result: ANDROID_CONTROL_AWB_LOCK_AVAILABLE_FALSE
    staticInfo.update(ANDROID_CONTROL_AWB_LOCK_AVAILABLE,
                      &awbLockAvailable, 1);

    // max input stream
    // (COMMENT) we do not want to support input stream
    // TODO: how input stream effects hal?
    // CTSTODO we have no choice to support 1 input stream
    int32_t max_input_streams = 1;
    staticInfo.update(ANDROID_REQUEST_MAX_NUM_INPUT_STREAMS,
                      &max_input_streams,
                      1);

    // (COMMENT) set by hal, choice limitedDevice
    int32_t max_latency = (limitedDevice) ?
                          CAM_MAX_SYNC_LATENCY : ANDROID_SYNC_MAX_LATENCY_PER_FRAME_CONTROL;
    staticInfo.update(ANDROID_SYNC_MAX_LATENCY,
                      &max_latency,
                      1);

    uint8_t available_hot_pixel_modes[] = {ANDROID_HOT_PIXEL_MODE_OFF};
    staticInfo.update(ANDROID_HOT_PIXEL_AVAILABLE_HOT_PIXEL_MODES,
                      available_hot_pixel_modes,
                      sizeof(available_hot_pixel_modes) / sizeof(available_hot_pixel_modes[0]));

    uint8_t available_shading_modes[] = {ANDROID_SHADING_MODE_OFF};
    staticInfo.update(ANDROID_SHADING_AVAILABLE_MODES,
                      available_shading_modes,
                      3);

    uint8_t available_lens_shading_map_modes[] = {ANDROID_STATISTICS_LENS_SHADING_MAP_MODE_OFF};
    staticInfo.update(ANDROID_STATISTICS_INFO_AVAILABLE_LENS_SHADING_MAP_MODES,
                      available_lens_shading_map_modes,
                      2);

    uint8_t available_edge_modes[] = {ANDROID_EDGE_MODE_ZERO_SHUTTER_LAG, ANDROID_EDGE_MODE_HIGH_QUALITY};

    staticInfo.update(ANDROID_EDGE_AVAILABLE_EDGE_MODES,
                      available_edge_modes,
                      sizeof(available_edge_modes) / sizeof(available_edge_modes[0]));

    uint8_t available_noise_red_modes[] = {ANDROID_NOISE_REDUCTION_MODE_ZERO_SHUTTER_LAG, ANDROID_NOISE_REDUCTION_MODE_HIGH_QUALITY};
    staticInfo.update(ANDROID_NOISE_REDUCTION_AVAILABLE_NOISE_REDUCTION_MODES,
                      available_noise_red_modes,
                      sizeof(available_noise_red_modes) / sizeof(available_noise_red_modes[0]));

    // tone map mode : fast (neet to check it later, seems only for FULL device)
    uint8_t available_tonemap_modes[] = {ANDROID_TONEMAP_MODE_FAST, ANDROID_TONEMAP_MODE_HIGH_QUALITY, ANDROID_TONEMAP_MODE_CONTRAST_CURVE};
    //uint8_t available_tonemap_modes[] = {ANDROID_TONEMAP_MODE_CONTRAST_CURVE};
    staticInfo.update(ANDROID_TONEMAP_AVAILABLE_TONE_MAP_MODES,
                      available_tonemap_modes,
                      sizeof(available_tonemap_modes) / sizeof(available_tonemap_modes[0]));

    // (COMMENT) on hot pixel map mode
    uint8_t available_hot_pixel_map_modes[] = {ANDROID_STATISTICS_HOT_PIXEL_MAP_MODE_OFF};
    staticInfo.update(ANDROID_STATISTICS_INFO_AVAILABLE_HOT_PIXEL_MAP_MODES,
                      available_hot_pixel_map_modes,
                      sizeof(available_hot_pixel_map_modes) / sizeof(available_hot_pixel_map_modes[0]));

    // BC to hal1
    camera_metadata_rational exposureCompensationStep = {
        gCamCapability->exp_compensation_step.numerator,
        gCamCapability->exp_compensation_step.denominator
    };
    staticInfo.update(ANDROID_CONTROL_AE_COMPENSATION_STEP,
                      &exposureCompensationStep, 1);


    // to fix android.hardware.camera2.cts.CaptureRequestTest#testColorCorrectionControl
#if 0
    static const camera_metadata_rational colorTransform[9] = {
        {1, 1}, {0, 1}, {0, 1},
        {0, 1}, {1, 1}, {0, 1},
        {0, 1}, {0, 1}, {1, 1}
    };
#endif

    // - ANDROID_COLOR_CORRECTION_TRANSFORM
    staticInfo.update(ANDROID_COLOR_CORRECTION_TRANSFORM, colorTransform, 9);

#if 0
    static const float colorGains[4] = {
        1.0f, 1.0f, 1.0f, 1.0f
    };
#endif

    // - ANDROID_COLOR_CORRECTION_GAINS
    staticInfo.update(ANDROID_COLOR_CORRECTION_GAINS, colorGains, 4);

#if 0
    // DEPTH_OUTPUT
    {
        static const int32_t maxDepthSamples = 100;
        staticInfo.update(ANDROID_DEPTH_MAX_DEPTH_SAMPLES, &maxDepthSamples, 1);

        static const int32_t availableDepthStreamConfigurations[] = {
            HAL_PIXEL_FORMAT_Y16, 160, 120, ANDROID_DEPTH_AVAILABLE_DEPTH_STREAM_CONFIGURATIONS_OUTPUT,
            HAL_PIXEL_FORMAT_BLOB, maxDepthSamples, 1, ANDROID_DEPTH_AVAILABLE_DEPTH_STREAM_CONFIGURATIONS_OUTPUT
        };

        staticInfo.update(ANDROID_DEPTH_AVAILABLE_DEPTH_STREAM_CONFIGURATIONS,
                          availableDepthStreamConfigurations,
                          sizeof(availableDepthStreamConfigurations) / sizeof(int32_t));
    }
#endif

    // (COMMENT) (TODO) supported request keys, need to review them later
    int32_t request_keys_basic[] = {ANDROID_COLOR_CORRECTION_MODE,
                                    ANDROID_COLOR_CORRECTION_TRANSFORM, ANDROID_COLOR_CORRECTION_GAINS,
                                    ANDROID_COLOR_CORRECTION_ABERRATION_MODE,
                                    ANDROID_CONTROL_AE_ANTIBANDING_MODE, ANDROID_CONTROL_AE_EXPOSURE_COMPENSATION,
                                    ANDROID_CONTROL_AE_LOCK, ANDROID_CONTROL_AE_MODE,
                                    ANDROID_CONTROL_AE_REGIONS, ANDROID_CONTROL_AE_TARGET_FPS_RANGE,
                                    ANDROID_CONTROL_AE_PRECAPTURE_TRIGGER, ANDROID_CONTROL_AF_MODE,
                                    ANDROID_CONTROL_AF_TRIGGER, ANDROID_CONTROL_AWB_LOCK,
                                    ANDROID_CONTROL_AWB_MODE, ANDROID_CONTROL_CAPTURE_INTENT,
                                    ANDROID_CONTROL_EFFECT_MODE, ANDROID_CONTROL_MODE,
                                    ANDROID_CONTROL_SCENE_MODE, ANDROID_CONTROL_VIDEO_STABILIZATION_MODE,
                                    ANDROID_DEMOSAIC_MODE, ANDROID_EDGE_MODE,
                                    ANDROID_FLASH_FIRING_POWER, ANDROID_FLASH_FIRING_TIME, ANDROID_FLASH_MODE,
                                    ANDROID_JPEG_GPS_COORDINATES,
                                    ANDROID_JPEG_GPS_PROCESSING_METHOD, ANDROID_JPEG_GPS_TIMESTAMP,
                                    ANDROID_JPEG_ORIENTATION, ANDROID_JPEG_QUALITY, ANDROID_JPEG_THUMBNAIL_QUALITY,
                                    ANDROID_JPEG_THUMBNAIL_SIZE, ANDROID_LENS_APERTURE, ANDROID_LENS_FILTER_DENSITY,
                                    ANDROID_LENS_FOCAL_LENGTH, ANDROID_LENS_FOCUS_DISTANCE,
                                    ANDROID_LENS_OPTICAL_STABILIZATION_MODE, ANDROID_NOISE_REDUCTION_MODE,
                                    ANDROID_REQUEST_ID, ANDROID_REQUEST_TYPE,
                                    ANDROID_SCALER_CROP_REGION, /*ANDROID_SENSOR_EXPOSURE_TIME,   should not exist in a no sensor camera */
                                    ANDROID_SENSOR_FRAME_DURATION, ANDROID_HOT_PIXEL_MODE,
                                    ANDROID_STATISTICS_HOT_PIXEL_MAP_MODE,
                                    ANDROID_SENSOR_SENSITIVITY, /* ANDROID_SHADING_MODE, */
                                    ANDROID_STATISTICS_FACE_DETECT_MODE,
                                    ANDROID_STATISTICS_HISTOGRAM_MODE, ANDROID_STATISTICS_SHARPNESS_MAP_MODE,
                                    ANDROID_STATISTICS_LENS_SHADING_MAP_MODE, ANDROID_TONEMAP_CURVE_BLUE,
                                    ANDROID_TONEMAP_CURVE_GREEN, ANDROID_TONEMAP_CURVE_RED, /* ANDROID_TONEMAP_MODE, */
#if 0
                                    ANDROID_DEPTH_MAX_DEPTH_SAMPLES,
                                    ANDROID_DEPTH_AVAILABLE_DEPTH_STREAM_CONFIGURATIONS_OUTPUT,
                                    ANDROID_DEPTH_AVAILABLE_DEPTH_STREAM_CONFIGURATIONS,
#endif
                                    ANDROID_BLACK_LEVEL_LOCK , ANDROID_CONTROL_AWB_REGIONS, ANDROID_CONTROL_AF_REGIONS
                                   };

    size_t request_keys_cnt =
        sizeof(request_keys_basic) / sizeof(request_keys_basic[0]);
    Vector<int32_t> available_request_keys;
    available_request_keys.appendArray(request_keys_basic, request_keys_cnt);
#if 0
    if (gCamCapability->supported_focus_modes_cnt > 1) {
        available_request_keys.add(ANDROID_CONTROL_AF_REGIONS);
    }
#endif

    staticInfo.update(ANDROID_REQUEST_AVAILABLE_REQUEST_KEYS,
                      available_request_keys.array(), available_request_keys.size());

    // (TODO) review these keys
    int32_t result_keys_basic[] = {ANDROID_COLOR_CORRECTION_TRANSFORM,
                                   ANDROID_COLOR_CORRECTION_GAINS, ANDROID_CONTROL_AE_MODE, ANDROID_CONTROL_AE_REGIONS,
                                   ANDROID_CONTROL_AE_STATE, ANDROID_CONTROL_AF_MODE,
                                   ANDROID_CONTROL_AF_STATE, ANDROID_CONTROL_AWB_MODE,
                                   ANDROID_CONTROL_AWB_STATE, ANDROID_CONTROL_MODE, ANDROID_EDGE_MODE,
                                   ANDROID_FLASH_FIRING_POWER, ANDROID_FLASH_FIRING_TIME, ANDROID_FLASH_MODE,
                                   ANDROID_FLASH_STATE, ANDROID_JPEG_GPS_COORDINATES, ANDROID_JPEG_GPS_PROCESSING_METHOD,
                                   ANDROID_JPEG_GPS_TIMESTAMP, ANDROID_JPEG_ORIENTATION, ANDROID_JPEG_QUALITY,
                                   ANDROID_JPEG_THUMBNAIL_QUALITY, ANDROID_JPEG_THUMBNAIL_SIZE, ANDROID_LENS_APERTURE,
                                   ANDROID_LENS_FILTER_DENSITY, ANDROID_LENS_FOCAL_LENGTH, ANDROID_LENS_FOCUS_DISTANCE,
                                   ANDROID_LENS_FOCUS_RANGE, ANDROID_LENS_STATE, ANDROID_LENS_OPTICAL_STABILIZATION_MODE,
                                   ANDROID_NOISE_REDUCTION_MODE, ANDROID_REQUEST_ID,
                                   ANDROID_SCALER_CROP_REGION, /* ANDROID_SHADING_MODE, */ /* ANDROID_SENSOR_EXPOSURE_TIME, */
                                   ANDROID_SENSOR_FRAME_DURATION, ANDROID_SENSOR_SENSITIVITY,
                                   ANDROID_SENSOR_TIMESTAMP, ANDROID_SENSOR_NEUTRAL_COLOR_POINT,
                                   ANDROID_SENSOR_PROFILE_TONE_CURVE, ANDROID_BLACK_LEVEL_LOCK, ANDROID_TONEMAP_CURVE_BLUE,
                                   ANDROID_TONEMAP_CURVE_GREEN, ANDROID_TONEMAP_CURVE_RED, /* ANDROID_TONEMAP_MODE, */
                                   ANDROID_STATISTICS_FACE_DETECT_MODE, ANDROID_STATISTICS_HISTOGRAM_MODE,
                                   ANDROID_STATISTICS_SHARPNESS_MAP, ANDROID_STATISTICS_SHARPNESS_MAP_MODE,
                                   ANDROID_STATISTICS_PREDICTED_COLOR_GAINS, ANDROID_STATISTICS_PREDICTED_COLOR_TRANSFORM,
                                   ANDROID_STATISTICS_SCENE_FLICKER, ANDROID_STATISTICS_FACE_RECTANGLES,
                                   ANDROID_STATISTICS_FACE_SCORES , ANDROID_CONTROL_AWB_REGIONS, ANDROID_CONTROL_AF_REGIONS
                                  };
    size_t result_keys_cnt =
        sizeof(result_keys_basic) / sizeof(result_keys_basic[0]);

    Vector<int32_t> available_result_keys;
    available_result_keys.appendArray(result_keys_basic, result_keys_cnt);
#if 0
    if (gCamCapability->supported_focus_modes_cnt > 1) {
        available_result_keys.add(ANDROID_CONTROL_AF_REGIONS);
    }
#endif
    if (CAM_SENSOR_YUV != gCamCapability->sensor_type.sens_type) {
        available_result_keys.add(ANDROID_SENSOR_NOISE_PROFILE);
        available_result_keys.add(ANDROID_SENSOR_GREEN_SPLIT);
    }
    if (supportedFaceDetectMode == 1) {
        available_result_keys.add(ANDROID_STATISTICS_FACE_RECTANGLES);
        available_result_keys.add(ANDROID_STATISTICS_FACE_SCORES);
    } else if ((supportedFaceDetectMode == 2) ||
               (supportedFaceDetectMode == 3)) {
        available_result_keys.add(ANDROID_STATISTICS_FACE_IDS);
        available_result_keys.add(ANDROID_STATISTICS_FACE_LANDMARKS);
    }
    staticInfo.update(ANDROID_REQUEST_AVAILABLE_RESULT_KEYS,
                      available_result_keys.array(), available_result_keys.size());

    int tonemapCurvePoints = 128;
    staticInfo.update(ANDROID_TONEMAP_MAX_CURVE_POINTS, &tonemapCurvePoints, 1);

    float tonemapCurve[4] = {
        0.f, 0.f,
        1.f, 1.f
    };

    staticInfo.update(ANDROID_TONEMAP_CURVE_RED, tonemapCurve, 4);
    staticInfo.update(ANDROID_TONEMAP_CURVE_BLUE, tonemapCurve, 4);
    staticInfo.update(ANDROID_TONEMAP_CURVE_GREEN, tonemapCurve, 4);



    // (COMMENT) (TODO) review these keys
    int32_t available_characteristics_keys[] = {ANDROID_CONTROL_AE_AVAILABLE_ANTIBANDING_MODES,
                                                ANDROID_CONTROL_AE_AVAILABLE_MODES, ANDROID_CONTROL_AE_AVAILABLE_TARGET_FPS_RANGES,
                                                ANDROID_CONTROL_AE_COMPENSATION_RANGE, ANDROID_CONTROL_AE_COMPENSATION_STEP,
                                                ANDROID_CONTROL_AF_AVAILABLE_MODES, ANDROID_CONTROL_AVAILABLE_EFFECTS,
                                                ANDROID_COLOR_CORRECTION_AVAILABLE_ABERRATION_MODES,
                                                ANDROID_SCALER_CROPPING_TYPE,
                                                ANDROID_SYNC_MAX_LATENCY,
                                                ANDROID_SENSOR_INFO_TIMESTAMP_SOURCE,
                                                ANDROID_CONTROL_AVAILABLE_SCENE_MODES,
                                                ANDROID_CONTROL_AVAILABLE_VIDEO_STABILIZATION_MODES,
                                                ANDROID_CONTROL_AWB_AVAILABLE_MODES, ANDROID_CONTROL_MAX_REGIONS,
                                                ANDROID_CONTROL_SCENE_MODE_OVERRIDES, ANDROID_FLASH_INFO_AVAILABLE,
                                                /*ANDROID_FLASH_INFO_CHARGE_DURATION,*/ ANDROID_JPEG_AVAILABLE_THUMBNAIL_SIZES,
                                                ANDROID_JPEG_MAX_SIZE, ANDROID_LENS_INFO_AVAILABLE_APERTURES,
                                                ANDROID_LENS_INFO_AVAILABLE_FILTER_DENSITIES,
                                                ANDROID_LENS_INFO_AVAILABLE_FOCAL_LENGTHS,
                                                ANDROID_LENS_INFO_AVAILABLE_OPTICAL_STABILIZATION,
                                                ANDROID_LENS_INFO_HYPERFOCAL_DISTANCE, ANDROID_LENS_INFO_MINIMUM_FOCUS_DISTANCE,
                                                ANDROID_LENS_INFO_SHADING_MAP_SIZE, ANDROID_LENS_INFO_FOCUS_DISTANCE_CALIBRATION,
                                                ANDROID_LENS_FACING,
                                                ANDROID_REQUEST_MAX_NUM_OUTPUT_STREAMS, ANDROID_REQUEST_MAX_NUM_INPUT_STREAMS,
                                                ANDROID_REQUEST_PIPELINE_MAX_DEPTH, ANDROID_REQUEST_AVAILABLE_CAPABILITIES,
                                                ANDROID_REQUEST_AVAILABLE_REQUEST_KEYS, ANDROID_REQUEST_AVAILABLE_RESULT_KEYS,
                                                ANDROID_REQUEST_AVAILABLE_CHARACTERISTICS_KEYS, ANDROID_REQUEST_PARTIAL_RESULT_COUNT,
                                                ANDROID_SCALER_AVAILABLE_MAX_DIGITAL_ZOOM,
                                                /*ANDROID_SCALER_AVAILABLE_INPUT_OUTPUT_FORMATS_MAP,*/
                                                ANDROID_SCALER_AVAILABLE_STREAM_CONFIGURATIONS,
                                                /*ANDROID_SCALER_AVAILABLE_STALL_DURATIONS,*/
                                                ANDROID_SCALER_AVAILABLE_MIN_FRAME_DURATIONS, /*ANDROID_SENSOR_FORWARD_MATRIX1,*/
                                                /*ANDROID_SENSOR_REFERENCE_ILLUMINANT1, ANDROID_SENSOR_REFERENCE_ILLUMINANT2,*/
                                                //ANDROID_SENSOR_FORWARD_MATRIX2, ANDROID_SENSOR_COLOR_TRANSFORM1,
                                                //ANDROID_SENSOR_COLOR_TRANSFORM2, ANDROID_SENSOR_CALIBRATION_TRANSFORM1,
                                                /*ANDROID_SENSOR_CALIBRATION_TRANSFORM2,*/ ANDROID_SENSOR_INFO_ACTIVE_ARRAY_SIZE,
                                                ANDROID_SENSOR_INFO_SENSITIVITY_RANGE, ANDROID_SENSOR_INFO_COLOR_FILTER_ARRANGEMENT,
                                                ANDROID_SENSOR_INFO_EXPOSURE_TIME_RANGE, ANDROID_SENSOR_INFO_MAX_FRAME_DURATION,
                                                ANDROID_SENSOR_INFO_PHYSICAL_SIZE, ANDROID_SENSOR_INFO_PIXEL_ARRAY_SIZE,
                                                /*ANDROID_SENSOR_INFO_WHITE_LEVEL,*/ /*ANDROID_SENSOR_BASE_GAIN_FACTOR,*/
                                                /*ANDROID_SENSOR_BLACK_LEVEL_PATTERN,*/ ANDROID_SENSOR_MAX_ANALOG_SENSITIVITY,
                                                ANDROID_SENSOR_ORIENTATION, ANDROID_SENSOR_AVAILABLE_TEST_PATTERN_MODES,
                                                ANDROID_STATISTICS_INFO_AVAILABLE_FACE_DETECT_MODES,
                                                /*ANDROID_STATISTICS_INFO_HISTOGRAM_BUCKET_COUNT,*/
                                                ANDROID_STATISTICS_INFO_MAX_FACE_COUNT, /*ANDROID_STATISTICS_INFO_MAX_HISTOGRAM_COUNT,*/
                                                /*ANDROID_STATISTICS_INFO_MAX_SHARPNESS_MAP_VALUE,*/
                                                /*ANDROID_STATISTICS_INFO_SHARPNESS_MAP_SIZE,*/ ANDROID_HOT_PIXEL_AVAILABLE_HOT_PIXEL_MODES,
                                                ANDROID_EDGE_AVAILABLE_EDGE_MODES,
                                                ANDROID_NOISE_REDUCTION_AVAILABLE_NOISE_REDUCTION_MODES,
                                                /*ANDROID_TONEMAP_AVAILABLE_TONE_MAP_MODES,*/
                                                ANDROID_STATISTICS_INFO_AVAILABLE_HOT_PIXEL_MAP_MODES,
                                                ANDROID_TONEMAP_MAX_CURVE_POINTS, // enabled for testToneMapControl
                                                ANDROID_CONTROL_AVAILABLE_MODES,
                                                ANDROID_CONTROL_AE_LOCK_AVAILABLE,
                                                ANDROID_CONTROL_AWB_LOCK_AVAILABLE,
                                                ANDROID_STATISTICS_INFO_AVAILABLE_LENS_SHADING_MAP_MODES,
                                                ANDROID_SHADING_AVAILABLE_MODES,
                                                ANDROID_INFO_SUPPORTED_HARDWARE_LEVEL
                                               };
    staticInfo.update(ANDROID_REQUEST_AVAILABLE_CHARACTERISTICS_KEYS,
                      available_characteristics_keys,
                      sizeof(available_characteristics_keys) / sizeof(int32_t));

#if 1
    /*available stall durations depend on the hw + sw and will be different for different devices */
    /*have to add for raw after implementation*/
    int32_t stall_formats[] = {HAL_PIXEL_FORMAT_BLOB, ANDROID_SCALER_AVAILABLE_FORMATS_RAW16};
    size_t stall_formats_count = sizeof(stall_formats) / sizeof(int32_t);

    count = MIN(gCamCapability->picture_sizes_tbl_cnt, MAX_SIZES_CNT);
    size_t raw_count = MIN(gCamCapability->supported_raw_dim_cnt,
                           MAX_SIZES_CNT);
    size_t available_stall_size = count * 4;
    int64_t available_stall_durations[available_stall_size];
    idx = 0;
    for (uint32_t j = 0; j < stall_formats_count; j++) {
        if (stall_formats[j] == HAL_PIXEL_FORMAT_BLOB) {
            for (uint32_t i = 0; i < count; i++) {
                available_stall_durations[idx]   = stall_formats[j];
                available_stall_durations[idx + 1] = gCamCapability->picture_sizes_tbl[i].width;
                available_stall_durations[idx + 2] = gCamCapability->picture_sizes_tbl[i].height;
                available_stall_durations[idx + 3] = gCamCapability->jpeg_stall_durations[i];

                ALOGV("[%s] HAL_PIXEL_FORMAT_BLOB [%d][%d] jpeg_stall_durations[%d] %lld",
                      __FUNCTION__,
                      gCamCapability->picture_sizes_tbl[i].width,
                      gCamCapability->picture_sizes_tbl[i].height,
                      i,
                      gCamCapability->jpeg_stall_durations[i]);

                idx += 4;
            }
        } else {
            for (uint32_t i = 0; i < raw_count; i++) {
                available_stall_durations[idx]   = stall_formats[j];
                available_stall_durations[idx + 1] = gCamCapability->raw_dim[i].width;
                available_stall_durations[idx + 2] = gCamCapability->raw_dim[i].height;
                available_stall_durations[idx + 3] = gCamCapability->raw16_stall_durations[i];
#if 0
                ALOGV("[%s] HAL_PIXEL_FORMAT_(other) raw16_stall_durations[%d] %d",
                      __FUNCTION__,
                      i,
                      gCamCapability->raw16_stall_durations[i]);
#endif
                idx += 4;
            }
        }
    }

    staticInfo.update(ANDROID_SCALER_AVAILABLE_STALL_DURATIONS,
                      available_stall_durations,
                      idx);
#endif
    mStaticMetadata = staticInfo.release();
    LOG_FUNCTION_END();
    return rc;
}

void CameraHardware3Base::initCamCap()
{
    LOG_FUNCTION_NAME();
    // clear camera capability struct first
    memset(&mCamCapability, 0x0, sizeof(cam_capability_t_rtk));

    cam_capability_t_rtk* gCamCapability = &mCamCapability;

    gCamCapability = &mCamCapability;
    gCamCapability->version = CAM_HAL_V3;

    // we are limited device
    gCamCapability->no_per_frame_control_support = 1;

    // define sensor info
    gCamCapability->sensor_type.sens_type = CAM_SENSOR_YUV;
    gCamCapability->sensor_type.native_format = CAM_FORMAT_YUV_420_NV12;

    // TODO: review me later
    cam_position_t pos = (mFacing == CAMERA_FACING_FRONT) ? CAM_POSITION_FRONT : CAM_POSITION_BACK;
    gCamCapability->position = pos;

    gCamCapability->min_focus_distance = 0.1f;
    gCamCapability->hyper_focal_distance = 0.0f;

    gCamCapability->focal_length = 5.0f;    // unit: mm , a meaningless values

    gCamCapability->apertures[0] = 2.8f;
    gCamCapability->apertures_count = 1;

    gCamCapability->filter_densities[0] = 0.0f;
    gCamCapability->filter_densities_count = 1;

    gCamCapability->optical_stab_modes[0] = CAM_OPT_STAB_OFF;
    gCamCapability->optical_stab_modes_count = 1;

    gCamCapability->lens_shading_map_size.width = 1;
    gCamCapability->lens_shading_map_size.height = 1;

    gCamCapability->sensor_physical_size[0] = 3.20f;
    gCamCapability->sensor_physical_size[1] = 2.40f;

    gCamCapability->exposure_time_range[0] = exposure_time_range[0];
    gCamCapability->exposure_time_range[1] = exposure_time_range[1];

    gCamCapability->max_frame_duration = frame_duration_range[1];

    gCamCapability->pixel_array_size.width = pixel_array_size[0];
    gCamCapability->pixel_array_size.height = pixel_array_size[1];

    // block to handle fps range
    {
        gCamCapability->fps_ranges_tbl_cnt = mFpsList.size();

        ALOGV("%s[%d]: fps_ranges_tbl_cnt %d", __FUNCTION__, __LINE__, gCamCapability->fps_ranges_tbl_cnt);

        int size = mFpsList.size();
        for (int i = 0; i < size; i++) {
            int fps = mFpsList.keyAt(i);
            //ALOGD("[initCamCap] fps: %d",fps);
            gCamCapability->fps_ranges_tbl[i].min_fps = fps;
            gCamCapability->fps_ranges_tbl[i].max_fps = fps;
            gCamCapability->fps_ranges_tbl[i].video_min_fps = fps;
            gCamCapability->fps_ranges_tbl[i].video_max_fps = fps;
        }
    }

    {
        int size = mSizeList.size();
        ALOGV("%s[%d]: RangeSize %d", __FUNCTION__, __LINE__, size);

        // extend picture size enum
        gCamCapability->picture_sizes_tbl_cnt = size;
        for(int i = 0; i < size; i++) {
            int w = mSizeList.keyAt(i).width;
            int h = mSizeList.keyAt(i).height;

            ALOGV("[%s] init size list [%d] w:%d h:%d", __FUNCTION__, i, w, h);

            gCamCapability->picture_sizes_tbl[i].width = w;
            gCamCapability->picture_sizes_tbl[i].height = h;

            int32_t fps = getMaxFpsOfRect(w, h);
            int64_t duration = (1000 / fps);

            ALOGV("%s[%d]: [%dx%d] duration %lld", __FUNCTION__, __LINE__, w, h, duration);

            gCamCapability->picture_min_duration[i] = (duration) * 1000 * 1000; //   ns
            gCamCapability->jpeg_stall_durations[i] = 100 * 1000 * 1000;
        }
    }

    gCamCapability->max_num_roi = 0;

    gCamCapability->exposure_compensation_min = -9;
    gCamCapability->exposure_compensation_max = 9;

    gCamCapability->supported_effects_cnt = 1;
    gCamCapability->supported_effects[0] = CAM_EFFECT_MODE_OFF;

    // COMMENT: support no scene mode
    gCamCapability->supported_scene_modes_cnt = 1;
    gCamCapability->supported_scene_modes[0] = CAM_SCENE_MODE_OFF; //CAM_SCENE_MODE_FACE_PRIORITY;

    gCamCapability->supported_antibandings_cnt = 1;
    gCamCapability->supported_antibandings[0] = CAM_ANTIBANDING_MODE_AUTO;

    gCamCapability->aberration_modes_count = 0;

    gCamCapability->supported_focus_modes_cnt = 1;
    gCamCapability->supported_focus_modes[0] = CAM_FOCUS_MODE_AUTO;

    gCamCapability->supported_white_balances_cnt = 1;
    gCamCapability->supported_white_balances[0] = CAM_WB_MODE_AUTO;

    gCamCapability->supported_flash_firing_level_cnt = 1;
    gCamCapability->supported_firing_levels[0] = CAM_FLASH_FIRING_LEVEL_0;

    gCamCapability->flash_available = 0;

    gCamCapability->supported_ae_modes_cnt = 1;
    // (COMMENT) (TODO) maybe we should report AE_MODE_ON
    gCamCapability->supported_ae_modes[0] = CAM_AE_MODE_ON;

    gCamCapability->sensitivity_range.min_sensitivity = sensitivity_range_min;
    gCamCapability->sensitivity_range.max_sensitivity = sensitivity_range_max;

    gCamCapability->max_analog_sensitivity = sensitivity_range_max;

    gCamCapability->sensor_mount_angle = 0;

    gCamCapability->focus_dist_calibrated = CAM_FOCUS_APPROXIMATE;

    gCamCapability->supported_test_pattern_modes_cnt = 1;
    gCamCapability->supported_test_pattern_modes[0] = CAM_TEST_PATTERN_OFF;

    gCamCapability->supported_raw_dim_cnt = 0;

    gCamCapability->color_arrangement = CAM_FILTER_ARRANGEMENT_YUYV;
    gCamCapability->active_array_size.left = 1;
    gCamCapability->active_array_size.top = 1;
    // TODO
    gCamCapability->active_array_size.width = 1;
    gCamCapability->active_array_size.height = 1;

    gCamCapability->white_level = 256;

    gCamCapability->exp_compensation_step.numerator = 1;
    gCamCapability->exp_compensation_step.denominator = 2;

}

/**
 * this is a very simple function to query v4l2 info which is called in a very early stage
 */
int CameraHardware3Base::queryV4L2Info(int32_t v4l2Id)
{
    LOG_FUNCTION_NAME();
    int rst = -1;
    size_t i, j, k;
    int idx, ret;
    int bestFmt;

    char device[64] = {'\0'};
    sprintf(device, "/dev/video%d", v4l2Id);

    mFormatList.clear();
    mSizeList.clear();
    mFpsList.clear();
    mFpsInfoList.clear();

    int mFd = open(device, O_RDWR);
    if(mFd == -1) {
        ALOGE("%s[%d]: open %s failed", __FUNCTION__, __LINE__, device);
        rst = -1;
        goto close_v4l2;
    }

    struct v4l2_capability cap;
    ret = ioctl (mFd, VIDIOC_QUERYCAP, &cap);
    if (ret < 0) {
        ALOGE("%s[%d]: Error opening device: unable to query device.", __FUNCTION__, __LINE__);
        rst = -1;
        goto close_v4l2;
    }

    if ((cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0) {
        ALOGE("%s[%d]: Error opening device: video capture not supported.", __FUNCTION__, __LINE__);
        rst = -1;
        goto close_v4l2;
    }

    if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
        ALOGE("%s[%d]: Capture device does not support streaming i/o", __FUNCTION__, __LINE__);
        rst = -1;
        goto close_v4l2;
    }

    // First, get supported pixel format
    rst = enumPixelFormat(mFd);

    if(rst < 0) {
        ALOGE("%s[%d]: query V4L2PixelFormat failed", __FUNCTION__, __LINE__);
        rst = -1;
        goto close_v4l2;
    }

    ALOGV("%s[%d]: enumPixelFormat done size:%d", __FUNCTION__, __LINE__, mFormatList.size());

    // then choice best format
    bestFmt = choiceBestV4L2Format();

    for (i = 0; i < mFormatList.size(); i++) {
        int fmt = mFormatList.keyAt(i);
        if(fmt != bestFmt) {
            continue;
        }

        ALOGV("%s[%d]: Device:%s get FrameSize of [%d] fmt:%c%c%c%c", __FUNCTION__, __LINE__, mDevStr, i, FCC2ASCII(fmt));
        ret = enumFrameSize(fmt, mFd);
        ALOGV("%s[%d]: enumFrameSize done size:%d", __FUNCTION__, __LINE__, mSizeList.size());
        for(j = 0; j < mSizeList.size(); j++) {
            Res size = mSizeList.keyAt(j);
            ALOGV("%s[%d]:          [%d] fmt:%c%c%c%c %dx%d", __FUNCTION__, __LINE__, j, FCC2ASCII(fmt), size.width, size.height);
            ret = enumFrameRate(fmt, size.width, size.height, mFd);
            ALOGV("%s[%d]:          enumFrameRate done size:%d", __FUNCTION__, __LINE__, mFpsList.size());

#if 0
            for(k = 0; k < mFpsList.size(); k++) {
                int fps = mFpsList.keyAt(k);
                ALOGV("[%d] fmt:%c%c%c%c %dx%d fps:%d", k, FCC2ASCII(fmt), size.width, size.height, fps);
            } //k
#endif
        } //j
    } //i

    rst = 0;

close_v4l2:
    if(mFd >= 0) {
        ALOGV("%s[%d]: close v4l2 fd:%d", __FUNCTION__, __LINE__, mFd);
        close(mFd);
    }
    return rst;
}

int CameraHardware3Base::enumPixelFormat(int v4l2fd)
{
    LOG_FUNCTION_NAME();
    int ret = 0;
    int idx;
    int fmt;
    v4l2_fmtdesc fmtdesc;

    fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    for(idx = 0; ; idx++) {
        fmtdesc.index = idx;
        ret = ioctl(v4l2fd, VIDIOC_ENUM_FMT, &fmtdesc);
        if (ret == -1) {
            if(errno == EINVAL) {
                ALOGV("%s[%d]: VIDIOC_ENUM_FMT ret:%d errno:%d, %s", __FUNCTION__, __LINE__, ret, errno, strerror(errno));
                ret = 0;
            } else {
                ALOGE("%s[%d]: VIDIOC_ENUM_FMT at index %d failed: %s", __FUNCTION__, __LINE__, idx, strerror(errno));
            }
            break;
        }

        fmt = fmtdesc.pixelformat;
        ALOGV("%s[%d]: [%d] fmt %d flags:%x fmt:%c%c%c%c name:%s", __FUNCTION__, __LINE__,
              idx, fmt, fmtdesc.flags, FCC2ASCII(fmt), fmtdesc.description);

#if 1
        char fmtBuf[8];
        sprintf(fmtBuf, "%c%c%c%c", FCC2ASCII(fmt));
        const char* fmtStr = fmtBuf;
#else
        const char* fmtStr = V4L2FmtToStr(fmt);
        if(fmtStr == NULL) {
            ALOGW("Unsupported format:%c%c%c%c", FCC2ASCII(fmt));
            continue;
        }
#endif
#if 0
        if(isCompressedFmt(fmt)) {
            ALOGW("Ignore compressed format:%s", fmtStr);
            continue;
        }
#endif

        if(mFormatList.indexOfKey(fmt) < 0) {
            mFormatList.add(fmt, String8(fmtStr));
        }
    }
    return ret;
}

int CameraHardware3Base::choiceBestV4L2Format()
{
    LOG_FUNCTION_NAME();
    int size = mFormatList.size();
    int fmt = -1;
    for (int i = 0; i < size; i++) {
#if 0 /* XChecked: only support YUYV tmp. It will support MJPEG later.*/
#if (H264_FIRST)
        if(mFormatList.keyAt(i) == V4L2_PIX_FMT_H264) {
            fmt = V4L2_PIX_FMT_H264;
            break;
        }
#endif

        // MJPG is first priority
        if(mFormatList.keyAt(i) == V4L2_PIX_FMT_MJPEG) {
            fmt = V4L2_PIX_FMT_MJPEG;
            break;
        }

#if (NV21_FIRST)
        // from 7.0, we would more like to choice NV21 as default
        if(mFormatList.keyAt(i) == V4L2_PIX_FMT_NV21) {
            fmt = V4L2_PIX_FMT_NV21;
            break;
        }
#endif

        if(mFormatList.keyAt(i) == V4L2_PIX_FMT_NV12) {
            fmt = V4L2_PIX_FMT_NV12;
            break;
        }
#endif
        if(mFormatList.keyAt(i) == V4L2_PIX_FMT_YUYV) {
            fmt = V4L2_PIX_FMT_YUYV;
            break;
        }
    }

    CHECK(fmt != -1);
    ALOGV("%s[%d]: selected fmt:%c%c%c%c", __FUNCTION__, __LINE__, FCC2ASCII(fmt));
    return fmt;
}

int CameraHardware3Base::enumFrameSize(int fmt, int v4l2fd)
{
    LOG_FUNCTION_NAME();
    int ret = 0;
    int idx;
    unsigned int width;
    unsigned int height;
    struct v4l2_frmsizeenum frmsize;

    frmsize.pixel_format = fmt;
    for(idx = 0; ; idx++) {
        frmsize.index = idx;

        ret = ioctl(v4l2fd, VIDIOC_ENUM_FRAMESIZES, &frmsize);
        if (ret == -1) {
            if (errno == EINVAL) {
                ALOGV("%s[%d]: VIDIOC_ENUM_FRAMESIZES fmt %c%c%c%c ret:%d errno:%d, %s", __FUNCTION__, __LINE__, FCC2ASCII(fmt), ret, errno, strerror(errno));
                ret = 0;
            } else {
                ALOGE("%s[%d]: VIDIOC_ENUM_FRAMESIZES at index %d failed: %s", __FUNCTION__, __LINE__, idx, strerror(errno));
            }
            break;
        }

        ALOGV("%s[%d]: frmsize type:%d", __FUNCTION__, __LINE__, frmsize.type);
        if (frmsize.type == V4L2_FRMSIZE_TYPE_DISCRETE) {
            width = frmsize.discrete.width;
            height = frmsize.discrete.height;
            ALOGV("%s[%d]: discrete [%d] fmt:%c%c%c%c [%dx%d]", __FUNCTION__, __LINE__, idx, FCC2ASCII(fmt), width, height);

#if (BLOCK_HIGH_RESOLUTION)
            // if(!((width == 1280 && height == 720) || (width == 320 && height == 240))) { // ok
            // if(!((width == 1280 && height == 720) || (width == 960 && height == 544))) { // ok
            // if(!((width == 1280 && height == 720) || (width == 640 && height == 480))) { // ok
            //if (width < 320 || height < 240 || width > 1280 || height > 720) {
            if ((width < 320 || height < 240) || (width > 640 || height > 480) || (width == 640 && height != 480)) {
                // if(height >= 1296){ // for cts-testing pass
                ALOGW("%s[%d]: Ignore size:%dx%d", __FUNCTION__, __LINE__, width, height);
                continue;
            }
#endif

#if (BLOCK_NG_RESOLUTIONS)
            if(isBlockedRes(width, height)) {
                continue;
            }
#endif

            Res size = Res(width, height);
            if(mSizeList.indexOfKey(size) < 0) {
                String8 sizeStr = String8::format("%dx%d", width, height);
                ALOGV("%s[%d]: fmt: %c%c%c%c add frameSize [%dx%d]", __FUNCTION__, __LINE__, FCC2ASCII(fmt), width, height);
                mSizeList.add(size, sizeStr);
                if((mMaxWidth * mMaxHeight) < (size.width * size.height)) {
                    mMaxWidth = size.width;
                    mMaxHeight = size.height;
                }

                if(mMinWidth * mMinHeight > size.width * size.height) {
                    mMinWidth = size.width;
                    mMinHeight = size.height;
                }
            } else {
                ALOGV("%s[%d]: fmt %c%c%c%c existing frameSize [%dx%d]", __FUNCTION__, __LINE__, FCC2ASCII(fmt), width, height);
            }
        }
    }
    ALOGV("%s[%d]:  max: [%dx%d], min: [%dx%d]", __FUNCTION__, __LINE__, mMaxWidth, mMaxHeight, mMinWidth, mMinHeight);

    return ret;
}

int CameraHardware3Base::enumFrameRate(int fmt, int width, int height, int v4l2fd)
{
    LOG_FUNCTION_NAME();
    int idx;
    int fps;
    int ret;
    struct v4l2_frmivalenum frmival;
    frmival.pixel_format = fmt;
    frmival.width = width;
    frmival.height = height;

    // TODO: review for cts
#if 0
    {
        // add fps (0,0)
        String8 fpsStr = String8::format("%d", 0);
        mFpsList.add(0, fpsStr);
    }
#endif

    for (idx = 0; ; idx++) {
        frmival.index = idx;
        ret = ioctl(v4l2fd, VIDIOC_ENUM_FRAMEINTERVALS, &frmival);
        if (ret == -1) {
            if (errno == EINVAL) {
                ALOGV("%s[%d]: VIDIOC_ENUM_FRAMEINTERVALS fmt %c%c%c%c ret:%d errno:%d, %s", __FUNCTION__, __LINE__, FCC2ASCII(fmt), ret, errno, strerror(errno));
                ret = 0;
            } else {
                ALOGE("%s[%d]: VIDIOC_ENUM_FRAMEINTERVALS at index %d failed: %s", __FUNCTION__, __LINE__, idx, strerror(errno));
            }
            break;
        }

        if(frmival.type != V4L2_FRMIVAL_TYPE_DISCRETE) {
            ALOGE("%s[%d]: unsupported frame interval type:%d", __FUNCTION__, __LINE__, frmival.type);
            return 1;
        }

        fps = frmival.discrete.denominator / frmival.discrete.numerator;

#if 0 // XCheck: support all frame rate.
        if(fps < 15) {
            ALOGW("Ignore low fps:%d", fps);
            continue;
        }
#endif

        ALOGV("%s[%d]: [%d] fmt:%c%c%c%c [%dx%d] fps:%d [%d/%d]", __FUNCTION__, __LINE__,
              idx, FCC2ASCII(fmt), width, height, fps, frmival.discrete.denominator, frmival.discrete.numerator);
        if(frmival.discrete.denominator % frmival.discrete.numerator) {
            ALOGV("%s[%d]: non-interger framerate is unsupported! [%d] fmt:%c%c%c%c %dx%d fps:%d/%d", __FUNCTION__, __LINE__,
                  idx, FCC2ASCII(fmt), width, height, frmival.discrete.denominator, frmival.discrete.numerator);
            continue;
        }

        if (mMaxFps < fps) {
            mMaxFps = fps;
            ALOGV("%s[%d]: mMaxFps %d", __FUNCTION__, __LINE__, mMaxFps);
        }

        if (mMinFps > fps) {
            mMinFps = fps;
            ALOGV("%s[%d]: mMinFps %d", __FUNCTION__, __LINE__, mMinFps);
        }

        if(mFpsList.indexOfKey(fps) < 0) {
            String8 fpsStr = String8::format("%d", fps);
            mFpsList.add(fps, fpsStr);
        }

        RECT_FPS_INFO info;
        info.width = width;
        info.height = height;
        info.fps = fps;

        mFpsInfoList.add(info);
    }

    return ret;
}

int32_t CameraHardware3Base::getMaxFpsOfRect(int32_t w, int32_t h)
{
    LOG_FUNCTION_NAME();
    int rst = -1;
    int size = mFpsInfoList.size();
    for (int i = 0; i < size; i++) {
        RECT_FPS_INFO info = mFpsInfoList[i];
        if (info.width == w && info.height == h) {
            if(info.fps > rst) {
                rst = info.fps;
            }
        }
    }

    CHECK(rst != -1);
    return rst;
}

void CameraHardware3Base::getPerferedFps()
{
    LOG_FUNCTION_NAME();
    int size = mFpsList.size();
    for(int i = 0; i < size; i++) {
        if(mFpsList.keyAt(i) > 0) {
#if (DEFAULT_MAX_FPS)
            if(mFpsList.keyAt(i) > mPerferedFps) {
#else
            if(mFpsList.keyAt(i) < mPerferedFps) {
#endif
                mPerferedFps = mFpsList.keyAt(i);
            }
        }
    }

    ALOGD("[%s] mPerferedFps:%d", __FUNCTION__, mPerferedFps);
}

status_t CameraHardware3Base::initialize(const camera3_callback_ops_t* ops)
{
    LOG_FUNCTION_NAME();
    ALOGE("%s[%d]: mStatus %d", __FUNCTION__, __LINE__, mStatus);
    mOps = ops;

#if 0
    if(mStatus != STATUS_ERROR) {
        ALOGE("%s: Already initialized! mStatus %d", __FUNCTION__, mStatus);
        return INVALID_OPERATION;
    }
#endif

    mStatus = STATUS_READY;
    return NO_ERROR;
}

const char* CameraHardware3Base::getStreamForamtString(int format)
{
    LOG_FUNCTION_NAME();
    switch(format) {
        case ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888:
            return "YCbCr";
        case ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED:
            return "ImplementDef";
        case ANDROID_SCALER_AVAILABLE_FORMATS_BLOB:
            return "Jpeg";
#ifdef HAL3_HWC_BUFFER
        case HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP:
            return "RTKYCbCr";
        case HAL_PIXEL_FORMAT_RTK_W16_H16_YCrCb_420_SP:
            return "RTKYCrCb";
#endif
        case HAL_PIXEL_FORMAT_RGBA_8888:
            return "RGBA8888";
        default:
            return "N/A";
    }
}

status_t CameraHardware3Base::process3A(CameraMetadata& settings)
{
    LOG_FUNCTION_NAME();
    /**
     * Extract top-level 3A controls
     */
    status_t res;
    bool facePriority = false;

    camera_metadata_entry e;

    // added from Android O , check fps and init
    e = settings.find(ANDROID_CONTROL_AE_TARGET_FPS_RANGE);

    // we need this field!!
    CHECK(e.count >= 2);
    mRuntimeFps0 = e.data.i32[0];
    mRuntimeFps1 = e.data.i32[1];

    ALOGV("%s[%d]: fps:mRuntimeFps0:%d,  mRuntimeFps1:%d, mPerferedFps:%d", __FUNCTION__, __LINE__, mRuntimeFps0, mRuntimeFps1, mPerferedFps);
    runtimeSetV4L2Device();

    e = settings.find(ANDROID_CONTROL_MODE);
    if (e.count == 0) {
        ALOGE("%s[%d]: No control mode entry!", __FUNCTION__, __LINE__);
        return BAD_VALUE;
    }
    uint8_t controlMode = e.data.u8[0];

    if (controlMode == ANDROID_CONTROL_MODE_OFF) {
        mAeState  = ANDROID_CONTROL_AE_STATE_INACTIVE;
        mAfState  = ANDROID_CONTROL_AF_STATE_INACTIVE;
        mAwbState = ANDROID_CONTROL_AWB_STATE_INACTIVE;
        update3A(settings);
        return OK;
    } else if (controlMode == ANDROID_CONTROL_MODE_USE_SCENE_MODE) {
#if 0
        if (!hasCapability(BACKWARD_COMPATIBLE)) {
            ALOGE("%s: Can't use scene mode when BACKWARD_COMPATIBLE not supported!",
                  __FUNCTION__);
            return BAD_VALUE;
        }
#endif
        e = settings.find(ANDROID_CONTROL_SCENE_MODE);
        if (e.count == 0) {
            ALOGE("%s[%d]: No scene mode entry!", __FUNCTION__, __LINE__);
            return BAD_VALUE;
        }
        uint8_t sceneMode = e.data.u8[0];

        switch(sceneMode) {
            case ANDROID_CONTROL_SCENE_MODE_FACE_PRIORITY:
                mFacePriority = true;
                break;
            default:
                ALOGE("%s[%d]: Emulator doesn't support scene mode %d", __FUNCTION__, __LINE__, sceneMode);
                return BAD_VALUE;
        }
    } else {
        mFacePriority = false;
    }

    // controlMode == AUTO or sceneMode = FACE_PRIORITY
    // Process individual 3A controls

    res = doFakeAE(settings);
    if (res != OK) {
        return res;
    }

    res = doFakeAF(settings);
    if (res != OK) {
        return res;
    }

    res = doFakeAWB(settings);
    if (res != OK) {
        return res;
    }

    update3A(settings);
    return OK;
}

status_t CameraHardware3Base::doFakeAE(CameraMetadata& settings)
{
    camera_metadata_entry e;

    // read original ae region settings
    e = settings.find(ANDROID_CONTROL_AE_REGIONS);
    //ALOGD("DBG0 ANDROID_CONTROL_AE_REGIONS count %d",e.count);

    int32_t org_region[5];

    for(int i = 0; i < 5; i++) {
        org_region[i] = 0;
    }

    if(e.count >= 5) {
        LOGVV("DBG0 ANDROID_CONTROL_AE_REGIONS in req %d %d %d %d %d", e.data.i32[0], e.data.i32[1], e.data.i32[2], e.data.i32[3], e.data.i32[4]);
        for(int i = 0; i < 5; i++) {
            org_region[i] = e.data.i32[i];
        }
    }

    // sync with crop region
    e = settings.find(ANDROID_SCALER_CROP_REGION);
    //ALOGD("DBG0 ANDROID_SCALER_CROP_REGION count %d",e.count);
    if(e.count >= 4) {
        LOGVV("DBG0 doFakeAE use crop region %d %d %d %d", e.data.i32[0], e.data.i32[1], e.data.i32[2], e.data.i32[3]);
        // final element refers to ANDROID_CONTROL_MAX_REGIONS
        int32_t region[5];
        region[0] = e.data.i32[0];
        region[1] = e.data.i32[1];
        region[2] = e.data.i32[2] + e.data.i32[0];
        region[3] = e.data.i32[3] + e.data.i32[1];
        region[4] = 1;
        settings.update(ANDROID_CONTROL_AE_REGIONS, region, 5);
    }

    // for cts..
    if(org_region[4] > 1) {
        settings.update(ANDROID_CONTROL_AE_REGIONS, org_region, 5);
    }

    e = settings.find(ANDROID_CONTROL_AE_MODE);
    if (e.count == 0 /*&& hasCapability(BACKWARD_COMPATIBLE)*/) {
        ALOGE("%s: No AE mode entry!", __FUNCTION__);
        return BAD_VALUE;
    }
    uint8_t aeMode = (e.count > 0) ? e.data.u8[0] : (uint8_t)ANDROID_CONTROL_AE_MODE_ON;

    switch (aeMode) {
        case ANDROID_CONTROL_AE_MODE_OFF:
            // AE is OFF
            mAeState = ANDROID_CONTROL_AE_STATE_INACTIVE;
            return OK;
        case ANDROID_CONTROL_AE_MODE_ON:
            // OK for AUTO modes
            break;
        default:
            // Mostly silently ignore unsupported modes
            ALOGV("%s[%d]: Emulator doesn't support AE mode %d, assuming ON", __FUNCTION__, __LINE__, aeMode);
            break;
    }

    e = settings.find(ANDROID_CONTROL_AE_LOCK);
    bool aeLocked = (e.count > 0) ? (e.data.u8[0] == ANDROID_CONTROL_AE_LOCK_ON) : false;

    e = settings.find(ANDROID_CONTROL_AE_PRECAPTURE_TRIGGER);
    bool precaptureTrigger = false;
    if (e.count != 0) {
        precaptureTrigger =
            (e.data.u8[0] == ANDROID_CONTROL_AE_PRECAPTURE_TRIGGER_START);
    }

    if (precaptureTrigger) {
        LOGVV("%s: Pre capture trigger = %d", __FUNCTION__, precaptureTrigger);
    } else if (e.count > 0) {
        LOGVV("%s: Pre capture trigger was present? %zu", __FUNCTION__, e.count);
    }

    if (precaptureTrigger || mAeState == ANDROID_CONTROL_AE_STATE_PRECAPTURE) {
        // Run precapture sequence
        if (mAeState != ANDROID_CONTROL_AE_STATE_PRECAPTURE) {
            mAeCounter = 0;
        }

        if (mFacePriority) {
            mAeTargetExposureTime = kFacePriorityExposureTime;
        } else {
            mAeTargetExposureTime = kNormalExposureTime;
        }

        if (mAeCounter > kPrecaptureMinFrames &&
            (mAeTargetExposureTime - mAeCurrentExposureTime) <
            mAeTargetExposureTime / 10) {
            // Done with precapture
            mAeCounter = 0;
            mAeState = aeLocked ? ANDROID_CONTROL_AE_STATE_LOCKED :
                       ANDROID_CONTROL_AE_STATE_CONVERGED;
        } else {
            // Converge some more
            mAeCurrentExposureTime +=
                (mAeTargetExposureTime - mAeCurrentExposureTime) *
                kExposureTrackRate;
            mAeCounter++;
            mAeState = ANDROID_CONTROL_AE_STATE_PRECAPTURE;
        }

    } else if (!aeLocked) {
        // Run standard occasional AE scan
        switch (mAeState) {
            case ANDROID_CONTROL_AE_STATE_CONVERGED:
            case ANDROID_CONTROL_AE_STATE_INACTIVE:
                mAeCounter++;
                if (mAeCounter > kStableAeMaxFrames) {
                    mAeTargetExposureTime =
                        mFacePriority ? kFacePriorityExposureTime :
                        kNormalExposureTime;
                    float exposureStep = ((double)random() / RAND_MAX) *
                                         (kExposureWanderMax - kExposureWanderMin) +
                                         kExposureWanderMin;
                    mAeTargetExposureTime *= std::pow(2, exposureStep);
                    mAeState = ANDROID_CONTROL_AE_STATE_SEARCHING;
                }
                break;
            case ANDROID_CONTROL_AE_STATE_SEARCHING:
                mAeCurrentExposureTime +=
                    (mAeTargetExposureTime - mAeCurrentExposureTime) *
                    kExposureTrackRate;
                if (abs(mAeTargetExposureTime - mAeCurrentExposureTime) <
                    mAeTargetExposureTime / 10) {
                    // Close enough
                    mAeState = ANDROID_CONTROL_AE_STATE_CONVERGED;
                    mAeCounter = 0;
                }
                break;
            case ANDROID_CONTROL_AE_STATE_LOCKED:
                mAeState = ANDROID_CONTROL_AE_STATE_CONVERGED;
                mAeCounter = 0;
                break;
            default:
                ALOGE("%s: Emulator in unexpected AE state %d", __FUNCTION__, mAeState);
                return INVALID_OPERATION;
        }
    } else {
        // AE is locked
        mAeState = ANDROID_CONTROL_AE_STATE_LOCKED;
    }

    return OK;
}

status_t CameraHardware3Base::doFakeAF(CameraMetadata& settings)
{
    camera_metadata_entry e;

    e = settings.find(ANDROID_CONTROL_AF_REGIONS);
    LOGVV("DBG0 ANDROID_CONTROL_AF_REGIONS count %d", e.count);

    int32_t org_region[5];

    for(int i = 0; i < 5; i++) {
        org_region[i] = 0;
    }

    if(e.count >= 5) {
        LOGVV("DBG0 ANDROID_CONTROL_AF_REGIONS in req %d %d %d %d %d",
              e.data.i32[0],
              e.data.i32[1],
              e.data.i32[2],
              e.data.i32[3],
              e.data.i32[4]);

        for(int i = 0; i < 5; i++) {
            org_region[i] = e.data.i32[i];
        }
    }
    // sync with crop region
    e = settings.find(ANDROID_SCALER_CROP_REGION);
    if(e.count >= 4) {
        LOGVV("DBG0 doFakeAF crop region %d %d %d %d", e.data.i32[0], e.data.i32[1], e.data.i32[2], e.data.i32[3]);
        int32_t region[5];
        region[0] = e.data.i32[0];
        region[1] = e.data.i32[1];
        region[2] = e.data.i32[2] + e.data.i32[0];
        region[3] = e.data.i32[3] + e.data.i32[1];
        region[4] = 1;
        settings.update(ANDROID_CONTROL_AF_REGIONS, region, 5);
    }

    // for cts..
    if(org_region[4] > 1) {
        settings.update(ANDROID_CONTROL_AF_REGIONS, org_region, 5);
    }

    e = settings.find(ANDROID_CONTROL_AF_MODE);
    if (e.count == 0 /*&& hasCapability(BACKWARD_COMPATIBLE)*/) {
        ALOGE("%s: No AF mode entry!", __FUNCTION__);
        return BAD_VALUE;
    }
    uint8_t afMode = (e.count > 0) ? e.data.u8[0] : (uint8_t)ANDROID_CONTROL_AF_MODE_OFF;

    e = settings.find(ANDROID_CONTROL_AF_TRIGGER);
    typedef camera_metadata_enum_android_control_af_trigger af_trigger_t;
    af_trigger_t afTrigger;
    if (e.count != 0) {
        afTrigger = static_cast<af_trigger_t>(e.data.u8[0]);

        LOGVV("%s: AF trigger set to 0x%x", __FUNCTION__, afTrigger);
        LOGVV("%s: AF mode is 0x%x", __FUNCTION__, afMode);
    } else {
        afTrigger = ANDROID_CONTROL_AF_TRIGGER_IDLE;
    }

    switch (afMode) {
        case ANDROID_CONTROL_AF_MODE_OFF:
            mAfState = ANDROID_CONTROL_AF_STATE_INACTIVE;
            return OK;
        case ANDROID_CONTROL_AF_MODE_AUTO:
        case ANDROID_CONTROL_AF_MODE_MACRO:
        case ANDROID_CONTROL_AF_MODE_CONTINUOUS_VIDEO:
        case ANDROID_CONTROL_AF_MODE_CONTINUOUS_PICTURE:
            // TODO: check it later
#if 0
            if (!mFacingBack) {
                ALOGE("%s: Front camera doesn't support AF mode %d",
                      __FUNCTION__, afMode);
                return BAD_VALUE;
            }
#endif
            // OK, handle transitions lower on
            break;
        default:
            ALOGE("%s: Emulator doesn't support AF mode %d",
                  __FUNCTION__, afMode);
            return BAD_VALUE;
    }

    bool afModeChanged = mAfMode != afMode;
    mAfMode = afMode;

    /**
     * Simulate AF triggers. Transition at most 1 state per frame.
     * - Focusing always succeeds (goes into locked, or PASSIVE_SCAN).
     */

    bool afTriggerStart = false;
    bool afTriggerCancel = false;
    switch (afTrigger) {
        case ANDROID_CONTROL_AF_TRIGGER_IDLE:
            break;
        case ANDROID_CONTROL_AF_TRIGGER_START:
            afTriggerStart = true;
            break;
        case ANDROID_CONTROL_AF_TRIGGER_CANCEL:
            afTriggerCancel = true;
            // Cancel trigger always transitions into INACTIVE
            mAfState = ANDROID_CONTROL_AF_STATE_INACTIVE;

            LOGVV("%s: AF State transition to STATE_INACTIVE", __FUNCTION__);

            // Stay in 'inactive' until at least next frame
            return OK;
        default:
            ALOGE("%s: Unknown af trigger value %d", __FUNCTION__, afTrigger);
            return BAD_VALUE;
    }

    // If we get down here, we're either in an autofocus mode
    // or in a continuous focus mode (and no other modes)

    int oldAfState = mAfState;
    switch (mAfState) {
        case ANDROID_CONTROL_AF_STATE_INACTIVE:
            if (afTriggerStart) {
                switch (afMode) {
                    case ANDROID_CONTROL_AF_MODE_AUTO:
                    // fall-through
                    case ANDROID_CONTROL_AF_MODE_MACRO:
                        mAfState = ANDROID_CONTROL_AF_STATE_ACTIVE_SCAN;
                        break;
                    case ANDROID_CONTROL_AF_MODE_CONTINUOUS_VIDEO:
                    // fall-through
                    case ANDROID_CONTROL_AF_MODE_CONTINUOUS_PICTURE:
                        mAfState = ANDROID_CONTROL_AF_STATE_NOT_FOCUSED_LOCKED;
                        break;
                }
            } else {
                // At least one frame stays in INACTIVE
                if (!afModeChanged) {
                    switch (afMode) {
                        case ANDROID_CONTROL_AF_MODE_CONTINUOUS_VIDEO:
                        // fall-through
                        case ANDROID_CONTROL_AF_MODE_CONTINUOUS_PICTURE:
                            mAfState = ANDROID_CONTROL_AF_STATE_PASSIVE_SCAN;
                            break;
                    }
                }
            }
            break;
        case ANDROID_CONTROL_AF_STATE_PASSIVE_SCAN:
            /**
             * When the AF trigger is activated, the algorithm should finish
             * its PASSIVE_SCAN if active, and then transition into AF_FOCUSED
             * or AF_NOT_FOCUSED as appropriate
             */
            if (afTriggerStart) {
                // Randomly transition to focused or not focused
                if (random() % 3) {
                    mAfState = ANDROID_CONTROL_AF_STATE_FOCUSED_LOCKED;
                } else {
                    mAfState = ANDROID_CONTROL_AF_STATE_NOT_FOCUSED_LOCKED;
                }
            }
            /**
             * When the AF trigger is not involved, the AF algorithm should
             * start in INACTIVE state, and then transition into PASSIVE_SCAN
             * and PASSIVE_FOCUSED states
             */
            else if (!afTriggerCancel) {
                // Randomly transition to passive focus
                if (rand() % 3 == 0) {
                    mAfState = ANDROID_CONTROL_AF_STATE_PASSIVE_FOCUSED;
                }
            }

            break;
        case ANDROID_CONTROL_AF_STATE_PASSIVE_FOCUSED:
            if (afTriggerStart) {
                // Randomly transition to focused or not focused
                if (rand() % 3) {
                    mAfState = ANDROID_CONTROL_AF_STATE_FOCUSED_LOCKED;
                } else {
                    mAfState = ANDROID_CONTROL_AF_STATE_NOT_FOCUSED_LOCKED;
                }
            }
            // TODO: initiate passive scan (PASSIVE_SCAN)
            break;
        case ANDROID_CONTROL_AF_STATE_ACTIVE_SCAN:
            // Simulate AF sweep completing instantaneously

            // Randomly transition to focused or not focused
            if (rand() % 3) {
                mAfState = ANDROID_CONTROL_AF_STATE_FOCUSED_LOCKED;
            } else {
                mAfState = ANDROID_CONTROL_AF_STATE_NOT_FOCUSED_LOCKED;
            }
            break;
        case ANDROID_CONTROL_AF_STATE_FOCUSED_LOCKED:
            if (afTriggerStart) {
                switch (afMode) {
                    case ANDROID_CONTROL_AF_MODE_AUTO:
                    // fall-through
                    case ANDROID_CONTROL_AF_MODE_MACRO:
                        mAfState = ANDROID_CONTROL_AF_STATE_ACTIVE_SCAN;
                        break;
                    case ANDROID_CONTROL_AF_MODE_CONTINUOUS_VIDEO:
                    // fall-through
                    case ANDROID_CONTROL_AF_MODE_CONTINUOUS_PICTURE:
                        // continuous autofocus => trigger start has no effect
                        break;
                }
            }
            break;
        case ANDROID_CONTROL_AF_STATE_NOT_FOCUSED_LOCKED:
            if (afTriggerStart) {
                switch (afMode) {
                    case ANDROID_CONTROL_AF_MODE_AUTO:
                    // fall-through
                    case ANDROID_CONTROL_AF_MODE_MACRO:
                        mAfState = ANDROID_CONTROL_AF_STATE_ACTIVE_SCAN;
                        break;
                    case ANDROID_CONTROL_AF_MODE_CONTINUOUS_VIDEO:
                    // fall-through
                    case ANDROID_CONTROL_AF_MODE_CONTINUOUS_PICTURE:
                        // continuous autofocus => trigger start has no effect
                        break;
                }
            }
            break;
        default:
            ALOGE("%s: Bad af state %d", __FUNCTION__, mAfState);
    }

    {
        char afStateString[100] = {0,};
        camera_metadata_enum_snprint(ANDROID_CONTROL_AF_STATE,
                                     oldAfState,
                                     afStateString,
                                     sizeof(afStateString));

        char afNewStateString[100] = {0,};
        camera_metadata_enum_snprint(ANDROID_CONTROL_AF_STATE,
                                     mAfState,
                                     afNewStateString,
                                     sizeof(afNewStateString));
        LOGVV("%s: AF state transitioned from %s to %s",
              __FUNCTION__, afStateString, afNewStateString);
    }

    return OK;
}

status_t CameraHardware3Base::doFakeAWB(CameraMetadata& settings)
{
    camera_metadata_entry e;

    e = settings.find(ANDROID_CONTROL_AWB_REGIONS);
    LOGVV("DBG0 ANDROID_CONTROL_AWB_REGIONS count %d", e.count);

    int32_t org_region[5];

    for(int i = 0; i < 5; i++) {
        org_region[i] = 0;
    }

    if(e.count >= 5) {
        LOGVV("DBG0 ANDROID_CONTROL_AWB_REGIONS in req %d %d %d %d %d",
              e.data.i32[0],
              e.data.i32[1],
              e.data.i32[2],
              e.data.i32[3],
              e.data.i32[4]);

        for(int i = 0; i < 5; i++) {
            org_region[i] = e.data.i32[i];
        }
    }

    // sync with crop region
    e = settings.find(ANDROID_SCALER_CROP_REGION);
    if(e.count >= 4) {
        LOGVV("DBG0 doFakeAWB crop region %d %d %d %d", e.data.i32[0], e.data.i32[1], e.data.i32[2], e.data.i32[3]);
        int32_t region[5];
        region[0] = e.data.i32[0];
        region[1] = e.data.i32[1];
        region[2] = e.data.i32[2] + e.data.i32[0];
        region[3] = e.data.i32[3] + e.data.i32[1];
        region[4] = 1;
        settings.update(ANDROID_CONTROL_AWB_REGIONS, region, 5);
    }

    // for cts..
    if(org_region[4] > 1) {
        settings.update(ANDROID_CONTROL_AWB_REGIONS, org_region, 5);
    }

    e = settings.find(ANDROID_CONTROL_AWB_MODE);
    if (e.count == 0 /*&& hasCapability(BACKWARD_COMPATIBLE)*/) {
        ALOGE("%s: No AWB mode entry!", __FUNCTION__);
        return BAD_VALUE;
    }
    uint8_t awbMode = (e.count > 0) ? e.data.u8[0] : (uint8_t)ANDROID_CONTROL_AWB_MODE_AUTO;

    // TODO: Add white balance simulation

    switch (awbMode) {
        case ANDROID_CONTROL_AWB_MODE_OFF:
            mAwbState = ANDROID_CONTROL_AWB_STATE_INACTIVE;
            return OK;
        case ANDROID_CONTROL_AWB_MODE_AUTO:
        case ANDROID_CONTROL_AWB_MODE_INCANDESCENT:
        case ANDROID_CONTROL_AWB_MODE_FLUORESCENT:
        case ANDROID_CONTROL_AWB_MODE_DAYLIGHT:
        case ANDROID_CONTROL_AWB_MODE_SHADE:
            // OK
            break;
        default:
            ALOGE("%s: Emulator doesn't support AWB mode %d",
                  __FUNCTION__, awbMode);
            return BAD_VALUE;
    }

    return OK;
}

void CameraHardware3Base::update3A(CameraMetadata& settings)
{
    if (mAeMode != ANDROID_CONTROL_AE_MODE_OFF) {
        settings.update(ANDROID_SENSOR_EXPOSURE_TIME,
                        &mAeCurrentExposureTime, 1);
        settings.update(ANDROID_SENSOR_SENSITIVITY,
                        &mAeCurrentSensitivity, 1);
    }

    settings.update(ANDROID_CONTROL_AE_STATE,
                    &mAeState, 1);
    settings.update(ANDROID_CONTROL_AF_STATE,
                    &mAfState, 1);
    settings.update(ANDROID_CONTROL_AWB_STATE,
                    &mAwbState, 1);

    uint8_t lensState;
    switch (mAfState) {
        case ANDROID_CONTROL_AF_STATE_PASSIVE_SCAN:
        case ANDROID_CONTROL_AF_STATE_ACTIVE_SCAN:
            lensState = ANDROID_LENS_STATE_MOVING;
            break;
        case ANDROID_CONTROL_AF_STATE_INACTIVE:
        case ANDROID_CONTROL_AF_STATE_PASSIVE_FOCUSED:
        case ANDROID_CONTROL_AF_STATE_FOCUSED_LOCKED:
        case ANDROID_CONTROL_AF_STATE_NOT_FOCUSED_LOCKED:
        case ANDROID_CONTROL_AF_STATE_PASSIVE_UNFOCUSED:
        default:
            lensState = ANDROID_LENS_STATE_STATIONARY;
            break;
    }
    settings.update(ANDROID_LENS_STATE, &lensState, 1);

#if 0
    // before update crop region, read it
    {
        camera_metadata_entry_t e;
        e = settings.find(ANDROID_SCALER_CROP_REGION);
        if(e.count >= 4) {
            ALOGD("DBG0 org crop region %d %d %d %d", e.data.i32[0], e.data.i32[1], e.data.i32[2], e.data.i32[3]);
        }
    }
#endif

#if 0
    // CTS: should also consider zoom factors...
    // for testing, update crop region
    {
        cam_capability_t_rtk* gCamCapability = &mCamCapability;
        int32_t scaler_crop_region[4];
        scaler_crop_region[0] = 0; //gCamCapability->active_array_size.left;
        scaler_crop_region[1] = 0; //gCamCapability->active_array_size.top;
        scaler_crop_region[2] = mMaxWidth; //gCamCapability->active_array_size.width + gCamCapability->active_array_size.left;
        scaler_crop_region[3] = mMaxHeight; //gCamCapability->active_array_size.height + gCamCapability->active_array_size.top;
        settings.update(ANDROID_SCALER_CROP_REGION, scaler_crop_region, 4);
    }
#endif
}

int CameraHardware3Base::convertV4L2FmtToHalFmt(int v4l2Fmt)
{
    LOG_FUNCTION_NAME();
#ifdef HAL_HWC_BUFFER
#if (NEW_HAL_PIXEL_FORMAT)
    ALOGV("%s[%d]: [new] V4L2 format : %c%c%c%c", __FUNCTION__, __LINE__, FCC2ASCII(v4l2Fmt));
    switch (v4l2Fmt) {
        case V4L2_PIX_FMT_NV12:
        case V4L2_PIX_FMT_YUYV:
            return HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP ; //HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP;
        case V4L2_PIX_FMT_NV21:
            return HAL_PIXEL_FORMAT_RTK_W16_H16_YCrCb_420_SP ; //HAL_PIXEL_FORMAT_RTK_16_YCrCb_420_SP;
        case V4L2_PIX_FMT_MJPEG:
        case V4L2_PIX_FMT_H264:
            return MJPG_HAL_FMT;
        default:
            return -1;
    }
#else
    LOGVV("[%s][legacy] V4L2 format : %c%c%c%c", __FUNCTION__, FCC2ASCII(v4l2Fmt));
    switch(v4l2Fmt) {
        case V4L2_PIX_FMT_NV12:
        case V4L2_PIX_FMT_YUYV:
        case V4L2_PIX_FMT_MJPEG:
        case V4L2_PIX_FMT_H264:
            return HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP;
#if 0
            // H264 is TODO
            //case V4L2_PIX_FMT_H264:
            //    return HAL_PIXEL_FORMAT_RGB_565;
#endif
        default:
            return -1;
    }
#endif
#else
    ALOGV("%s[%d]: [new] V4L2 format : %c%c%c%c", __FUNCTION__, __LINE__, FCC2ASCII(v4l2Fmt));
    switch (v4l2Fmt) {
        case V4L2_PIX_FMT_NV12:
        case V4L2_PIX_FMT_YUYV:
        case V4L2_PIX_FMT_NV21:
        case V4L2_PIX_FMT_MJPEG:
        case V4L2_PIX_FMT_H264:
            return HAL_PIXEL_FORMAT; //XCheck
        default:
            return -1;
    }

#endif
}

void CameraHardware3Base::dumpData(
    void* vaddr,
    int size,
    char* path,
    uint32_t* pCnt)
{
    FILE* fd = fopen(path, "w");
    if(fd != NULL) {
        fwrite(vaddr, 1, size, fd);
        int rst = fclose(fd);
        ALOGV("[%s] path %s size %d rst %d", __FUNCTION__, path, size, rst);
        (*pCnt)++;
    }
}

bool CameraHardware3Base::addExif(JPEG_RST_INFO* pInfo,
                                  uint8_t* metaBuf,
                                  /* , uint32_t bufSize */
                                  MISC_FIELD_INFO* pMiscInfo)
{
    return add_exif(pInfo, metaBuf, pMiscInfo);
}

void CameraHardware3Base::collectRequestMiscInfo(CameraMetadata& settings, MISC_FIELD_INFO* pInfo)
{
    LOG_FUNCTION_NAME();
    camera_metadata_entry_t e;
#if (SUPPORT_EXIF)
    // jpeg Thumbnail
    e = settings.find(ANDROID_JPEG_THUMBNAIL_SIZE);
    if(e.count >= 2) {
        pInfo->thumbWidth = e.data.i32[0];
        pInfo->thumbHeight = e.data.i32[1];
    }
    ALOGV("%s[%d]: Thumbnail info in request [%dx%d]", __FUNCTION__, __LINE__, pInfo->thumbWidth, pInfo->thumbHeight);

    // gps coordinates
    e = settings.find(ANDROID_JPEG_GPS_COORDINATES);
    if(e.count >= 3) {
        pInfo->gpsCoordinates[0] = e.data.d[0];
        pInfo->gpsCoordinates[1] = e.data.d[1];
        pInfo->gpsCoordinates[2] = e.data.d[2];

        ALOGV("%s[%d]: GPS coordinates %.2f %.2f %.2f", __FUNCTION__, __LINE__,
              (float)pInfo->gpsCoordinates[0], (float)pInfo->gpsCoordinates[1], (float)pInfo->gpsCoordinates[2]);
    }

    // gps timestamp
    e = settings.find(ANDROID_JPEG_GPS_TIMESTAMP);
    if(e.count >= 1) {
        pInfo->gpsTimestamp = e.data.i64[0];
        ALOGV("%s[%d]: GPS timestamp %lld", __FUNCTION__, __LINE__, pInfo->gpsTimestamp);
    }

    // gps method
    e = settings.find(ANDROID_JPEG_GPS_PROCESSING_METHOD);
    if(e.count > 0) {
        ALOGV("%s[%d]: GPS METHOD : \"%s\"", __FUNCTION__, __LINE__, e.data.u8);
        memcpy(pInfo->gpsProcessingMethod, e.data.u8, GPS_METHOD_STRLEN);
    }
#endif
}

#ifdef HAL3_HWC_BUFFER
int CameraHardware3Base::getStreamBufferSize(StreamBuffer* stream_buffer)
{
    buffer_handle_t* buffer = stream_buffer->buffer;
    int rst = pGrallocWrapper->GetSize(*buffer);
    return rst;
}

int CameraHardware3Base::getStreamBufferSize(buffer_handle_t* buffer)
{
    int rst = pGrallocWrapper->GetSize(*buffer);
    return rst;
}

int CameraHardware3Base::getStreamBufferSize(camera3_stream_buffer_t* stream_buffer)
{
    LOG_FUNCTION_NAME();
    buffer_handle_t* buffer = stream_buffer->buffer;
    int rst = pGrallocWrapper->GetSize(*buffer);
    return rst;
}

void CameraHardware3Base::getStreamBufferRect(buffer_handle_t* buffer, int* w, int* h)
{
    *w = pGrallocWrapper->GetWidth(*buffer);
    *h = pGrallocWrapper->GetHeight(*buffer);
}

void CameraHardware3Base::getStreamBufferVAddr(buffer_handle_t* buffer, void** vPtr, int format, void** cb, void** cr)
{
    int w, h, usage;
    getStreamBufferRect(buffer, &w, &h);
    usage = pGrallocWrapper->GetUsage(*buffer);
    getBufferVAddr(buffer, *w, *h, usage, vPtr, format, cb, cr);
}
#else
int CameraHardware3Base::getStreamBufferSize(buffer_handle_t __unused* buffer, int width, int height, int format)
{
    LOG_FUNCTION_NAME();
    int size;
    if (format != HAL_PIXEL_FORMAT_BLOB) {
        size = width * height * 3 / 2;
    } else {
        size = width * height * 2; // XCheck!!!
    }
    ALOGV("%s[%d]: [%dx%d], fmt:0x%x, size:%d  ", __FUNCTION__, __LINE__, width, height, format, size);
    return size;
}

void CameraHardware3Base::getStreamBufferVAddr(buffer_handle_t* buffer, void** vPtr, int w, int h, int usage, int format, void** cb, void** cr)
{
    LOG_FUNCTION_NAME();
    getBufferVAddr(buffer, w, h, usage, vPtr, format, cb, cr);
}
#endif

void CameraHardware3Base::getStreamBufferPhyAddr(buffer_handle_t* buffer, unsigned int* phyAddr)
{
    LOG_FUNCTION_NAME();
    unsigned int phy_info[2];
    phy_info[0] = 0;
    phy_info[1] = 0;
    CHECK(buffer != NULL);
#ifdef HAL3_HWC_BUFFER
    GrallocWrapper_GetPhyAddrs(pGrallocWrapper, *buffer, (unsigned int*)phy_info);
#endif
    *phyAddr = phy_info[0];
}

void CameraHardware3Base::getBufferVAddr(buffer_handle_t* pHandle, int w, int h, int usage, void** vPtr, int format, void** cb, void** cr)
{
    LOG_FUNCTION_NAME();
    const Rect rect(w, h);
    if(IS_YUV_BUFFER(format)) {
        android_ycbcr ycbcr = android_ycbcr();
        ycbcr.y = (void*)0xffffffff;
        memset(&ycbcr, 0x0, sizeof(android_ycbcr));

        status_t rst = GraphicBufferMapper::get().lockYCbCr(*pHandle, usage, rect, &ycbcr);
        if (format == HAL_PIXEL_FORMAT_YCbCr_420_888) {
            ALOGV("%s[%d]: lockYCbCr [fmt:%s] rst:0x%x, y:%p, cb:%p, cr:%p, cr-cb:%d, usage:0x%x", __FUNCTION__, __LINE__, getStreamForamtString(format), rst, ycbcr.y, ycbcr.cb, ycbcr.cr, (uint32_t) ycbcr.cr - (uint32_t) ycbcr.cb, usage);
        } else {
            ALOGV("%s[%d]: lockYCbCr [fmt:%s] rst:0x%x, usage:0x%x", __FUNCTION__, __LINE__, getStreamForamtString(format), rst, usage);
        }

        *vPtr = ycbcr.y;
        *cb = ycbcr.cb;
        *cr = ycbcr.cr;

        if(*vPtr == NULL) {
            ALOGE("%s[%d]: lock NULL with lockYCbCr, try with lock again", __FUNCTION__, __LINE__);
            GraphicBufferMapper::get().lock(*pHandle, usage, rect, vPtr);
            ALOGV("%s[%d]: retry with lock: %p", __FUNCTION__, __LINE__, *vPtr);
        }
    } else {
        *cb = NULL;
        *cr = NULL;
        ALOGV("%s[%d]: [%dx%d], fmt: 0x%x", __FUNCTION__, __LINE__, w, h, format);
        GraphicBufferMapper::get().lock(*pHandle, usage, rect, vPtr);
        ALOGV("%s[%d]: fmt %s rect: [%d][%d] handle %p -> vAddr %p", __FUNCTION__, __LINE__, getStreamForamtString(format), w, h, pHandle, *vPtr);
    }
    LOG_FUNCTION_END();
    GraphicBufferMapper::get().unlock(*pHandle);
}

status_t CameraHardware3Base::connectCameraBase()
{
    LOG_FUNCTION_NAME();
    status_t res;
    if(mStatus != STATUS_ERROR) {
        ALOGE("%s[%d]: Can't connect in state %d", __FUNCTION__, __LINE__, mStatus);
        return INVALID_OPERATION;
    }

    mStatus = STATUS_ERROR;
    mFlushed = false;

    // Initialize fake 3A
    mControlMode  = ANDROID_CONTROL_MODE_AUTO;
    mFacePriority = false;
    mAeMode       = ANDROID_CONTROL_AE_MODE_ON;
    mAfMode       = ANDROID_CONTROL_AF_MODE_AUTO;
    mAwbMode      = ANDROID_CONTROL_AWB_MODE_AUTO;
    mAeState      = ANDROID_CONTROL_AE_STATE_INACTIVE;
    mAfState      = ANDROID_CONTROL_AF_STATE_INACTIVE;
    mAwbState     = ANDROID_CONTROL_AWB_STATE_INACTIVE;
    mAeCounter    = 0;
    mAeTargetExposureTime = kNormalExposureTime;
    mAeCurrentExposureTime = kNormalExposureTime;
    mAeCurrentSensitivity  = kNormalSensitivity;

    mCamOpened = true;
    mStatus = STATUS_OPEN;
#ifdef HAL3_HWC_BUFFER
    char bufNumStr[10];
    memset(bufNumStr, 0x0, 10);
    int bufNum = kMaxBufferCount + 4;
    sprintf(bufNumStr, "%d", bufNum);
    property_set("rtk.hwc.v1.flip.buffer", bufNumStr);
#endif
    return NO_ERROR;
}

status_t CameraHardware3Base::closeCameraBase()
{
    LOG_FUNCTION_NAME();
    clearStreams();
    mStatus = STATUS_CLOSED;
    mCamOpened = false;
    return NO_ERROR;
}

status_t CameraHardware3Base::configureStreamsBase(camera3_stream_configuration_t* streamList)
{
    LOG_FUNCTION_NAME();
    stream_width = 0;
    stream_height = 0;

    jpg_width = 0;
    jpg_height = 0;

    mRuntimeFps0 = 0;
    mRuntimeFps1 = 0;

    int stateLoop = 100;
    ALOGV("%s[%d]: begin mStatus:%d", __FUNCTION__, __LINE__, mStatus);
    // wait for a while
    while ((mStatus != STATUS_OPEN && mStatus != STATUS_READY) && (stateLoop > 0)) {
        stateLoop--; // 100ms
        usleep(1000);
    }

    if (mStatus != STATUS_OPEN && mStatus != STATUS_READY) {
        ALOGE("%s[%d]: Cannot configure streams in state %d", __FUNCTION__, __LINE__, mStatus);
        return NO_INIT;
    }

    /**
     * Sanity-check input list.
     */
    if (streamList == NULL) {
        ALOGE("%s[%d]: NULL stream configuration", __FUNCTION__, __LINE__);
        return BAD_VALUE;
    }

    if (streamList->streams == NULL) {
        ALOGE("%s[%d]: NULL stream list", __FUNCTION__, __LINE__);
        return BAD_VALUE;
    }

    if (streamList->num_streams < 1) {
        ALOGE("%s[%d]: Bad number of streams requested: %d", __FUNCTION__, __LINE__, streamList->num_streams);
        return BAD_VALUE;
    }

    bool hasInputStream = false;

    camera3_stream_t* inputStream = NULL;
    for (size_t i = 0; i < streamList->num_streams; i++) {
        camera3_stream_t* newStream = streamList->streams[i];

        if (newStream == NULL) {
            ALOGE("%s[%d]: Stream index %zu was NULL", __FUNCTION__, __LINE__, i);
            return BAD_VALUE;
        }

        ALOGV("%s[%d]: [old] (id %zu), T:%d, [%dx%d], FMT:0x%x(%s), U:0x%x", __FUNCTION__, __LINE__,
              i, newStream->stream_type, newStream->width, newStream->height, newStream->format,
              getStreamForamtString(newStream->format), newStream->usage);

        if (newStream->stream_type == CAMERA3_STREAM_INPUT ||
            newStream->stream_type == CAMERA3_STREAM_BIDIRECTIONAL) {
            if (inputStream != NULL) {
                ALOGE("%s: Multiple input streams requested!", __FUNCTION__);
                return BAD_VALUE;
            }
            inputStream = newStream;
            hasInputStream = true;
        }

        bool validFormat = false;
        for (size_t f = 0; f < sizeof(kAvailableFormats) / sizeof(kAvailableFormats[0]); f++) {
            if (newStream->format == kAvailableFormats[f]) {
                validFormat = true;
                break;
            }
        }
        if (!validFormat) {
            ALOGE("%s[%d]: Unsupported stream format 0x%x requested", __FUNCTION__, __LINE__, newStream->format);
            return BAD_VALUE;
        }

        ALOGV("%s[%d]: (id %zu), T:%d, [%dx%d], FMT:0x%x(%s), U:0x%x", __FUNCTION__, __LINE__,
              i, newStream->stream_type, newStream->width, newStream->height, newStream->format,
              getStreamForamtString(newStream->format), newStream->usage);

#if 0
        // only modify preview & record stream format
        if(newStream->format == ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED ||
           newStream->format == ANDROID_SCALER_AVAILABLE_FORMATS_YCbCr_420_888) {
            // RTK hack, chenge the format in Hal
            newStream->format = convertV4L2FmtToHalFmt(mFormat); /*HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP*/
        }
#else
        // only modify preview & record stream format, do not modify input stream
        if(newStream->stream_type != CAMERA3_STREAM_INPUT &&
           newStream->format == ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED ) {
            //!hasInputStream ){ // not a reprocessing request
            if(hasInputStream) {
#if 1
                ALOGW("[%s] overwrite ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED with reprocessing request",
                      __FUNCTION__);
#endif
            }
            newStream->format = convertV4L2FmtToHalFmt(mFormat);
        }
#endif

    }
    // only one inputStream
    mInputStream = inputStream;
    ALOGV("%s[%d]: InputStream %p", __FUNCTION__, __LINE__, mInputStream);
    if(mInputStream != NULL) {
        ALOGV("%s[%d]: input stream info format %.2x", __FUNCTION__, __LINE__, mInputStream->format);
    }

    /**
     * Initially mark all existing streams as not alive
     */
    for (StreamIterator s = mStreams.begin(); s != mStreams.end(); ++s) {
        PrivateStreamInfo* privStream = static_cast<PrivateStreamInfo*>((*s)->priv);
        privStream->alive = false;
    }

    /**
     * Find new streams and mark still-alive ones
     */
    for (size_t i = 0; i < streamList->num_streams; i++) {
        camera3_stream_t* newStream = streamList->streams[i];

        if(newStream->rotation != CAMERA3_STREAM_ROTATION_0) {
            ALOGV("%s[%d]: rotation %.8x check failed", __FUNCTION__, __LINE__, newStream->rotation);
            return -EINVAL;
        }

        if (newStream->priv == NULL) {
            ALOGV("%s[%d]: Create PrivateStreamInfo", __FUNCTION__, __LINE__);
            // New stream, construct info
            PrivateStreamInfo* privStream = new PrivateStreamInfo();
            privStream->alive = true;
            newStream->max_buffers = kMaxBufferCount;

            newStream->priv = privStream;
            // only push the new one
            mStreams.push_back(newStream);
        } else {
            // Existing stream, mark as still alive.
            PrivateStreamInfo* privStream = static_cast<PrivateStreamInfo*>(newStream->priv);
            privStream->alive = true;
        }

        if(newStream->stream_type != CAMERA3_STREAM_INPUT) {
            // Always update usage and max buffers
            newStream->max_buffers = kMaxBufferCount;
        }

        switch (newStream->stream_type) {
            case CAMERA3_STREAM_OUTPUT:
                ALOGV("[%s][%d] detect CAMERA3_STREAM_OUTPUT, format:0x%x. only modify implementation defined stream usage:0x%x",
                      __FUNCTION__, __LINE__, newStream->format, newStream->usage);
                if(newStream->format != ANDROID_SCALER_AVAILABLE_FORMATS_BLOB) {
                    newStream->usage |= mGrallocUsage;
                }
                break;
            case CAMERA3_STREAM_INPUT:
                ALOGV("[%s][%d] handle reprocessing request", __FUNCTION__, __LINE__);
                newStream->usage |= mGrallocUsage;
                break;
            case CAMERA3_STREAM_BIDIRECTIONAL:
                ALOGE("[%s][%d] Wtf, CAMERA3_STREAM_BIDIRECTIONAL", __FUNCTION__, __LINE__);
                newStream->usage |= mGrallocUsage;
                break;
        }

        ALOGV("%s[%d]: [new] (id %zu), T:%d, [%dx%d], FMT:0x%x(%s), U:0x%x, bufNum:%i", __FUNCTION__, __LINE__,
              i, newStream->stream_type, newStream->width, newStream->height, newStream->format,
              getStreamForamtString(newStream->format), newStream->usage, newStream->max_buffers);

        // only use output stream size to configure v4l2
        if ((newStream->format != ANDROID_SCALER_AVAILABLE_FORMATS_BLOB) &&
            (newStream->stream_type != CAMERA3_STREAM_INPUT) &&
            isValidSize(newStream->width, newStream->height)) {
            if((stream_width * stream_height) < (int)(newStream->width * newStream->height)) {
                stream_width = newStream->width;
                stream_height = newStream->height;
                ALOGV("%s[%d]: save preview stream size info: [%dx%d]", __FUNCTION__, __LINE__, stream_width, stream_height);
            }
        }

        // save jpg stream size
        if (newStream->format == ANDROID_SCALER_AVAILABLE_FORMATS_BLOB &&
            newStream->stream_type != CAMERA3_STREAM_INPUT &&
            isValidSize(newStream->width, newStream->height)) {
            if ((jpg_width * jpg_height) < (int)(newStream->width * newStream->height)) {
                jpg_width = newStream->width;
                jpg_height = newStream->height;
                ALOGV("%s[%d]: save jpeg stream size info: [%dx%d]", __FUNCTION__, __LINE__, stream_width, stream_height);
            }
        }
    }

    /**
     * Reap the dead streams
     */
    for (StreamIterator s = mStreams.begin(); s != mStreams.end();) {
        PrivateStreamInfo* privStream = static_cast<PrivateStreamInfo*>((*s)->priv);
        if (!privStream->alive) {
            (*s)->priv = NULL;
            ALOGV("%s[%d]: [DEBUG] reap dead streams", __FUNCTION__, __LINE__);
            delete privStream;
            s = mStreams.erase(s);
        } else {
            ++s;
        }
    }

    /**
     * Can't reuse settings across configure call
     */
    ALOGV("%s[%d]: clear mPrevSettings", __FUNCTION__, __LINE__);
    mPrevSettings.clear();

    CHECK(mFormat != -1);
    CHECK(mPerferedFps != -1);

    if (stream_width * stream_height <= 0) {
        ALOGV("%s[%d]: No YUV stream detected, use BLOB stream rectangle info [%dx%d]", __FUNCTION__, __LINE__, jpg_width, jpg_height);
        stream_width = jpg_width;
        stream_height = jpg_height;
    }

    return NO_ERROR;
}

status_t CameraHardware3Base::processCaptureRequest(camera3_capture_request_t* request)
{
    LOG_FUNCTION_NAME();
    status_t res;
    uint32_t frameNumber = request->frame_number;
    mLatestFrameNumber = frameNumber;

    /** Validation */
    if (mStatus < STATUS_READY) {
        ALOGE("%s[%d]: Can't submit capture requests in state %d", __FUNCTION__, __LINE__, mStatus);
        return INVALID_OPERATION;
    }

    if (request == NULL) {
        ALOGE("%s[%d]: NULL request!", __FUNCTION__, __LINE__);
        return BAD_VALUE;
    }

    if (request->settings == NULL && mPrevSettings.isEmpty()) {
        ALOGE("%s[%d]: Request %d: NULL settings for first request after configureStreams()", __FUNCTION__, __LINE__, frameNumber);
        return BAD_VALUE;
    }

    if (request->input_buffer != NULL && request->input_buffer->stream != mInputStream) {
        ALOGE("%s[%d]: Request %d: Input buffer not from input stream!", __FUNCTION__, __LINE__, frameNumber);
        ALOGV("%s[%d]: Bad stream %p, expected: %p", __FUNCTION__, __LINE__, request->input_buffer->stream, mInputStream);
        ALOGV("%s[%d]: Bad stream type %d, expected stream type %d", __FUNCTION__, __LINE__,
              request->input_buffer->stream->stream_type, mInputStream ? mInputStream->stream_type : -1);

        return BAD_VALUE;
    }

    if (request->num_output_buffers < 1 || request->output_buffers == NULL) {
        ALOGE("%s[%d]: Request %d: No output buffers provided!", __FUNCTION__, __LINE__, frameNumber);
        return BAD_VALUE;
    }

    // Validate all buffers, starting with input buffer if it's given
    ssize_t idx;
    const camera3_stream_buffer_t* b;
    // set initial value of idx
    // if streams contain input buffer, then
    if (request->input_buffer != NULL) {
        ALOGV("%s[%d]: processCaptureRequest has input_buffer", __FUNCTION__, __LINE__);
        idx = -1;
        b = request->input_buffer;
    } else {
        idx = 0;
        b = request->output_buffers;
    }
    // check input/output buffers info
    do {
        PrivateStreamInfo* priv =
            static_cast<PrivateStreamInfo*>(b->stream->priv);

        // priv is set in configureStreams callback
        if (priv == NULL) {
            ALOGE("%s[%d]: Request %d: Buffer %zu: Unconfigured stream!", __FUNCTION__, __LINE__, frameNumber, idx);
            return BAD_VALUE;
        }
        // priv is set to alive in configureStreams callback
        if (!priv->alive) {
            ALOGE("%s[%d]: Request %d: Buffer %zu: Dead stream!", __FUNCTION__, __LINE__, frameNumber, idx);
            return BAD_VALUE;
        }
        // check buffer status
        if (b->status != CAMERA3_BUFFER_STATUS_OK) {
            ALOGE("%s[%d]: Request %d: Buffer %zu: Status not OK!", __FUNCTION__, __LINE__, frameNumber, idx);
            return BAD_VALUE;
        }
        // check fence status
        if (b->release_fence != -1) {
            ALOGE("%s[%d]: Request %d: Buffer %zu: Has a release fence!", __FUNCTION__, __LINE__, frameNumber, idx);
            return BAD_VALUE;
        }
        // check internal buffer
        if (b->buffer == NULL) {
            ALOGE("%s[%d]: Request %d: Buffer %zu: NULL buffer handle!", __FUNCTION__, __LINE__, frameNumber, idx);
            return BAD_VALUE;
        }

        {
            camera3_stream_t* stream = b->stream;
            //ALOGD("stream_type: %d",stream->stream_type);
        }
        idx++;
        b = &(request->output_buffers[idx]);
    } while (idx < (ssize_t)request->num_output_buffers);

    /**
     * Start processing this request
     */
    mStatus = STATUS_ACTIVE;

    CameraMetadata settings;

    // if settings in request is null, then, use previous settings
    if (request->settings == NULL) {
        settings.acquire(mPrevSettings);
    } else {
        settings = request->settings;
    }

    // use this settings to process3A
    res = process3A(settings);
    if (res != OK) {
        ALOGE("%s[%d]: process3A failed", __FUNCTION__, __LINE__);
        return res;
    }

    /**
     * Get ready for sensor config
     */
    nsecs_t  exposureTime;
    nsecs_t  frameDuration;
    uint32_t sensitivity;
    camera_metadata_entry_t entry;

    entry = settings.find(ANDROID_SENSOR_EXPOSURE_TIME);
    exposureTime = (entry.count > 0) ? entry.data.i64[0] : 1000L; //Sensor::kExposureTimeRange[0];
    entry = settings.find(ANDROID_SENSOR_FRAME_DURATION);
    frameDuration = (entry.count > 0) ? entry.data.i64[0] : 33331760L; //Sensor::kFrameDurationRange[0];
    entry = settings.find(ANDROID_SENSOR_SENSITIVITY);
    sensitivity = (entry.count > 0) ? entry.data.i32[0] : 100; //Sensor::kSensitivityRange[0];

    if (exposureTime > frameDuration) {
        frameDuration = exposureTime + 10000L ; //Sensor::kMinVerticalBlank;
        settings.update(ANDROID_SENSOR_FRAME_DURATION, &frameDuration, 1);
    }

    // Buffers: Vector<StreamBuffer>
    // create two Vectors that carry request->num_output_buffers
    Buffers* sensorBuffers = new Buffers(); // Internal buffers for passing across threads.
    sensorBuffers->setCapacity(request->num_output_buffers + 10);

    HalBufferVector* buffers = new HalBufferVector();
    buffers->setCapacity(request->num_output_buffers + 10);

    int num_jpeg_output_buffers = 0;

    // Process all the buffers we got for output, constructing internal buffer
    // structures for them, and lock them for writing.
    ALOGV("%s[%d]: request->num_output_buffers %d", __FUNCTION__, __LINE__, request->num_output_buffers);
    for (size_t i = 0; i < request->num_output_buffers; i++) {
        camera3_stream_buffer srcBuf = request->output_buffers[i]; // This buffer needs filled with image data from capture
        CHECK(equipeStream(&srcBuf, request->output_buffers[i]) == OK);

        ALOGV("%s[%d]: srcBuf %p orgBuf %p", __FUNCTION__, __LINE__, &srcBuf, &(request->output_buffers[i]));

        // Wait on fence first
        sp<Fence> bufferAcquireFence = new Fence(srcBuf.acquire_fence);
        res = bufferAcquireFence->wait(kFenceTimeoutMs);
        if (res == TIMED_OUT) {
            ALOGE("%s[%d]: Request %d: Buffer %zu: Fence timed out after %d ms", __FUNCTION__, __LINE__, frameNumber, i, kFenceTimeoutMs);
        }

        void* vaddr = NULL;
        void* cb = NULL;
        void* cr = NULL;
        uint32_t phyAddr = 0;

        ALOGV("%s[%d]: srcBuf.stream: [%dx%d] fmt:0x%x usg:0x%x", __FUNCTION__, __LINE__, srcBuf.stream->width, srcBuf.stream->height, srcBuf.stream->format, srcBuf.stream->usage);
#ifdef HAL3_HWC_BUFFER
        getStreamBufferVAddr(srcBuf.buffer, &vaddr, srcBuf.stream->format, &cb, &cr);
        getStreamBufferPhyAddr(srcBuf.buffer, &phyAddr);
#else
        getStreamBufferVAddr(srcBuf.buffer, &vaddr, (srcBuf.stream->width), (srcBuf.stream->height), srcBuf.stream->usage, srcBuf.stream->format, &cb, &cr);
#endif

        // create destBuf (StreamBuffer)
        StreamBuffer destBuf;
        destBuf.streamId        = kGenericStreamId;
        destBuf.width           = srcBuf.stream->width;
        destBuf.height          = srcBuf.stream->height;
        destBuf.format          = srcBuf.stream->format; //privBuffer->format; // Use real private format
        destBuf.stride          = srcBuf.stream->width; // TODO: query from gralloc
        destBuf.dataSpace       = srcBuf.stream->data_space;
        destBuf.usage           = srcBuf.stream->usage;
        destBuf.buffer          = srcBuf.buffer;
#ifdef HAL3_HWC_BUFFER
        destBuf.buf_size        = getStreamBufferSize(srcBuf.buffer);
#else
        destBuf.buf_size        = getStreamBufferSize(srcBuf.buffer, srcBuf.stream->width, srcBuf.stream->height, srcBuf.stream->format);
#endif
        destBuf.y               = vaddr;
        destBuf.cb              = cb;
        destBuf.cr              = cr;
        destBuf.phyAddr         = phyAddr;
#ifdef HAL3_HWC_BUFFER
        destBuf.gralloc_usage   = pGrallocWrapper->GetUsage(*(srcBuf.buffer));
#else
        destBuf.gralloc_usage   = 0;
        ALOGV("%s[%d]: destBuf [%dx%d], fmt:0x%x, buf_size: %d", __FUNCTION__, __LINE__, destBuf.width, destBuf.height, destBuf.format, destBuf.buf_size);
#endif
#if 0
        if((destBuf.gralloc_usage & GRALLOC_USAGE_PRIVATE_3) == 0) {
            ALOGW("%s XXX no GRALLOC_USAGE_PRIVATE_3 in buffer %d usage %.8x", __FUNCTION__, i, destBuf.gralloc_usage);
        }
#endif

        // BLOB is a jpeg capture request
        if (destBuf.format == HAL_PIXEL_FORMAT_BLOB) {
            num_jpeg_output_buffers = num_jpeg_output_buffers + 1;
            ALOGV("%s[%d]: BLOB Buffer [%dx%d], size:%d", __FUNCTION__, __LINE__, srcBuf.stream->width, srcBuf.stream->height, destBuf.buf_size);
        }

        // add dstBuf to sensorBuffers
        sensorBuffers->push_back(destBuf);
        // add srcBuf to HalBufferVector
        buffers->push_back(srcBuf);
    }

    // queue to ReadoutThread.
    Request r = Request();
    r.frameNumber = request->frame_number;
    r.settings = settings;
    r.sensorBuffers = sensorBuffers;        // dst buffer
    r.buffers = buffers;                    // src buffer
    r.num_jpeg_output_buffers = num_jpeg_output_buffers;
    r.input_buffer = request->input_buffer; // input buffer
    r.num_output_buffers = request->num_output_buffers;
    r.flushed = false; //mFlushed;
    ALOGV("%s[%d]: request: frame_number: %d, sensorBuffers %p", __FUNCTION__, __LINE__, request->frame_number, r.sensorBuffers);

    if (r.input_buffer != NULL) {
        ALOGV("%s[%d]: input buffer [%d] acq: %d rel: %d", __FUNCTION__, __LINE__, r.frameNumber, r.input_buffer->acquire_fence, r.input_buffer->release_fence);
        // Wait input buffer Fence
        sp<Fence> bufferAcquireFence = new Fence(r.input_buffer->acquire_fence);
        status_t res = bufferAcquireFence->wait(kFenceTimeoutMs);
        if (res == TIMED_OUT) {
            ALOGE("%s[%d]: Wait InputBuffer acquire fence timeout", __FUNCTION__, __LINE__);
        }

        int sync_timeline_fd = -1;
#if 0
        // set release_fence of input_buffer
        int sync_timeline_fd = sw_sync_timeline_create();
        ALOGD("sync_timeline_fd %d", sync_timeline_fd);

        // no permission to access sw fence , so this block of code would not be executed.
        if(sync_timeline_fd >= 0) {
            CHECK(sync_timeline_fd >= 0);
            char str[256] = {'\0'};
            sprintf(str, "input-fence-%d", r.frameNumber);
            unsigned val = (unsigned) r.frameNumber;
            int fd = sw_sync_fence_create(sync_timeline_fd,
                                          str,
                                          val);
            CHECK(fd >= 0);
            r.input_buffer->release_fence = fd;
        }
#endif
        r.input_buffer_sync_fd = sync_timeline_fd;
        // then, clone input_buffer structure to prevent caller release structure memory.
        camera3_stream_buffer_t* input_buffer_2 = (camera3_stream_buffer_t*) malloc(sizeof(camera3_stream_buffer_t));
        memcpy(input_buffer_2, r.input_buffer, sizeof(camera3_stream_buffer_t));
        CHECK(equipeStream(input_buffer_2, *(r.input_buffer)) == OK);
        r.input_buffer = input_buffer_2;
    }

    if (num_jpeg_output_buffers > 0) {
        ALOGV("%s[%d]: Queued frame for JPEG request frameNumber %d sensorBuffers %p", __FUNCTION__, __LINE__, request->frame_number, r.sensorBuffers);
    } else {
        ALOGV("%s[%d]: Queued frame for PREV request frameNumber %d sensorBuffers %p", __FUNCTION__, __LINE__, request->frame_number, r.sensorBuffers);
    }

    // call virtual functions
    processCaptureRequest(r);
    // Cache the settings for next time
    mPrevSettings.acquire(settings);

    LOG_FUNCTION_END();
    return OK;
}

void CameraHardware3Base::clearStreams()
{
    LOG_FUNCTION_NAME();
    // Clear out private stream information
    for (StreamIterator s = mStreams.begin(); s != mStreams.end(); s++) {
        PrivateStreamInfo* privStream =
            static_cast<PrivateStreamInfo*>((*s)->priv);
        if(privStream != NULL) {
            delete privStream;
        }
        (*s)->priv = NULL;
    }
    mStreams.clear();
}

status_t CameraHardware3Base::equipeStream(camera3_stream_buffer* dst, const camera3_stream_buffer& src)
{
    LOG_FUNCTION_NAME();
    // Compare src stream with mStreams, then got strem form mStreams
    for (StreamIterator s = mStreams.begin(); s != mStreams.end(); s++) {
        camera3_stream_t* stream = (*s);
        CHECK(stream != NULL);
        if(memcmp(stream, src.stream, sizeof(camera3_stream_t)) == 0) {
            dst->stream = stream;
            return OK;
        }
    }

    return INVALID_OPERATION;
}

bool CameraHardware3Base::isBlockedRes(int w, int h)
{
    int array_length = sizeof(blocked_sizes) / sizeof(Res);

    //ALOGD("blocked_sizes length %d",array_length);

    for (int i = 0; i < array_length; i++) {
        Res r = blocked_sizes[i];
        if (r.width == w && r.height == h) {
            ALOGV("%s[%d]: Res %dx%d is blocked", __FUNCTION__, __LINE__, w, h);
            return true;
        }
    }

    return false;
}

}; // namespace android
