//#define LOG_NDEBUG 0
#define LOG_TAG "V4L2Device2MJPG"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <log/log.h>
#ifdef HAL3_HWC_BUFFER
#include <utils/backtrace.h>
#endif
#include <cutils/properties.h>
#include <gui/Surface.h>
#include <OMX_IVCommon.h>
#include <media/stagefright/foundation/AMessage.h>
#include <media/stagefright/foundation/ABuffer.h>
#include <media/stagefright/MediaErrors.h>
#include <media/ICrypto.h>
#include <ui/GraphicBufferMapper.h>
#ifdef HAL3_HWC_BUFFER
#include <libRTKAllocator/include/allocator/MallocAllocator.h>
#endif
#include "V4L2Device2MJPG.h"
#include "MediaCodecDecodeHelper.h"
#include "OMXComponentDecodeHelper.h"
#include "misc.h"
#include "Debug.h"

#define QUEUE_SIZE      (16)

#define SAVE_RAW_DATA   (0)

#define USE_OMX_COMPONENT   (1)

namespace androidcamera3 {

static const nsecs_t kWaitNs = (1*1E9);

V4L2Device2MJPG::V4L2Device2MJPG(
    int v4l2Id,
    V4L2Params *params,
    V4L2Device2Callback *callback,
    int maxBufferCount,
#ifdef HAL3_HWC_BUFFER
    GrallocWrapper *pGrallocWrapper,
#endif
    FormatConvertHelper *pConvertHelper,
    bool forceAllocator,
    bool usePhyAddr)
: V4L2Device2Common(v4l2Id,
        params,
        callback,
        maxBufferCount,
#ifdef HAL3_HWC_BUFFER
        pGrallocWrapper,
#endif
        pConvertHelper,
        forceAllocator,
        usePhyAddr),
  mMediaCodecInputQueue(NULL),
  mDecodeHelper(NULL),
  mDeviceRunning(true),
  mV4L2RawFd(NULL)
{
    ALOGD("V4L2Device2MJPG()");
    initAllocBufs(params->w,params->h);

    mMediaCodecInputQueue = new BufferQueue<AllocatorBuf*>(QUEUE_SIZE, (char*)"MediaCodecInputQueue");
#ifdef HAL3_HWC_BUFFER
	mDecodeHelper = dynamic_cast<OMXDecodeHelper *>
#else
    //mDecodeHelper = (OMXDecodeHelper *)
#endif
#if (USE_OMX_COMPONENT)
        (new OMXComponentDecodeHelper(
#else
        (new MediaCodecDecodeHelper(
#endif
            this,
            mParams.v4l2format,
            mParams.w,
            mParams.h));

#if (SAVE_RAW_DATA)
    openRawFd();
#endif

}

V4L2Device2MJPG::~V4L2Device2MJPG()
{
    if(mRequestQueue.size() > 0) {
        ALOGW("~V4L2Device2MJPG() mRequestQueue %d",mRequestQueue.size() );
    }

    // there should be no pending Request in queue, test to remove CHECK here
    // CHECK(mRequestQueue.size() == 0);

    stop();
    closeDev();
    releaseAllocBufs();

    // clear BufferQueue(s)
    if(mMediaCodecInputQueue != NULL){
        delete mMediaCodecInputQueue;
        mMediaCodecInputQueue = NULL;
    }

#if (SAVE_RAW_DATA)
    closeRawFd();
#endif

    if(mDecodeHelper != NULL) {
        delete mDecodeHelper;
        mDecodeHelper = NULL;
    }

    ALOGD("~V4L2Device2MJPG() %d",__LINE__);
}

void V4L2Device2MJPG::stop()
{
    ALOGD("%s : E",__FUNCTION__);
    stopDevice();
    // stop streaming
    stopStreaming();

    ALOGD("%s : [%d]",__FUNCTION__,__LINE__);

    // stop V4L2Thread
    if(mV4L2Thread.get() != NULL) {
        mV4L2Thread->stop();
        mV4L2Thread->requestExitAndWait();
        mV4L2Thread.clear();
        mV4L2Thread = NULL;
    }

    ALOGD("%s : [%d]",__FUNCTION__,__LINE__);
#if 0
    // dequeue all buffers in V4L2
    struct v4l2_buffer buf;
    while(ioctl(mFd, VIDIOC_DQBUF, &buf) == 0);
#endif

    mDecodeHelper->stop();

    ALOGD("%s : Exit",__FUNCTION__);
}

status_t V4L2Device2MJPG::start()
{
    mDecodeHelper->start();

    // run V4L2
    queueAllRawBufs();
    startStreaming();

    CHECK(mV4L2Thread.get() == NULL);
    mV4L2Thread = new V4L2Thread(this);
    mV4L2Thread->run("V4L2");

    return OK;
}

void V4L2Device2MJPG::flush(bool reportError)
{
    CHECK(reportError);

    stop();

    while(mRequestQueue.size()>0) {
        Request *r = new Request();
        dequeueRequest(r);
        r->pts = 0x2379;
        r->mSuccess = false;
        mCallback->returnRequest(r);
    }
}

void V4L2Device2MJPG::dumpDeviceStatus()
{
    // Do nothing now
}

int32_t V4L2Device2MJPG::getOccupiedRequestCount()
{
    int reqs = mRequestQueue.size();
    return reqs;
}

/**
 * get buffer from v4l2 queue, and copy to codec input buffer
 * then queue buffer back to v4l2
 */
status_t
V4L2Device2MJPG::onInputBufferAvailable(
    uint8_t *vAddr,
    int32_t __unused phyAddr,
    size_t capacity,
    nsecs_t *pts,
    size_t *range)
{
    LOG_FUNCTION_NAME();
    // wait input from V4L2 ready
    while(mMediaCodecInputQueue->size() <= 0 && mDeviceRunning){
        mMediaCodecInputQueue->wait();
    }

    if(!mDeviceRunning) {
        return INVALID_OPERATION;
    }

    CHECK(mMediaCodecInputQueue->size()>0);

    AllocatorBuf *buf = mMediaCodecInputQueue->peek();
    CHECK(buf != NULL);

    mMediaCodecInputQueue->dequeue();
#ifdef HAL3_HWC_BUFFER
    buf->pAllocator->Invalidate();
#endif
    ALOGV("%s [%d] - dst %p src %p",
        __FUNCTION__,
        __LINE__,
        vAddr,
        buf->vAddr);

    CHECK((int)capacity >= buf->mBytesUsed);

    // do copy
    memcpy(vAddr,buf->vAddr,buf->mBytesUsed);
    *pts = buf->mPts;
    *range = buf->mBytesUsed;

    // enqueue to V4L2Device
    enqueueRawBuffer(buf);
    return OK;
}

/**
 * get output result from decode helper, and report back to hal
 */
status_t
V4L2Device2MJPG::onOutputBufferAvailable(
    uint8_t *vAddr,
    int32_t phyAddr,
    size_t range,
    nsecs_t pts)
{
    LOG_FUNCTION_NAME();
    while(mRequestQueue.size() <= 0 && mDeviceRunning) {
        Mutex::Autolock l(mRequestQueueWaitLock);
        mRequestQueueSignal.waitRelative(mRequestQueueWaitLock,(kWaitNs*10));
        ALOGV("No Request in RequestQueue");
    }

    if(!mDeviceRunning){
        return OK;
    }

    CHECK(mRequestQueue.size() > 0);

    Request *r = new Request();
    dequeueRequest(r);
    r->mSuccess = true;

    // we do not trust device pts
    r->pts = pts;

    // Complete Request use FormatConvertHelper
    BUF_CONVERT_INFO ext;
    ext.w = mParams.w;
    ext.h = mParams.h;
    ext.format = mHalPixelFormat;
    ext.vAddr = vAddr;
    ext.phyAddr = phyAddr;
    ext.bufferSize = range;
    ext.pHandle = NULL;
#ifdef HAL3_HWC_BUFFER
    ext.allocator = NULL;
#endif
    CHECK(mFormatConvertHelper != NULL);

    if(phyAddr != 0) {
        mFormatConvertHelper->convertBuffers(r,&ext);
        mFormatConvertHelper->setRtkPrivData(r,mParams.fps);
    }else{
        //r->mSuccess = false;
    }

    // report request to HAL
    mCallback->returnRequest(r);
    return OK;
}

// -------- V4L2Thread --------
V4L2Device2MJPG::V4L2Thread::V4L2Thread(V4L2Device2MJPG *parent)
: mParent(parent), mRunning(true)
{
    ALOGD("V4L2Thread()");
}

V4L2Device2MJPG::V4L2Thread::~V4L2Thread()
{
    ALOGD("~V4L2Thread()");
}

bool V4L2Device2MJPG::V4L2Thread::threadLoop()
{
    LOG_FUNCTION_NAME();
    if(!mRunning) return false;

    int32_t bufId = -1;
    status_t rst = OK;
    rst = mParent->dequeueRawBuffer(&bufId);
    if(rst != OK) {
        ALOGE("V4L2Thread dequeueRawBuffer failed");
        mRunning = false;
        return false;
    }

    CHECK(bufId >= 0);
    CHECK(mParent->mAllocBufs[bufId].mEnqueued);

    mParent->mAllocBufs[bufId].mEnqueued = false;

    // check if it is a validate MJPG frame
    AllocatorBuf *buf = &(mParent->mAllocBufs[bufId]);
    if(mParent->mParams.v4l2format == V4L2_PIX_FMT_MJPEG && !mParent->checkMJPEGFrameValidate(buf)){
        ALOGE("Unable to pass MJPG frame checking bufID:%d",bufId);
        mParent->enqueueRawBuffer(buf);
        return true;
    }

    // push to MediaCodecInputQueue and notify
    mParent->mMediaCodecInputQueue->enqueue(&(mParent->mAllocBufs[bufId]));
    mParent->mMediaCodecInputQueue->signal();
    ALOGV("V4L2Thread loop mMediaCodecInputQueue size %d",
        mParent->mMediaCodecInputQueue->size());

#if 0
    if(mParent->mV4L2InputQueue->size() > 0) {
        AllocatorBuf *buf = mParent->mV4L2InputQueue->peek();
        mParent->mV4L2InputQueue->dequeue();
        rst = mParent->enqueueRawBuffer(buf);
        if(rst != OK) {
            ALOGE("V4L2Thread enqueueRawBuffer failed");
            mRunning = false;
            return false;
        }
    }
#endif

    return true;
}

void V4L2Device2MJPG::V4L2Thread::signal()
{
    Mutex::Autolock l(mLock);
    mSignal.signal();
}

void V4L2Device2MJPG::V4L2Thread::stop()
{
    mRunning = false;
    signal();
}

void V4L2Device2MJPG::initAllocBufs(int32_t w, int32_t h)
{
    LOG_FUNCTION_NAME();
    // content is jpeg, w*h is enough.
    int Bpp = V4L2FmtToBpp(mParams.v4l2format);
    int size = ROUNDUP16(w)*ROUNDUP16(h)*Bpp/8;

    for(int i=0;i<V4L2_BUF_NUM;i++)
    {
        mAllocBufs[i].mIndex = i;
#ifdef HAL3_HWC_BUFFER
        mAllocBufs[i].pAllocator = (AllocatorDef *) new MallocAllocator();
        mAllocBufs[i].pAllocator->Alloc(size);
        mAllocBufs[i].vAddr = mAllocBufs[i].pAllocator->GetVirAddr();
        mAllocBufs[i].mWidth = w;
        mAllocBufs[i].mHeight = h;
        mAllocBufs[i].mSize = size;
#else
        mAllocBufs[i].vAddr = (void*)malloc(sizeof(char) * size);
        mAllocBufs[i].mWidth = w;
        mAllocBufs[i].mHeight = h;
        mAllocBufs[i].mSize = size;
        ALOGD("%s [%d] ",__FUNCTION__,__LINE__);
#endif
    }
}

void V4L2Device2MJPG::releaseAllocBufs()
{
    for(int i=0;i<V4L2_BUF_NUM;i++)
    {
#ifdef HAL3_HWC_BUFFER
        if(NULL != mAllocBufs[i].pAllocator)
        {
            ALOGV("%s id:%d",__FUNCTION__,i);
            mAllocBufs[i].pAllocator->Free();
            delete mAllocBufs[i].pAllocator;
            mAllocBufs[i].pAllocator = NULL;
        }
#endif
        ALOGD("%s [%d] ",__FUNCTION__,__LINE__);
    }
}

status_t V4L2Device2MJPG::queueAllRawBufs()
{
    LOG_FUNCTION_NAME();
    int ret;
    // Queue all buffers to V4L2
    for(int i=0;i<V4L2_BUF_NUM;i++)
    {
        memset(&videoIn.buf,0x0,sizeof(videoIn.buf));

        videoIn.buf.index = i;
        CHECK(i==mAllocBufs[i].mIndex);
        videoIn.buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        videoIn.buf.memory = V4L2_MEMORY_USERPTR;
        videoIn.buf.m.userptr = (unsigned long)(mAllocBufs[i].vAddr);
        videoIn.buf.length = mAllocBufs[i].mSize;

        ALOGV("VIDIOC_QBUF [%d] len:%d addr:%lx",
            i,
            videoIn.buf.length,
            videoIn.buf.m.userptr);

        ret = ioctl(mFd, VIDIOC_QBUF, &videoIn.buf);
        if (ret < 0) {
            ALOGE("Init: VIDIOC_QBUF Failed: %s", strerror(errno));
            mAllocBufs[i].mEnqueued = false;
            return INVALID_OPERATION;
        }
        mAllocBufs[i].mEnqueued = true;
    }

    return OK;
}

status_t V4L2Device2MJPG::enqueueRawBuffer(AllocatorBuf *buf)
{
    LOG_FUNCTION_NAME();
    if(!isStreaming()){
        return INVALID_OPERATION;
    }

    CHECK(mFd != -1);
    CHECK(buf != NULL);
    CHECK(!buf->mEnqueued);

    int ret;
    struct v4l2_buffer v4l2Buf;
    struct v4l2_buffer *pBuf = &v4l2Buf;
    memset(pBuf, 0x0, sizeof(struct v4l2_buffer));

    pBuf->index = buf->mIndex;
    pBuf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    pBuf->memory = V4L2_MEMORY_USERPTR;
    pBuf->m.userptr = (unsigned long) buf->vAddr;
    pBuf->length = buf->mSize;
    ret = ioctl(mFd, VIDIOC_QBUF, pBuf);

    if (ret < 0) {
        ALOGE("[%s]: QBUF Failed: %s",
            __FUNCTION__,
            strerror(errno));
        return INVALID_OPERATION;
    }else{
        buf->mEnqueued = true;
        return OK;
    }
}

status_t V4L2Device2MJPG::dequeueRawBuffer(int32_t *id)
{
    LOG_FUNCTION_NAME();
    if(!isStreaming()){
        return INVALID_OPERATION;
    }

    struct v4l2_buffer buf;
    int ret;

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_USERPTR;

    ret = ioctl(mFd, VIDIOC_DQBUF, &buf);

    if(ret<0){
        closeDev();
        return INVALID_OPERATION;
    }

    if ((int)buf.bytesused <= 0) {
        ALOGW("[%s]: Invalid frame[%d] size:%d",
            __FUNCTION__,
            buf.index,
            (int)buf.bytesused);
        return INVALID_OPERATION;
    }

    uint32_t bytes = buf.bytesused;
    int index = buf.index;

    nsecs_t pts = systemTime();

#if 0
    nsecs_t pts = (nsecs_t)buf.timestamp.tv_sec * 1000000000 +
        (nsecs_t)buf.timestamp.tv_usec * 1000;
#endif

    mAllocBufs[index].mBytesUsed = bytes;
    // We do not trust pts from device
    mAllocBufs[index].mPts = pts;

#if (SAVE_RAW_DATA)
    writeRawFd(
        (char *) (mAllocBufs[index].vAddr),
        bytes);
#endif

    *id = index;
    return OK;
}

bool V4L2Device2MJPG::enqueueRequest(const Request& r)
{
    LOG_FUNCTION_NAME();
    {
        Mutex::Autolock l(mRequestQueueLock);
        mRequestQueue.push_back(r);
    }

    {
        Mutex::Autolock l(mRequestQueueWaitLock);
        mRequestQueueSignal.signal();
    }
    return true;
}

bool V4L2Device2MJPG::dequeueRequest(Request *r)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mRequestQueueLock);

    CHECK(mRequestQueue.size()>0);
    *r = mRequestQueue[0];
    mRequestQueue.removeItemsAt(0);

    return true;
}

void V4L2Device2MJPG::stopDevice()
{
    ALOGD("%s : E",__FUNCTION__);
    mDeviceRunning = false;

    {
        Mutex::Autolock l(mRequestQueueWaitLock);
        mRequestQueueSignal.signal();
    }

    mMediaCodecInputQueue->signal();
}

void V4L2Device2MJPG::dumpEnqueuedBufCount()
{
    int count = 0;
    for(int i=0;i<V4L2_BUF_NUM;i++) {
        if(mAllocBufs[i].mEnqueued){
            count++;
        }
    }
    ALOGD("%s %d",__FUNCTION__,count);
}

bool V4L2Device2MJPG::checkMJPEGFrameValidate(AllocatorBuf *buf)
{
    LOG_FUNCTION_NAME();
    uint8_t *data = (uint8_t *) (buf->vAddr);
    uint32_t size = buf->mBytesUsed;
    bool bIsValidFrame = false;
    // Invalidate first
#ifdef HAL3_HWC_BUFFER
    buf->pAllocator->Invalidate();
#endif
    if(size>=4 && data[0]==0xff && data[1]==0xd8){
        bIsValidFrame = true;
    }

    if(bIsValidFrame) {
        // sync latest MJPEG frame checking logic
        int checksize = 1024;
        if (checksize > (int)(size-4)) {
            checksize = (size-4);
        }

        unsigned char *ptr = (unsigned char *)data + size -1;

        bIsValidFrame = false;

        for(int i=0;i<checksize;i++){
            if(*ptr==0xd9 && *(ptr-1)==0xff){
                ALOGV("Find tailing token at %p round %d",ptr,i);
                bIsValidFrame = true;
                break;
            }
            // update ptr
            ptr = ptr - 1;
        }
    }
    return bIsValidFrame;
}

void V4L2Device2MJPG::openRawFd()
{
    LOG_FUNCTION_NAME();
    CHECK(mV4L2RawFd==NULL);
    char path[128]={'\0'};
    nsecs_t time = systemTime();
    int32_t t = time / 1E9;
    snprintf(path,128,"/data/v4l2-%d.raw",t);
    mV4L2RawFd = fopen(path,"w");
}

void V4L2Device2MJPG::closeRawFd()
{
    LOG_FUNCTION_NAME();
    if(mV4L2RawFd != NULL) {
        fclose(mV4L2RawFd);
        mV4L2RawFd = NULL;
    }
}

void V4L2Device2MJPG::writeRawFd(char *buf, size_t length)
{
    LOG_FUNCTION_NAME();
    if(mV4L2RawFd != NULL) {
        ALOGD("write buffer length %d",length);
        fwrite(buf,1,length,mV4L2RawFd);
    }else{
        ALOGE("Write mV4L2RawFd NG file is NULL");
    }
}

};
