//#define LOG_NDEBUG 0
#define LOG_TAG     "MediaCodecDecodeHelper"

#include <linux/videodev2.h>
#include <log/log.h>
#include <gui/Surface.h>
#include <OMX_IVCommon.h>
#include "MediaCodecDecodeHelper.h"

#define QUEUE_SIZE      (16)

using namespace android;

namespace androidcamera3 {

MediaCodecDecodeHelper::MediaCodecDecodeHelper(
    OMXDecodeHelperCallback *callback,
    int32_t format,
    int32_t w,
    int32_t h)
: mInputBufIdQueue(NULL),
  mOutputBufIdQueue(NULL),
  mLooper(NULL),
  mCodec(NULL),
  mHandler(NULL),
  mReply(NULL)
{
    ALOGD("MediaCodecDecodeHelper()");

    mCallback = callback;

    // callback should not be null
    CHECK(mCallback != NULL);

    mFormat = format;
    mWidth = w;
    mHeight = h;
    mDeviceRunning = true;

    mInputBufIdQueue = new BufferQueue<int>(QUEUE_SIZE,"inputId");
    mOutputBufIdQueue = new BufferQueue<int>(QUEUE_SIZE,"outputId");
}

MediaCodecDecodeHelper::~MediaCodecDecodeHelper()
{
    release();
    ALOGD("~MediaCodecDecodeHelper()");
}

status_t
MediaCodecDecodeHelper::init()
{
    ALOGD("%s - do nothing now",__FUNCTION__);
    return OK;
}

status_t
MediaCodecDecodeHelper::start()
{
    ALOGD("%s : E",__FUNCTION__);
    status_t rst = createMediaCodec();

    CHECK(rst == OK);

    CHECK(mInputBufThread.get() == NULL);
    mInputBufThread = new MediaCodecThread(this,false);
    mInputBufThread->run("MediaCodec-In");

    CHECK(mOutputBufThread.get() == NULL);
    mOutputBufThread = new MediaCodecThread(this,true);
    mOutputBufThread->run("MediaCodec-Out");

    ALOGD("%s : X",__FUNCTION__);

    return rst;
}

status_t
MediaCodecDecodeHelper::stop()
{
    mDeviceRunning = false;

    ALOGD("%s : E",__FUNCTION__);
    // stop MediaCodec threads

    // 1 input
    if(mInputBufThread.get() != NULL) {
        mInputBufThread->stop();
        mInputBufIdQueue->signal();
        mInputBufThread->requestExitAndWait();
        mInputBufThread.clear();
        mInputBufThread = NULL;
    }

    // 2 output
    if(mOutputBufThread.get() != NULL) {
        mOutputBufThread->stop();
        mOutputBufIdQueue->signal();
        mOutputBufThread->requestExitAndWait();
        mOutputBufThread.clear();
        mOutputBufThread = NULL;
    }

    // stop MediaCodec
    if(mCodec != NULL) {
        // stop MediaCodec
        mCodec->stop();

        // free pending output buffers
        while(mOutputBufIdQueue->size() > 0) {
            int index = mOutputBufIdQueue->peek();
            mOutputBufIdQueue->dequeue();
            ALOGD("stop release MediaCodec output buffer %d",index);
            mCodec->releaseOutputBuffer(index);
        }

        mCodec->release();
        mCodec.clear();
        mCodec = NULL;
        ALOGD("%s release MediaCodec done",
            __FUNCTION__);
    }

    if(mLooper != NULL) {
        mLooper->stop();
        mLooper.clear();
        mLooper = NULL;
        ALOGV("[%s][%d] release ALooper done",
            __FUNCTION__,
            __LINE__);
    }

    if(mHandler != NULL) {
        mHandler.clear();
        mHandler = NULL;
        ALOGV("[%s][%d] release AHandler done",
            __FUNCTION__,
            __LINE__);
    }

    if(mReply != NULL) {
        mReply.clear();
        mReply = NULL;
        ALOGV("[%s][%d] release mReply done",
            __FUNCTION__,
            __LINE__);
    }

    ALOGD("%s : X",__FUNCTION__);

    return OK;
}

status_t
MediaCodecDecodeHelper::release()
{
    ALOGD("%s : E",__FUNCTION__);

    if(mInputBufIdQueue != NULL){
        delete mInputBufIdQueue;
        mInputBufIdQueue = NULL;
    }

    if(mOutputBufIdQueue != NULL){
        delete mOutputBufIdQueue;
        mOutputBufIdQueue = NULL;
    }

    CHECK(mCodec == NULL);
    CHECK(mLooper == NULL);
    CHECK(mHandler == NULL);
    CHECK(mReply == NULL);

    CHECK(mInputBufThread == NULL);
    CHECK(mOutputBufThread == NULL);

    ALOGD("%s : X",__FUNCTION__);
    return OK;
}

// ------------ MediaCodecDecodeHelper Thread ------------------

MediaCodecDecodeHelper::MediaCodecThread::MediaCodecThread(
    MediaCodecDecodeHelper *parent,
    bool out)
: mParent(parent), mRunning(true), mOut(out)
{
    ALOGD("MediaCodecThread()");
}

MediaCodecDecodeHelper::MediaCodecThread::~MediaCodecThread()
{
    ALOGD("~MediaCodecThread() %s",(mOut?"Out":"In"));
}

bool MediaCodecDecodeHelper::MediaCodecThread::loopIn()
{
    while(mParent->mInputBufIdQueue->size()<=0 && mRunning){
        mParent->mInputBufIdQueue->wait();
    }

    if(!mRunning) return false;
    CHECK(mParent->mInputBufIdQueue->size() > 0);

    int index = mParent->mInputBufIdQueue->peek();
    mParent->mInputBufIdQueue->dequeue();

    mParent->codecInputAvailable(index);

    return true;
}

bool MediaCodecDecodeHelper::MediaCodecThread::loopOut()
{
    while(mParent->mOutputBufIdQueue->size()<=0 && mRunning){
        mParent->mOutputBufIdQueue->wait();
    }

    if(!mRunning) return false;
    CHECK(mParent->mOutputBufIdQueue->size() > 0);

    int index = mParent->mOutputBufIdQueue->peek();
    mParent->mOutputBufIdQueue->dequeue();
    mParent->mCodec->releaseOutputBuffer(index);
    return true;
}

bool MediaCodecDecodeHelper::MediaCodecThread::threadLoop()
{
    if(mOut) {
        return loopOut();
    }else{
        return loopIn();
    }
}

void MediaCodecDecodeHelper::MediaCodecThread::signal()
{
    Mutex::Autolock l(mLock);
    mSignal.signal();
}

void MediaCodecDecodeHelper::MediaCodecThread::stop()
{
    mRunning = false;
    signal();
}

// ------------ MediaCodecDecodeHelper private implementation --------------
MediaCodecDecodeHelper::MyDecodeHandler::MyDecodeHandler(MediaCodecDecodeHelper *parent)
{
    ALOGD("MyDecodeHandler()");
    mParent = parent;
}

MediaCodecDecodeHelper::MyDecodeHandler::~MyDecodeHandler()
{
    ALOGD("~MyDecodeHandler()");
}

void
MediaCodecDecodeHelper::MyDecodeHandler::onMessageReceived(const sp <AMessage> &msg)
{
    ALOGV("MediaCodec onMessage %s",
        msg->debugString().c_str());

    switch(msg->what()) {
        case kWhatCodecNotify:
        {
            int32_t cbID;
            msg->findInt32("callbackID", &cbID);
            ALOGV("MediaCodec kWhatCodecNotify: cbID=%d",cbID);

            switch(cbID) {
                case MediaCodec::CB_INPUT_AVAILABLE:
                {
                    int32_t index=0;
                    CHECK(msg->findInt32("index", &index));
                    ALOGV("CB_INPUT_AVAILABLE index: %d",index);
                    // push to queue
                    mParent->mInputBufIdQueue->enqueue(index);
                    mParent->mInputBufIdQueue->signal();
                    break;
                }
                case MediaCodec::CB_OUTPUT_AVAILABLE:
                {
                    int32_t index=0;
                    size_t offset=0;
                    size_t size=0;
                    int64_t timeUs=0;
                    int32_t flags=0;

                    CHECK(msg->findInt32("index", &index));
                    CHECK(msg->findSize("offset", &offset));
                    CHECK(msg->findSize("size", &size));
                    CHECK(msg->findInt64("timeUs", &timeUs));
                    CHECK(msg->findInt32("flags", &flags));
                    ALOGV("CB_OUTPUT_AVAILABLE index:%d",index);
                    mParent->codecOutputAvailable(
                        index,
                        offset,
                        size,
                        timeUs,
                        flags);
                    break;
                }
                case MediaCodec::CB_OUTPUT_FORMAT_CHANGED:
                {
                    ALOGV("CB_OUTPUT_FORMAT_CHANGED is TODO");
                    break;
                }
                case MediaCodec::CB_ERROR:
                {
                    ALOGE("CB_ERROR is TODO");
                    break;
                }
                default:
                {
                    ALOGE("onMessageReceived kWhatCodecNotify default cbID:%d",cbID);
                    break;
                }
            }

            break;
        }
        default:
        {
            ALOGE("onMessageReceived default");
        }
    }
}

void
MediaCodecDecodeHelper::codecOutputAvailable(
    int32_t index,
    size_t offset,
    size_t size,
    int64_t timeUs,
    int32_t flags)
{
    ALOGV("[%s] index %d",__FUNCTION__,index);

    sp<ABuffer> buffer;
    mCodec->getOutputBuffer((size_t)index, &buffer);

    // get metadata in ABuffer
    sp<AMessage> metadata = buffer->meta();

    // seems phys in ABuffer metadata is invalid.
    int32_t phyAddr = 0;
    metadata->findInt32("RealPhy",&phyAddr);

    CHECK(phyAddr!=0);

    mCallback->onOutputBufferAvailable(
        buffer->data(),
        phyAddr,
        size,
        timeUs);

    // queue buffer index , and another thread will consume this queue
    mOutputBufIdQueue->enqueue(index);
    mOutputBufIdQueue->signal();
}

void
MediaCodecDecodeHelper::codecInputAvailable(int index)
{
    sp<ABuffer> buffer;
    status_t rst = mCodec->getInputBuffer(index, &buffer);

    if(rst != OK) {
        ALOGE("%s getInputBuffer error %.x",
            __FUNCTION__,
            rst);
        return;
    }

    CHECK(buffer.get() != NULL);

    nsecs_t pts = -1;
    size_t range = -1;

    rst = mCallback->onInputBufferAvailable(
        buffer->data(),
        0x0,
        buffer->capacity(),
        &pts,
        &range);

    if(rst != OK) {
        ALOGW("onInputBufferAvailable failed");
        return;
    }

    if(!mDeviceRunning) return;

    CHECK(pts > 0);
    CHECK(range > 0);

    buffer->setRange(0, range);
    buffer->meta()->setInt64("timeUs", pts);

    mCodec->queueInputBuffer(
        index,
        0,
        range,
        pts,
        0);
}

char *
MediaCodecDecodeHelper::getMediaCodecTypeString()
{
    if(mFormat == V4L2_PIX_FMT_MJPEG){
        return "video/x-motion-jpeg";
    }else{
        return "video/avc";
    }
}

status_t
MediaCodecDecodeHelper::createMediaCodec()
{
    CHECK(mCodec == NULL);
    CHECK(mLooper == NULL);
    CHECK(mHandler == NULL);

    status_t err;

    mLooper = new ALooper;
    mLooper->setName("MediaCodecDecodeHelperLooper");
    //mLooper->start(false, false, ANDROID_PRIORITY_AUDIO);
    mLooper->start();
    ALOGV("Creating codec");

    char *type = getMediaCodecTypeString();
    mCodec = MediaCodec::CreateByType(
        mLooper,
        type,
        false,
        NULL,
        0x2379);

    if(mCodec == NULL) {
        ALOGE("create mCodec failed");
        mLooper->stop();
        cleanUpMediaCodec();
        return UNKNOWN_ERROR;
    }

    sp<AMessage> format = new AMessage;
    format->setInt32("width", mWidth);
    format->setInt32("height", mHeight);
    format->setString("mime", type);
    format->setInt32("color-format",OMX_COLOR_FormatYUV420SemiPlanar);
    format->setInt32("realtek-camera", 1);

    mHandler = new MyDecodeHandler(this);
    mLooper->registerHandler(mHandler);

    // set callback.
    mReply = new AMessage(kWhatCodecNotify, mHandler);
    mCodec->setCallback(mReply);

    err = mCodec->configure(format, NULL,NULL,0);

    if(err != NO_ERROR) {
        ALOGE("configure mCodec failed (err=%d)",err);
        mCodec->release();
        mLooper->stop();
        cleanUpMediaCodec();
        return UNKNOWN_ERROR;
    }

    err = mCodec->start();
    if(err != NO_ERROR) {
        ALOGE("start codec error (err=%d)",err);
        mCodec->release();
        mLooper->stop();
        cleanUpMediaCodec();
        return UNKNOWN_ERROR;
    }

    ALOGD("createMediaCodec complete");

    return OK;
}

void
MediaCodecDecodeHelper::cleanUpMediaCodec()
{
    if(mCodec != NULL){
        mCodec.clear();
        mCodec = NULL;
    }

    if(mLooper != NULL){
        mLooper.clear();
        mLooper = NULL;
    }

    if(mHandler != NULL) {
        mHandler.clear();
        mHandler = NULL;
    }

    if(mReply != NULL) {
        mReply.clear();
        mReply = NULL;
    }

}

};
